#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/alias.c,v 1.4 1999/08/07 00:27:02 kc Exp $")

#include "bb.h"

struct alias {
    char *alias;
    char *command;
    boolean parameters;
}    *Aliases = NULL;
int Alias_size = 0, Alias_next = 0;

static int alias_cmp(char *a1, char b1, char *a2, char b2);
static int alias_order(struct alias * ui1, struct alias * ui2);

void alias(char *parameter)
{
    char alias[BUFFER_SIZE], command[BUFFER_SIZE];
    struct alias *p;
    boolean new_alias;
    char params[BUFFER_SIZE];

    strcpy(params, parameter);
    strcpy(alias, getword(params));
    strcpy(command, getline(params, FALSE));
    if (!*command) {
	char format[BUFFER_SIZE];
	int l = 0;

	if (!*alias)
	    for (p = Aliases; p < Aliases + Alias_next; p++)
		l = MAX(l, strlen(bb_lit(p->alias)));
	sprintf(format, "alias %%-%ds %%s\n", l);
	for (p = Aliases; p < Aliases + Alias_next; p++)
	    if (!*alias || (alias_cmp(alias, '*', p->alias, '*') == 0)) {
		strcpy(command, bb_lit(p->command));
		fprintf(cmdout, format, bb_lit(p->alias), command);
	    }
    } else {
	for (p = Aliases; p < Aliases + Alias_next; p++)
	    if (alias_cmp(alias, '*', p->alias, '*') == 0)
		break;
	if (p >= Aliases + Alias_size) {
	    Alias_next = p - Aliases;
	    if (Alias_size == 0) {
		Alias_size = 20;
		Aliases = (struct alias *)
		  malloc(Alias_size * sizeof(*Aliases));
	    } else {
		Alias_size *= 2;
		Aliases = (struct alias *)
		  realloc(Aliases, Alias_size * sizeof(*Aliases));
	    }
	    p = Aliases + Alias_next;
	}
	new_alias = p == Aliases + Alias_next;
	if (new_alias) {
	    Alias_next++;
	    p->alias = (char *)malloc(strlen(alias) + 1);
	    strcpy(p->alias, alias);
	    p->command = (char *)malloc(strlen(command) + 1);
	} else {
	    if (strcmp(p->alias, alias) != 0) {
		fprintf(cmdout, "(alias %s replaced %s)\n", alias, p->alias);
		p->alias = (char *)realloc(p->alias, strlen(alias) + 1);
		strcpy(p->alias, alias);
	    }
	    p->command = (char *)realloc(p->command, strlen(command) + 1);
	}
	strcpy(p->command, command);
	p->parameters = (char *)index(p->command, '!') != NULL;
	if (new_alias)
	    qsort(Aliases, Alias_next, sizeof(*Aliases), alias_order);
    }
}

static int alias_cmp(char *a1, char b1, char *a2, char b2)
{
    boolean a1_done = FALSE, a2_done = FALSE;

    for (; *a1 && *a2; a1++, a2++) {
	if (*a1 == b1) {
	    a2--;
	    a1_done = TRUE;
	    continue;
	}
	if (*a2 == b2) {
	    a1--;
	    a2_done = TRUE;
	    continue;
	}
	if (*a1 != *a2)
	    return (*a1) - (*a2);
    }
    a1_done = a1_done || *a1 == b1;
    a2_done = a2_done || *a2 == b2;
    if ((*a1 && a1_done) || (*a2 && a2_done))
	return 0;
    return (*a1) - (*a2);
}

static int alias_order(struct alias * ui1, struct alias * ui2)
{
        return alias_cmp(ui1->alias, '*', ui2->alias, '*');
}

char *alias_lookup(char *alias, char *parameter)
{
    struct alias *a;
    static char command[BUFFER_SIZE];
    char *p, *q;

    for (a = Aliases; a < Aliases + Alias_next; a++)
	if (alias_cmp(alias, '\0', a->alias, '*') == 0) {
	    boolean got_parameters = FALSE;

	    if (!a->parameters) {
		if (Debug & DEBUG_PARSE)
		    fprintf(Debug_file, "Alias %s -> %s\n", alias, a->command);
		return a->command;
	    }
	    for (p = a->command, q = command; *p; p++, q++) {
		switch (*p) {
		case '!':
		    if (p > a->command && *(p - 1) == '\\') {
			*--q = *p;
		    } else {
			switch (*++p) {
			case '*':
			    got_parameters = TRUE;
			    strcpy(q, getline(parameter, TRUE));
			    while (*q)
				q++;
			    q--;
			    break;
			case '@':{
				boolean again;	/* getline stopped on
						 * redirect--do again */

				got_parameters = TRUE;
				do {
				    strcpy(q, getline(parameter, TRUE));
				    while (*q)
					q++;
				    if ((again = isdirect(*parameter))) {
				    /*
				     * Copy redirection operator to q, shift
				     * out of parameter 
				     */
					char *r;

					*q++ = *parameter;
					for (r = parameter; (*r = *(r + 1)); r++);
				    }
				} while (again);
				q--;
				break;
			    }
			default:
			    *q = *--p;
			}
		    }
		    break;
		default:
		    *q = *p;
		}
	    }
	    if (got_parameters)
		*q++ = ';';
	    *q = '\0';
	    if (Debug & DEBUG_PARSE)
		fprintf(Debug_file, "Alias %s -> %s\n", alias, command);
	    return command;
	}
    return alias;
}

void unalias(char *parameter)
{
    char alias[BUFFER_SIZE];
    struct alias *p;
    char params[BUFFER_SIZE];

    strcpy(params, parameter);
    strcpy(alias, getword(params));
    if (*params) {
	fprintf(stderr, "unexpected input: '%s'\n", params);
	return;
    }
    for (p = Aliases; p < Aliases + Alias_next; p++)
	if (alias_cmp(alias, '*', p->alias, '*') == 0)
	    break;
    if (p == Aliases + Alias_next)
	return;
    free(p->alias);
    free(p->command);
    Alias_next--;
    for (; p < Aliases + Alias_next; p++) {
	p->alias = (p + 1)->alias;
	p->command = (p + 1)->command;
    }
}
