#ifndef VMS
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <strings.h>
#include <fcntl.h>
#else
#include <stdlib.h>
#include <types.h>
#include <stat.h>
#include <time.h>
#include <string.h>
#include <unixio.h>
#endif
#include <ctype.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>

#include "config.h"
#include "../lib/config.h"
#include "../lib/types.h"
#include "../lib/client.h"
#include "../lib/macros.h"
#include "../lib/misc.h"
#include "../lib/config.h"

#ifndef FALSE
#define FALSE	0
#define TRUE	1
#endif

#ifndef OLD_SUMMARY_DEFAULT
#define OLD_SUMMARY_DEFAULT TRUE
#endif

#define BUFFERSIZE 4096
#define BLOCK_SIZE 32768

#define SUMMARY_NUMBER(i) \
	(Old_summary ? Summary[i]->number : Message_numbers[i])
#define SUMMARY(i) \
	(Old_summary ? Summary[i] : \
		       get_message_entry(Room_id, Message_numbers[i]))

extern char *Booleans[2];
extern char *Classes[5];
extern char *Archives[3];
extern header_info **Summary;
extern int *Message_numbers;
extern boolean Old_summary;
extern char Roomname[];
extern int Room_id;
extern int CurrentMessage, Lastmessage, Last_deleted;
extern bb_priority Cutoff_priority;
extern int Terminal_width;
extern char Id[], Password[], Server[];
extern int Port_number;
extern char Nickname[], Mail_address[];
extern char Pager[], Editor[], Mailer[], Header_format[], File_header_format[];
extern char Helpfile[];
extern FILE *Infile, *cmdin, *cmdout, *Page;
extern boolean No_clobber, Snide, Environment, Crypt, Abbrev, Editmode;
extern boolean Redirected_input, Redirected_output;
extern char *Infilename;
extern int Source_level;
extern char lib_error[];
extern int lib_severity;
extern char Path[];
extern boolean Hooks;

extern char *getenv(char *var);
char *Date(time_t date);

int readmessage(int message, boolean permanent);
char *format_message_header(header_info header, char *format);

#define DEBUG_TRANSPORT (1 << 0)
#define DEBUG_PARSE (1 << 1)
#define DEBUG_MISC (1 << 2)

#define isopt(c) ((c) == '/' /* || (c) == '-' */ )
#define isdirect(c) ((c) == '<' || (c) == '>' || (c) == '|')
#define isend(c) ((c) == '\0' || (c) == '\n' || (c) == ';')
#define isquote(c) ((c) == '"')
#define isquotenext(c) ((c) == '\\')
#define iscommand(c) (!isspace(c) && !isopt(c) && !isdirect(c) && \
		      !isend(c) && (c) != '=')

#define crypt_string(src, dest) { \
  register char *p, *q; \
  for(p=src, q=dest; *p;) *q++ = *p++ ^ 0x5a; \
  *q = *p; \
}
#define get_pw(str) crypt_string(Password, str);
#define put_pw(str) crypt_string(str, Password);

boolean streqv(char *s1, char *s2);
void printlist(char **list, int count);
char *get_word(char *input, boolean leave_quotes);
#define getword(input) get_word(input, FALSE)
char *prompt(char *prompt_string, char *input);
char *getline(char *input, boolean leave_quotes);
int ieq(char c1, char c2);
#define match(input, list, count) match_list(input, list, count, TRUE)
int match_list(char *input, char **list, int count, boolean abbrev);
void process_message(char *buf, int buf_l);
void crypt_message(char *buf, int buf_l);
int get_message_number(char *parameter, int default_message);
FILE *page();
void page_end();
char *get_room(char *input);
char *valid_room(char *input);
#define get_valid_room(input) (valid_room(getword(input)))
char *strdup(const char *input);
char *get_group(char *input);
int get_gid(char *group);
group_info *get_group_info(char *group);
char *literal(char *string);
char *bb_lit(char *string);
void selectroom(char *parameter);
void update_summary(int current_message);
char *find_file(char *filename);
void do_connect();
char *group_names(group_set *groups, result_list *group_directory);

#define LIB_FATAL 3
#define LIB_ERROR 2
#define LIB_WARNING 1

void help(char *parameter);
void readcommand(char *parameter);
void back(char *parameter);
void next(char *parameter);
void first(char *parameter);
void all(char *parameter);
void last(char *parameter);
void summary(char *parameter);
void create(char *parameter);
void goroom(char *parameter);
void nextroom(char *parameter);
void backroom(char *parameter);
void nextroomany(char *parameter);
void backroomany(char *parameter);
void directory(char *parameter);
void post(char *parameter);
void respond(char *parameter);
void delete_command(char *parameter);
void bb_remove(char *parameter);
void changeclass(char *parameter);
void change(char *parameter);
void changepriority(char *parameter);
void hide(char *parameter);
void show(char *parameter);
void info(char *parameter);
void setmail(char *parameter);
void setname(char *parameter);
void setnick(char *parameter);
void password(char *parameter);
void users(char *parameter);
void list(char *parameter);
void signon(char *parameter);
void set(char *parameter);
void quit(char *parameter);
void shutdown_command(char *parameter);
void add(char *parameter);
void subtract(char *parameter);
void cutoff(char *parameter);
void preserve(char *parameter);
void bb_rename(char *parameter);
void grouppriority(char *parameter);
void reply(char *parameter);
void mail(char *parameter);
void reset(char *parameter);
void kill_client(char *parameter);
void undelete(char *parameter);
void source(char *parameter);
void alias(char *parameter);
void unalias(char *parameter);
void echo(char *parameter);
void search(char *parameter);
int system(char *parameter);
char *alias_lookup(char *alias, char *parameter);
void catchup(char *parameter);
void run(char *parameter);
