#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/change.c,v 1.4 1999/08/07 00:27:04 kc Exp $")

#include "bb.h"

void change(char *parameter)
{
    char roomname[BUFFER_SIZE] = "";
    directory_info *room_info;
    class_t class = no_class;
    char archive_string[BUFFER_SIZE] = "";
    char class_string[BUFFER_SIZE] = "", expire_string[BUFFER_SIZE] = "";
    char owner[BUFFER_SIZE] = "", summary[BUFFER_SIZE] = "";
    char group[BUFFER_SIZE] = "", priority_s[BUFFER_SIZE] = "";
    time_t expire = 0;
    bb_priority priority = 0;
    bb_archive_t archive;
    boolean input = FALSE;
    char *optnames[] = {"owner", "class_default", "summary", "expiration",
			"group", "priority", "archive"};
    enum {
	owner_f, class_default_f, summary_f, expiration_f, group_f,
	priority_f, archive_f
    };
    int opt, tmp = -1;

    while (*parameter) {
	if (isopt(*parameter)) {
	    char *word;

	    strcpy(parameter, parameter + 1);
	    if ((opt = match(word = getword(parameter),
			     optnames, lengthof(optnames))) != -1) {
		input = input || !(opt == group_f);
		if (*parameter != '=') {
		    fprintf(stderr, "value expected for /%s\n", optnames[opt]);
		    return;
		}
		strcpy(parameter, parameter + 1);
		switch (opt) {
		case archive_f:
		    tmp = match(strcpy(archive_string, getword(parameter)),
				Archives, lengthof(Archives));
		    if (tmp == -1) {
			fprintf(stderr, "no such archive %s\n",
				archive_string);
			return;
		    } else {
			archive = (bb_archive_t)tmp;
		    }
		    break;
		case class_default_f:
		    tmp = match(strcpy(class_string, getword(parameter)),
				Classes, lengthof(Classes) - 1);
		    if (tmp == -1) {
			fprintf(stderr, "no such class %s\n", class_string);
			return;
		    } else {
			class = (class_t)tmp;
		    }
		    break;
		case expiration_f:
		    strcpy(expire_string, getword(parameter));
		    if (*expire_string != '\0') {
		        if (sscanf(expire_string, "%d", (int *)&expire)
			    != 1) {
			    fprintf(stderr, "The expiration must be an integer specifying the default number of days before\na message expires.\n");
			    return;
			}
		    }
		    break;
		case group_f:
		    if (*summary) {
			fprintf(stderr, "/group and /summary are incompatible options\n");
			return;
		    }
		    strcpy(group, get_group(parameter));
		    break;
		case owner_f:
		    strcpy(owner, getword(parameter));
		    break;
		case priority_f:
		    strcpy(priority_s, getword(parameter));
		    if (*priority_s != '\0') {
		        if (sscanf(priority_s, "%d", (int *)&priority) != 1) {
			    fprintf(stderr, "The priority must be an integer specifying the offset priority of the room.\n");
			    return;
			}
		    }
		    break;
		case summary_f:
		    if (*group) {
			fprintf(stderr, "/group and /summary are incompatible options\n");
			return;
		    }
		    strcpy(summary, getword(parameter));
		    break;
		}
	    } else {
		fprintf(stderr, "unrecognized qualifier /%s\n", word);
		return;
	    }
	} else {
	    if (!*roomname) {
		strcpy(roomname, getword(parameter));
	    } else {
		fprintf(stderr, "unexpected input: '%s'\n", parameter);
		return;
	    }
	}
    }

    do_connect();
    if (!*group) {
	if (!*roomname)
	    strcpy(roomname, Roomname);
	if (!(room_info = client_roominfo(roomname))) {
	    fprintf(stderr, "%s: %s\n", roomname, client_error);
	    return;
	}
    }
    if (!input) {
    /* owner */
	prompt("owner's id: ", owner);

    /* class_default */
	printlist(Classes, lengthof(Classes) - 1);
	prompt("default class: ", class_string);
	while (*class_string &&
	       (tmp = match(class_string, Classes, lengthof(Classes) - 1))
	       == -1) {
	    fprintf(stderr, "no such class %s\n", class_string);
	    prompt("default class: ", class_string);
	}
	if (*class_string)
	    class = (class_t)tmp;

    /* summary */
	if (!*group)
	    prompt("summary? ", summary);

    /* expiration */
	prompt("expiration? ", expire_string);
	while (*expire_string &&
	       sscanf(expire_string, "%d", (int *)&expire) != 1) {
	    fprintf(stderr, "The expiration must be an integer specifying the default number of days before\na message expires.\n");
	    prompt("expiration? ", expire_string);
	}
    /* priority */
	prompt("offset priority? ", priority_s);
	while (*priority_s &&
	       sscanf(priority_s, "%d", (int *)&priority) != 1) {
	    fprintf(stderr, "The priority must be an integer specifying the offset priority of the room.\n");
	    prompt("offset priority? ", priority_s);
	}
    /* archive */
	printlist(Archives, lengthof(Archives));
	prompt("archive: ", archive_string);
	while (*archive_string &&
	       (tmp = match(archive_string, Archives, lengthof(Archives)))
	       == -1) {
	    fprintf(stderr, "no such archive %s\n", archive_string);
	    prompt("archive: ", archive_string);
	}
	if (*archive_string)
	    archive = (bb_archive_t)tmp;
    }
    expire *= 86400;
    if (*group) {
	if (client_group_change(group,
				NULL /* new group name */ ,
				*owner ? owner : NULL,
				class != no_class ? &class : NULL,
				(*priority_s ? &priority : NULL),
				*expire_string ? &expire : NULL))
	    fprintf(cmdout, "group %s changed\n", group);
	else
	    fprintf(stderr, "%s\n", client_error);
	return;
    }
    if (client_changeroom(roomname,
			  (*owner ? owner : NULL),
			  (*expire_string ? &expire : NULL),
			  (class != no_class ? &class : NULL),
			  (*summary ? summary : NULL),
			  (*priority_s ? &priority : NULL),
			  (*archive_string ? &archive : NULL)))
	fprintf(cmdout, "room %s changed\n", roomname);
    else
	fprintf(stderr, "%s\n", client_error);
}
