#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/crypt.c,v 1.4 1999/08/07 00:27:05 kc Exp $")

#include <strings.h>

char *crypt(char *key, char *salt);

/*
 *	A one-rotor machine designed along the lines of Enigma
 *	but considerably trivialized.  Based on BSD's crypt.
 */

#define ROTORSZ 256
#define MASK 0377

static char t1[ROTORSZ];
static char t2[ROTORSZ];
static char t3[ROTORSZ];
static char deck[ROTORSZ];
static char buf[13];

int Crypt_secure = 0;

static void shuffle(char deck[]);

void crypt_setup(char *pw)
{
  int ic, i, k, temp;
  unsigned random;
  long seed;

  strncpy(buf, crypt(pw, pw), 13);
  seed = 123;
  for(i=0; i < 13; i++)
    seed = seed*buf[i] + i;
  for(i=0; i < ROTORSZ; i++) {
    t1[i] = i;
    deck[i] = i;
    t2[i] = t3[i] = 0;
  }
  for(i=0; i < ROTORSZ; i++) {
    seed = 5*seed + buf[i%13];
    random = seed % 65521;
    k = ROTORSZ-1 - i;
    ic = (random&MASK)%(k+1);
    random >>= 8;
    temp = t1[k];
    t1[k] = t1[ic];
    t1[ic] = temp;
    if(t3[k]!=0) continue;
    ic = (random&MASK) % k;
    while(t3[ic]!=0) ic = (ic+1) % k;
    t3[k] = ic;
    t3[ic] = k;
  }
  for(i=0;i<ROTORSZ;i++)
    t2[t1[i]&MASK] = i;
}

void crypt_message(char *buf, int buf_l)
{
  register i, n1 = 0, n2 = 0, nr1, nr2 = 0;
  char *p;
  char xdeck[ROTORSZ];

  if(Crypt_secure) bcopy(deck, xdeck, sizeof(xdeck));
  for(p=buf; p < buf+buf_l; p++) {
    i = *p;
    if (Crypt_secure) {
      nr1 = deck[n1]&MASK;
      nr2 = deck[nr1]&MASK;
    } else {
      nr1 = n1;
    }
    i = t2[(t3[(t1[(i+nr1)&MASK]+nr2)&MASK]-nr2)&MASK]-nr1;
    *p = i;
    n1++;
    if(n1==ROTORSZ) {
      n1 = 0;
      n2++;
      if(n2==ROTORSZ) n2 = 0;
      if (Crypt_secure) {
	shuffle(deck);
      } else {
	nr2 = n2;
      }
    }
  }
  if(Crypt_secure) bcopy(xdeck, deck, sizeof(deck));
}

static void shuffle(char deck[])
{
  int i, ic, k, temp;
  unsigned random;
  static long seed = 123;

  for(i=0;i<ROTORSZ;i++) {
    seed = 5*seed + buf[i%13];
    random = seed % 65521;
    k = ROTORSZ-1 - i;
    ic = (random&MASK)%(k+1);
    temp = deck[k];
    deck[k] = deck[ic];
    deck[ic] = temp;
  }
}
