#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/directory.c,v 1.4 1999/08/07 00:27:06 kc Exp $")

#include "bb.h"
#define hidden(x) (!opts[hidden_f] && ((x)->priority < cutoff_priority))

void directory(char *parameter)
{
  int gid = -1;
  result_list *directory;
  directory_info *entry, *last_entry = NULL;
  int i, c, opt, cutoff_priority = Cutoff_priority;
  char *optnames[] = {"full", "hidden", "messages", "count",
		      "priority", "cutoff", "old", "groups"};
  enum {full, hidden_f, messages, count, priority_f, cutoff_f, old_f, group_f};
  boolean opts[lengthof(optnames)];
  char params[BUFFER_SIZE];

  strcpy(params, parameter);
  for(opt=0; opt<lengthof(opts); opt++) opts[opt] = FALSE;
  while(*params) {
    if(isopt(*params)) {
      char *word;
      strcpy(params, params+1);
      if((opt=match(word=getword(params), optnames, lengthof(opts)))!=-1) {
	opts[opt] = TRUE;
	switch(opt) {
	case group_f:
	  if(*params == '=') {
	    opts[group_f] = FALSE;
	    strcpy(params, params+1);
	    gid = get_gid(get_group(params));
	  }
	  break;
	case priority_f:
	  opts[old_f] = TRUE;
	  break;
	case cutoff_f:
	  if(*params != '=') {
	    fprintf(stderr, "value expected for /%s\n", optnames[opt]);
	    return;
	  }
	  strcpy(params, params+1);
	  sscanf(getword(params), "%d", &cutoff_priority);
	  break;
	default:
	  break;
	}
      } else {
	fprintf(stderr, "unrecognized qualifier /%s\n", word);
	return;
      }
    } else {
      fprintf(stderr, "unexpected input: '%s'\n", params);
      return;
    }
  }
  if(opts[group_f] && (opts[messages] || opts[count])) {
    if(opts[count]) fprintf(stderr, "/count is invalid with /group\n");
    if(opts[messages]) fprintf(stderr, "/messages is invalid with /group\n");
    return;
  }
  do_connect();
  directory = opts[group_f] ? client_group_directory() : client_directory();
  if(!directory) {
    fprintf(stderr, "%s\n", client_error);
    return;
  }
  page();
  fprintf(Page, "\n");
  if(opts[group_f]) {
    group_info *entry;
    for(i=0; i<directory->count; i++) {
      entry = (group_info *) directory->list[i];
      fprintf(Page, "%5d %-20s %-12s %-10s %3d day%c  %-10s\n",
	      (int)entry->priority,
	      entry->name,
	      entry->owner,
	      Classes[entry->default_class],
	      entry->max_expiration/=86400,
	      entry->max_expiration == 1 ? ' ' : 's',
	      Classes[entry->class]);
    }
    fprintf(Page, "\n");
    return;
  }
  if(opts[full]) {
    int name_width = 0, priority_width = 0, count_width = 1, owner_width = 0;
    int summary_width = 0, l, w, pointer = -1;
    char fmt[20], count_fmt[20], summary_fmt[20], owner_fmt[20];
    char summary[BUFFER_SIZE];

    for(i=0; i<directory->count; i++) {
      entry = (directory_info *) directory->list[i];
      if(strcmp(entry->name, Roomname)==0) pointer = i;
      if(!hidden(entry) &&
	 (!opts[messages] || entry->new_messages) &&
	 (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	if(opts[priority_f]) {
	  sprintf(fmt, "%d", (int)entry->priority);
	  if((l=strlen(fmt)+1) > priority_width) priority_width = l;
	}
	if(opts[count]) {
	  sprintf(fmt, "%d", (int)entry->new_messages);
	  if((l=strlen(fmt)+2) > count_width) count_width = l;
	}
	if((l=strlen(entry->name)) > name_width) name_width = l;
	if((l=strlen(entry->owner)) > owner_width) owner_width = l;
	if((l=strlen(entry->summary)) > summary_width) summary_width = l;
      }
    }
    summary_width = MIN(summary_width, Terminal_width - 1 -
			priority_width - 1 - count_width - 1 - name_width - 1 -
			1 - owner_width);
    if(opts[priority_f]) sprintf(fmt, "%%%dd ", priority_width-1);
    if(opts[count]) sprintf(count_fmt, "(%%%dd)", count_width-2);
    sprintf(summary_fmt, "%%-%ds", summary_width);
    sprintf(owner_fmt, " %%%ds\n", owner_width);
    for(i=0; i<directory->count; i++) {
      entry = (directory_info *) directory->list[i];
      if(!hidden(entry) &&
	 (!opts[messages] || entry->new_messages) &&
	 (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	  if(opts[priority_f]) {
	    if(!last_entry || entry->priority != last_entry->priority) {
	      fprintf(Page, fmt, entry->priority);
	      last_entry = entry;
	    } else for(w=0; w<priority_width; w++) putc(' ', Page);
	  }
	  if(i == pointer) putc('>', Page);
	  else putc(' ', Page);
	  if(opts[count]) fprintf(Page, count_fmt, entry->new_messages);
	  else putc(entry->new_messages ? '*' : ' ', Page);
	  putc(Hidden(entry) ? '-' : ' ', Page);
	  fprintf(Page, "%s ", entry->name);
	  for(w=strlen(entry->name); w<name_width; w++) putc(' ', Page);
	  strncpy(summary, entry->summary, summary_width);
	  summary[summary_width]='\0';
	  fprintf(Page, summary_fmt, summary);
	  fprintf(Page, owner_fmt, entry->owner);
      }
    }
  } else if(opts[old_f]) {
    int total_new_messages = 0, total_rooms = 0;
    int name_width = 0, priority_width = 0, count_width = 1;
    int l, w, n, pointer = -1;
    char fmt[20], count_fmt[20];

    for(i=0; i<directory->count; i++) {
      entry = (directory_info *) directory->list[i];
      if(strcmp(entry->name, Roomname)==0) pointer = -2;
      if(!hidden(entry) &&
	 (!opts[messages] || entry->new_messages) &&
	 (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	if(pointer==-2) pointer = i;
	total_new_messages += entry->new_messages;
	total_rooms++;
	if(opts[priority_f]) {
	  sprintf(fmt, "%d", (int)entry->priority);
	  if((l=strlen(fmt)+1) > priority_width) priority_width = l;
	}
	if(opts[count]) {
	  sprintf(fmt, "%d", (int)entry->new_messages);
	  if((l=strlen(fmt)+2) > count_width) count_width = l;
	}
	if((l=strlen(entry->name)) > name_width) name_width = l;
      }
    }
    if(opts[priority_f]) sprintf(fmt, "%%s%%%dd ", priority_width-1);
    if(opts[count]) sprintf(count_fmt, "(%%%dd)", count_width-2);
    n = (Terminal_width-priority_width) / (1 + count_width + 1 + name_width);
    if(n <= 0) n = 1;
    for(i=0, c=0; i<directory->count; i++) {
      entry = (directory_info *) directory->list[i];
      if(!hidden(entry) &&
	 (!opts[messages] || entry->new_messages) &&
	 (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	if(opts[priority_f]) {
	  if(!last_entry || entry->priority != last_entry->priority) {
	    fprintf(Page, fmt, c==0 ? "" : "\n", entry->priority);
	    c = 0;
	  } else if(c==0) for(w=0; w<priority_width; w++) putc(' ', Page);
	  last_entry = entry;
	}
	if(i == pointer) putc('>', Page);
	else putc(' ', Page);
	if(opts[count]) fprintf(Page, count_fmt, entry->new_messages);
	else putc(entry->new_messages ? '*' : ' ', Page);
	putc(Hidden(entry) ? '-' : ' ', Page);
	fprintf(Page, "%s", entry->name);
	if(++c >= n) {
	  fprintf(Page, "\n");
	  c = 0;
	} else for(w=strlen(entry->name); w<name_width; w++) putc(' ', Page);
      }
    }
    if (c != 0) fprintf(Page, "\n");
    fprintf(Page,
	    total_new_messages == 1 ? 
	    "\nThere is %d unread message in %d room%s.\n" :
	    "\nThere are %d unread messages in %d room%s.\n",
	    total_new_messages,
	    total_rooms,
	    total_rooms == 1 ? "" : "s");
  } else {
    int total_new_messages = 0, total_rooms = 0;
    int name_width = 0, priority_width = 0, count_width = 1;
    int l, lines, j, w, width, pointer = -1;
    char fmt[20], count_fmt[20];
    directory_info **rooms = (directory_info **)
      malloc(directory->count * sizeof(directory_info *));

    for(i=0; i<directory->count; i++) {
      entry = (directory_info *) directory->list[i];
      if(strcmp(entry->name, Roomname)==0) pointer = total_rooms;
/*    fprintf(Page,
	      "gid=%d, groups=0x%04x, group_isset(gid, &groups) == %d\n",
	      gid, entry->groups, GROUP_ISSET(gid, &entry->groups));
*/    if(!hidden(entry) &&
	 (!opts[messages] || entry->new_messages) &&
	 (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	rooms[total_rooms++] = entry;
	total_new_messages += entry->new_messages;
	if(opts[priority_f]) {
	  sprintf(fmt, "%d", (int)entry->priority);
	  if((l=strlen(fmt)+1) > priority_width) priority_width = l;
	}
	if(opts[count]) {
	  sprintf(fmt, "%d", (int)entry->new_messages);
	  if((l=strlen(fmt)+3) > count_width) count_width = l;
	}
	if((l=strlen(entry->name)) > name_width) name_width = l;
      }
    }
    name_width += 2;
    width = priority_width + 1 + count_width + 1 + name_width;
    c = Terminal_width / width;
    if(c == 0) c = 1;
    lines = (total_rooms + c - 1) / c;
    if(opts[priority_f]) sprintf(fmt, "%%%dd ", priority_width-1);
    if(opts[count]) sprintf(count_fmt, "(%%%dd) ", count_width-3);
    for(i=0; i<lines; i++) {
      for(j=0; j<c; j++) {
	entry = rooms[j*lines+i];
	if(!hidden(entry) &&
	   (!opts[messages] || entry->new_messages) &&
	   (gid == -1 || GROUP_ISSET(gid, &entry->groups))) {
	  if(opts[priority_f]) {
	    if(i==0 ||
	       entry->priority != (rooms[j*lines+i-1])->priority)
	      fprintf(Page, fmt, entry->priority);
	    else for(w=0; w<priority_width; w++) putc(' ', Page);
	  }
	  if(j*lines+i == pointer) putc('>', Page);
	  else putc(' ', Page);
	  if(opts[count]) fprintf(Page, count_fmt, entry->new_messages);
	  else putc(entry->new_messages ? '*' : ' ', Page);
	  putc(Hidden(entry) ? '-' : ' ', Page);
	  fprintf(Page, "%s", entry->name);
	  if((j+1)*lines+i >= total_rooms) {
	    fprintf(Page, "\n");
	    break;
	  }
	  for(w=strlen(entry->name); w<name_width; w++) putc(' ', Page);
	}
      }
    }
    free(rooms);
    fprintf(Page,
	    total_new_messages == 1 ? 
	    "\nThere is %d unread message in %d room%s.\n" :
	    "\nThere are %d unread messages in %d room%s.\n",
	    total_new_messages,
	    total_rooms,
	    total_rooms == 1 ? "" : "s");
  }
  fprintf(Page, "\n");
}
