#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/main.c,v 1.5 1999/08/07 00:27:10 kc Exp $")

/*
 * bb.c - client program for bb
 */

#include "bb.h"
#include "../lib/misc.h"
#include "../lib/protocol.h"
#ifndef VMS
#include <sys/ioctl.h>
#else
#include <descrip.h>
#include <smgdef.h>
#define signal_t void
#endif

typedef struct {
    char *command;
    void (*function) (char *parameter);
    boolean connect;
} command_t;

/* GLOBALS */

command_t Commands[] = {
			{"help", help},
			{"?", help},

			{"read", readcommand},
			{"next", next},
			{"back", back},
			{"first", first},
			{"last", last},
			{"all", all},
			{"reset", reset},
			{"summary", summary},

			{"directory", directory},
			{"goto", goroom},
			{"+", nextroomany},
			{"-", backroomany},
			{"priority", changepriority},
			{"cutoff", cutoff},
			{"hide", hide},
			{"show", show},

			{"post", post},
			{"respond", respond},
			{"reply", reply},
			{"delete", delete_command},
			{"preserve", preserve},

			{"create", create},
			{"remove", bb_remove},
			{"rename", bb_rename},
			{"change", change},
			{"changeclass", changeclass},
			{"list", list},

			{"grouppriority", grouppriority},
			{"add", add},
			{"subtract", subtract},

			{"info", info},
			{"mail", mail},
			{"set", set},
			{"setmail", setmail},
			{"setnick", setnick},
			{"setname", setname},
			{"login", signon},
			{"password", password},
			{"users", users},
			{"quit", quit},
			{"exit", quit},
			{"goodbye", quit},
			{"bye", quit},
			{"end", quit},
			{"undelete", undelete},
			{"expire", preserve},
			{"source", source},
			{"run", run},
			{"shutdown", shutdown_command},
			{"kill", kill_client},
			{"alias", alias},
			{"unalias", unalias},
			{"echo", echo},
			{"search", search},
			{"signon", signon},
			{"signoff", quit},
			{"logout", quit},
			{"catchup", catchup},
			{NULL, NULL}
};

char *Classes[] =
{"no_access", "read_only", "read_write", "moderator", "no_class"};
char *Booleans[] = {"false", "true"};
char *Archives[] = {"common", "mail", "none"};

char Bbrc[BUFFERSIZE] = "";
result_list Commandlist;
header_info **Summary = NULL;
int *Message_numbers = NULL;
char Roomname[ROOM_SIZE] = "";
int Room_id = -1;
int CurrentMessage = 0;
int Lastmessage = 0;
int Terminal_width = 80;
int Terminal_rows = 24;
boolean Redirected_input = FALSE;
boolean Redirected_output = FALSE;
static boolean Piped_output = FALSE;
jmp_buf Main_loop_env;
FILE *Infile = NULL;
FILE *cmdin = NULL;
FILE *cmdout = NULL;
char *Infilename = "(standard input)";
extern int Port_number;
FILE *Prog_cmdin = stdin;
FILE *Prog_cmdout = stdout;
boolean Prog_redirected_input = FALSE;
boolean Prog_redirected_output = FALSE;

#define Cmdout_Close { \
  if(Page == cmdout) page_end(); \
  if(Piped_output) { \
    pclose(cmdout); \
    Piped_output = FALSE; \
  } else fclose(cmdout); \
}

#ifdef VMS
int Virtual_Keyboard = -1, Key_Table;
#endif VMS

void getterm()
{
#ifdef VMS
    $DESCRIPTOR(dev, "SYS$COMMAND");

    if (Virtual_Keyboard == -1) {
	smg$create_virtual_keyboard(&Virtual_Keyboard, &dev);
	smg$create_key_table(&Key_Table);
    }
#else
#include <sgtty.h>
#ifdef TIOCGWINSZ
    struct winsize w;

#else
#ifdef WIOCGETD
    struct uwdata w;

#endif WIOCGETD
#endif TIOCGWINSZ
#ifdef AIX
    static char lines_env[BUFFER_SIZE], columns_env[BUFFER_SIZE];

#endif AIX

#ifdef TIOCGWINSZ
    if (ioctl(2, TIOCGWINSZ, &w) == 0 && w.ws_col > 0)
	Terminal_width = w.ws_col;
    else
#else
#ifdef WIOCGETD
    if (ioctl(2, WIOCGETD, &w) == 0 && w.uw_width > 0)
	Terminal_width = w.uw_width / w.uw_hs;
    else
#endif WIOCGETD
#endif TIOCGWINSZ
    {
	char buf[1024];

	if (tgetent(buf, getenv("TERM")) == 1)
	    Terminal_width = tgetnum("co");
    }
    if (Terminal_width <= 0)
	Terminal_width = 80;
#ifdef TIOCGWINSZ
    if (ioctl(2, TIOCGWINSZ, &w) == 0 && w.ws_row > 0)
	Terminal_rows = w.ws_row;
    else
#else
#ifdef WIOCGETD
    if (ioctl(2, WIOCGETD, &w) == 0 && w.uw_height > 0)
	Terminal_rows = w.uw_height / w.uw_vs;
    else
#endif WIOCGETD
#endif TIOCGWINSZ
    {
	char buf[1024];

	if (tgetent(buf, getenv("TERM")) == 1)
	    Terminal_rows = tgetnum("li");
    }
    if (Terminal_rows <= 0)
	Terminal_rows = 24;
#ifdef AIX
    sprintf(lines_env, "LINES=%d", Terminal_rows);
    putenv(lines_env);
    sprintf(columns_env, "COLUMNS=%d", Terminal_width);
    putenv(columns_env);
#endif AIX
#endif VMS
}

#ifdef SIGWINCH
signal_t winch()
{
#ifdef VMS
    signal(SIGWINCH, (signal_t (*) (int,...))winch);
#else
    signal(SIGWINCH, winch);
#endif VMS
    getterm();
    return (signal_t)0;
}

#endif SIGWINCH

signal_t interrupt()
{
#ifdef VMS
    signal(SIGINT, (signal_t (*) (int,...))interrupt);
#else
    signal(SIGINT, interrupt);
    signal(SIGTSTP, SIG_DFL);
#endif VMS
    printf("\n");
    if (Infile != stdin) {
	printf("aborting %s\n", Infilename);
	source("/return");
    }
    longjmp(Main_loop_env, 0);
    return (signal_t)0;
}

signal_t die()
{
    printf("\n");
    server_close();
    exit(0);
    return (signal_t)0;
}

void exec_command(char *);

main(int argc, char **argv)
{
    char input_line[BUFFERSIZE], prompt_s[BUFFERSIZE];
    int i, ch;
    char *bbrc = ".bbrc";
    boolean readrc = TRUE, tty = FALSE;
    int connect_tries = 0;

    extern char *optarg;

    printf("Welcome to BB.\n\n");
    getterm();
    set("/reset");

 /*
  * Read in flags 
  */

    while ((ch = getopt(argc, argv, "F:S:dfp:s")) != EOF)
	switch (ch) {
	case 'F':
	    bbrc = optarg;
	    break;
	case 'S':
	    strcpy(Server, optarg);
	    break;
	case 'd':
	    Debug = DEBUG_TRANSPORT | DEBUG_PARSE | DEBUG_MISC;
	    break;
	case 'f':
	    readrc = FALSE;
	    break;
	case 'p':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid port number %s.\n",
			    optarg);
		else
		    Port_number = tmp;
		break;
	    }
	case 's':
	    Snide = TRUE;
	    break;
	}

 /*
  * Setups 
  */

    Commandlist.list = (void **)malloc(lengthof(Commands) * sizeof(void *));
    for (i = 0; Commands[i].command != NULL; i++) {
	Commandlist.list[i] = Commands[i].command;
	Commandlist.count = i + 1;
    }

    if (!cmdin)
	cmdin = stdin;
    if (!cmdout)
	cmdout = stdout;

    if (!server_open(Server, Port_number)) {
	fprintf(stderr, "%s\n", client_error);
	return 1;
    }
 /*
  * Read from initialization file 
  */

    if (readrc) {
#ifndef VMS
	if (*bbrc == '/' || *bbrc == '~')
	    strcpy(Bbrc, bbrc);
	else
	    sprintf(Bbrc, "~/%s", bbrc);
#else
	sprintf(Bbrc, "%s%s", getenv("HOME"), bbrc);
#endif VMS
	if (Infile = fopen(tilde(Bbrc), "r")) {
	    fclose(Infile);
	    chmod(Bbrc, 0600);
	    source(Bbrc);
	} else {
	    help(".no_init_file");
	}
    }
    if (!Infile)
	Infile = stdin;
    cmdin = stdin;
    cmdout = stdout;

 /*
  * Main loop 
  */

    setjmp(Main_loop_env);
#ifdef VMS
    signal(SIGQUIT, (signal_t (*) (int,...))interrupt);
    signal(SIGPIPE, (signal_t (*) (int,...))SIG_IGN);
    signal(SIGINT, (signal_t (*) (int,...))interrupt);
#ifdef SIGWINCH
    signal(SIGWINCH, (signal_t (*) (int,...))winch);
#endif SIGWINCH
#else
    signal(SIGQUIT, interrupt);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGINT, interrupt);
#ifdef SIGWINCH
    signal(SIGWINCH, winch);
#endif SIGWINCH
#endif VMS

    for (;;) {
	tty = isatty(fileno(Infile));
	if (tty && !*Id)
	    signon("");
    /*
     * Input loop 
     */
	connect_tries = 0;
	if (Infile == stdin)
	    if (*Roomname)
		sprintf(prompt_s, "bb-%s:%d/%d> ",
			Roomname, CurrentMessage, Lastmessage);
	    else
		sprintf(prompt_s, "bb> ");

	cmdin = Infile;
	if (!prompt(prompt_s, input_line)) {
	    if (feof(Infile)) {
		cmdin = stdin;
		source("/return");
		longjmp(Main_loop_env, 0);
	    } else {
#ifdef READLINE
		if (Editmode)
		    printf("\n");
#endif READLINE
		quit("");
	    }
	}
	cmdin = stdin;

#ifdef READLINE
	if (tty && *input_line)
	    add_history(input_line);
#endif READLINE
	exec_command(input_line);
    }
}

void exec_command(char *input_line)
{
    int i, which;
    char command[BUFFERSIZE], parameters[BUFFERSIZE], keyword[BUFFERSIZE];
    char *room;

    extern void *Aliases;

    for (i = 0;; i++) {
	if (cmdin && cmdin != Prog_cmdin)
	    fclose(cmdin);
	if (cmdout && cmdout != Prog_cmdout)
	    Cmdout_Close;
	cmdin = Prog_cmdin;
	Redirected_input = Prog_redirected_input;
	cmdout = Prog_cmdout;
	Redirected_output = Prog_redirected_output;
	if (i && !*input_line)
	    break;
	if (Aliases) {
	    strcpy(command, input_line);
	    sprintf(input_line, "%s %s",
		    alias_lookup(get_word(command, TRUE), command), command);
	}
	strcpy(command, getline(input_line, TRUE));
	while (isdirect(*input_line)) {
	    const char *file;

	    switch (*input_line) {
	    case '<':
		Redirected_input = TRUE;
		strcpy(input_line, input_line + 1);
		file = getline(input_line, FALSE);
		if (cmdin && cmdout) {
		    if (cmdin != Prog_cmdin)
			fclose(cmdin);
		    if (!*file) {
			cmdin = Prog_cmdin;
		    } else if (!(cmdin = fopen(file = tilde(file), "r"))) {
			perror(file);
		    }
		}
		break;
	    case '>':
		Redirected_output = TRUE;
		strcpy(input_line, input_line + 1);
		if (*input_line == '>') {
		    strcpy(input_line, input_line + 1);
		    file = getline(input_line, FALSE);
		    if (cmdin && cmdout) {
			if (cmdout != Prog_cmdout)
			    Cmdout_Close;
			if (!*file) {
			    cmdout = Prog_cmdout;
			} else if (!(cmdout =
				     fopen(file = tilde(file), "a"))) {
			    perror(file);
			}
		    }
		} else {
		    file = getline(input_line, FALSE);
		    if (cmdin && cmdout) {
			if (cmdout != Prog_cmdout)
			    Cmdout_Close;
			if (!*file) {
			    cmdout = Prog_cmdout;
			} else if (No_clobber &&
				   (cmdout =
				    fopen(file = tilde(file), "r"))) {
			    fprintf(stderr, "No_clobber set and %s exists.\n",
				    file);
			    fclose(cmdout);
			    cmdout = NULL;
			} else if (!(cmdout =
				     fopen(file = tilde(file), "w"))) {
			    perror(file);
			}
		    }
		}
		break;
	    case '|':
		Piped_output = Redirected_output = TRUE;
		strcpy(input_line, input_line + 1);
		file = getline(input_line, FALSE);
		if (cmdin && cmdout) {
		    if (cmdout != Prog_cmdout)
			Cmdout_Close;
		    if (!*file) {
			Piped_output = FALSE;
			cmdout = Prog_cmdout;
		    } else if (!(cmdout = popen(file = tilde(file), "w"))) {
			perror(file);
		    } else {
			setlinebuf(cmdout);
		    }
		}
		break;
	    }
	}
	if (!cmdin || !cmdout)
	    continue;
	strcpy(parameters, command);
	strcpy(keyword, getword(parameters));
	if (i && !*keyword)
	    continue;

	which = match(keyword, (char **)Commandlist.list, Commandlist.count);
	if (which != -1) {
	    if (Commands[which].function)
		Commands[which].function(parameters);
	} else if (isend(*keyword)) {
	    if (Infile == stdin)
		next("");
	} else if (*keyword == '!') {
	    if (*parameters)
		sprintf(command, "%s %s", keyword + 1,
			getline(parameters, FALSE));
	    else
		strcpy(command, keyword + 1);
	    system(command);
	} else if (*keyword == '#' && Infile != stdin) {
	    *input_line = '\0';
	} else if (get_message_number(strdup(keyword), CurrentMessage) != -1 ||
		   lib_severity != LIB_FATAL) {
	    readcommand(command);
	} else if (!*parameters && (room = valid_room(keyword))) {
	    selectroom(room);
	} else if (Infile == stdin) {
	    if (Snide)
		printf("You know, there IS help.\n");
	    else
		printf("Unrecognized command.  Use 'help' to get a list of recognized commands.\n");
	} else
	    printf("Unrecognized command in source file (%s)\n- %s.\n",
		   Infilename, command);
    }
}
