#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/bb/run.c,v 1.5 1999/08/07 00:27:15 kc Exp $")

/*
 * Copyright 1991 by Greg Titus. 
 *
 * Do with this as you will, but it'd be nice if you told me about any
 * uses you find for it, enhancements, etc.
 *
 */

#include "bb.h"

typedef struct _var {
    char *name;
    char *value;
    struct _var *next;
} VAR;

typedef struct {
    VAR *var;
    char type;
    int number;
    result_list *room_list;
} FORDATA;

#define FORDATA_ROOMS    1
#define FORDATA_MESSAGES 2
#define FORDATA_NEW      3
#define FORDATA_LIST     4

typedef struct _stack {
    long position;
    char type;
    FORDATA *data;
    struct _stack *next;
} STACK;

typedef struct run_record RUN_RECORD;
struct run_record {
    RUN_RECORD *prev;
    STACK *top;
    VAR *temps;
    char *params[10];
    char *all_params;
    FILE *cmdin;
    FILE *cmdout;
    boolean redirected_input;
    boolean redirected_output;
    FILE *prog_cmdin;
    FILE *prog_cmdout;
    boolean prog_redirected_input;
    boolean prog_redirected_output;
};

RUN_RECORD *current_run = 0;
STACK *stack = 0;
VAR *temps = 0;
VAR *globals = 0;
char *params[10];
char *all_params;

extern FILE *Prog_cmdin;
extern FILE *Prog_cmdout;
extern boolean Prog_redirected_input;
extern boolean Prog_redirected_output;

#define TYPE_NULL  0
#define TYPE_RUN   1
#define TYPE_SKIP  2
#define TYPE_FOR   3
#define TYPE_WHILE 4
#define TYPE_DO    5

static void push(t, p, d)
    char t;
    long p;
    void *d;
{
    STACK *s;

    s = (STACK *) malloc(sizeof(STACK));
    s->type = t;
    s->position = p;
    s->data = d;
    s->next = stack;
    stack = s;
}

static void pop()
{
    STACK *s;
    VAR *v;
    int i;

    if (!stack) {
	fprintf(stderr, "program: tried to pop while nothing on the stack\n");
	return;
    }
    s = stack;
    stack = s->next;

    if (s->data) {
	if (s->data->var == temps)
	    temps = temps->next;
	else {
	    for (v = temps; v->next != s->data->var; v = v->next);
	    v->next = s->data->var->next;
	}
	if (s->data->type == FORDATA_LIST) {
	    for (i = 0; i < s->data->room_list->count; i++)
		free(s->data->room_list->list[i]);
	    free(s->data->room_list->list);
	    free(s->data->room_list);
	}
	free(s->data->var->name);
	free(s->data->var->value);
	free(s->data->var);
	free(s->data);
    }
    free(s);
}

static void push_run_record()
{
    int i;
    RUN_RECORD *rr = (RUN_RECORD *) malloc(sizeof(RUN_RECORD));

    rr->top = stack;
    stack = 0;
    rr->temps = temps;
    temps = 0;
    for (i = 0; i < 10; i++) {
	rr->params[i] = params[i];
    }
    rr->all_params = all_params;
    rr->cmdin = cmdin;
    rr->cmdout = cmdout;
    rr->redirected_input = Redirected_input;
    rr->redirected_output = Redirected_output;
    rr->prog_cmdin = Prog_cmdin;
    rr->prog_cmdout = Prog_cmdout;
    rr->prog_redirected_input = Prog_redirected_input;
    rr->prog_redirected_output = Prog_redirected_output;
    rr->prev = rr;
    current_run = rr;
}

static void pop_run_record()
{
    VAR *p;
    int i;
    RUN_RECORD *rr;

    while (stack)
	pop();

    while (temps) {
	p = temps->next;
	free(temps->value);
	free(temps->name);
	free(temps);
	temps = p;
    }
    rr = current_run;
    stack = rr->top;
    temps = rr->temps;
    for (i = 0; i < 10; i++)
	params[i] = rr->params[i];
    all_params = rr->all_params;
    cmdin = rr->cmdin;
    cmdout = rr->cmdout;
    Redirected_input = rr->redirected_input;
    Redirected_output = rr->redirected_output;
    Prog_cmdin = rr->prog_cmdin;
    Prog_cmdout = rr->prog_cmdout;
    Prog_redirected_input = rr->prog_redirected_input;
    Prog_redirected_output = rr->prog_redirected_output;
    current_run = rr->prev;
    free(rr);
}

static void run_bb_program();

void run(char *parameters)
{
    push_run_record();
    Prog_cmdin = cmdin;
    Prog_cmdout = cmdout;
    Prog_redirected_input = Redirected_input;
    Prog_redirected_output = Redirected_output;
    run_bb_program(parameters);
    pop_run_record();
}

static char *bbvariables();
static char *setvariables();

static int expand_macro();

static int letcmd();
static int forcmd();
static int endcmd();
static int ifcmd();
static int elsecmd();
static int ficmd();
static int promptcmd();
static int globalcmd();
static int returncmd();
static int whilecmd();
static int docmd();
static int untilcmd();

typedef struct _command {
    char *name;
    int (*func) ();
}        COMMAND;

static COMMAND ProgramCommands[] = {
				    {"let", letcmd},
				    {"for", forcmd},
				    {"end", endcmd},
				    {"if", ifcmd},
				    {"else", elsecmd},
				    {"fi", ficmd},
				    {"prompt", promptcmd},
				    {"global", globalcmd},
				    {"return", returncmd},
				    {"while", whilecmd},
				    {"do", docmd},
				    {"until", untilcmd},
				    {NULL, NULL}};

static result_list *room_directory, *group_directory;

void run_bb_program(char *parameter)
{
    int t;
    char *cp, *bp;
    char buf[BUFFER_SIZE];
    char parambuf[BUFFER_SIZE];
    char a_params[BUFFER_SIZE];
    char vbuf[25];
    COMMAND *comp;
    FILE *f;
    char dotbuf[25];
    int active, tmp;
    char activeroom[25];
    int roomindex;
    directory_info *rinfo;
    long file_position;

    room_directory = client_directory();
    group_directory = client_group_directory();
    strcpy(parambuf, parameter);
    cp = parambuf;
    while (*cp == ' ')
	cp++;
    while (*cp != ' ' && *cp != '\0')
	cp++;
    if (cp)
	strcpy(a_params, ++cp);
    else
	*a_params = '\0';
    all_params = a_params;
    cp = parambuf;
    for (tmp = 0; tmp < 10; tmp++)
	params[tmp] = 0;
    tmp = 0;
    while (1) {
	while (*cp == ' ')
	    cp++;
	params[tmp++] = cp;
	while (*cp != ' ' && *cp != '\0')
	    cp++;
	if (*cp == '\0')
	    break;
	*cp++ = 0;
    }

    if (!**params) {
	cp = *params;
	prompt("program: ", cp);
	if (!*cp)
	    return;
    }
    if ((f = fopen(find_file(*params), "r")) == NULL) {
	fprintf(stderr, "Couldn't find program file.\n");
	return;
    }
    while (1) {
	file_position = ftell(f);
	bp = buf;
	while ((t = getc(f)) != '\n' && t != EOF) {
	    if (t == '$') {
		if ((t = getc(f)) != '{') {
		    *bp++ = '$';
		    *bp++ = t;
		    continue;
		}
		active = CurrentMessage - 1;
		strcpy(activeroom, Roomname);
		roomindex = match_list(activeroom,
				       (char **)room_directory->list,
				       room_directory->count, FALSE);
		if (roomindex == -1) {
		    rinfo = NULL;
		} else {
		    rinfo = (directory_info *) room_directory->list[roomindex];
		}
		while (1) {
		    cp = vbuf;
		    t = getc(f);
		    while (t != '}' && t != '.') {
			*cp++ = t;
			t = getc(f);
		    }
		    *cp = 0;
		    if (t != '.')
			break;
		    if (*vbuf == '"') {
			if (isalpha(vbuf[1])) {
			    for (cp = vbuf + 1; *cp != '"' && *cp; cp++);
			    *cp = 0;
			    strcpy(activeroom, vbuf + 1);
			    roomindex = match_list(activeroom,
					      (char **)room_directory->list,
						   room_directory->count,
						   FALSE);
			    if (roomindex == -1) {
				rinfo = NULL;
			    } else {
				rinfo = (directory_info *)
				  room_directory->list[roomindex];
			    }
			} else if (isdigit(*vbuf)) {
			    sscanf(vbuf, "%d", &active);
			    active--;
			} else if (*vbuf == '#') {
			    sscanf(vbuf, "#%d", &tmp);
			    for (active = 0; active <= Lastmessage; active++) {
				if (SUMMARY_NUMBER(active) == tmp) {
				    tmp = -1;
				    break;
				}
			    }
			    if (tmp != -1) {
				fprintf(stderr, "%s%s",
				      "program: dot-macro expands to post ",
					"which doesn't exist\n");
				return;
			    }
			}
		    } else {
			cp = dotbuf;
			if (!stack || (stack->type != TYPE_SKIP &&
				       stack->type != TYPE_NULL)) {
			    if (!expand_macro(vbuf, &cp, active, rinfo))
				return;
			}
			*cp = 0;
			if (isdigit(*dotbuf)) {
			    sscanf(dotbuf, "%d", &active);
			    active--;
			} else {
			    strcpy(activeroom, dotbuf);
			    roomindex = match_list(activeroom,
					      (char **)room_directory->list,
						   room_directory->count,
						   FALSE);
			    if (roomindex == -1) {
				rinfo = NULL;
			    } else {
				rinfo = (directory_info *)
				  room_directory->list[roomindex];
			    }
			}
		    }
		}
		if (!stack || (stack->type != TYPE_SKIP &&
			       stack->type != TYPE_NULL)) {
		    if (!expand_macro(vbuf, &bp, active, rinfo))
			return;
		}
	    } else
		*bp++ = t;
	}
	if (t == EOF)
	    break;

    /* have a full line now. */
	*bp = 0;
	if (!*buf)
	    continue;

	bp = buf;
	while (*bp == ' ' || *bp == '\t')
	    bp++;
	cp = vbuf;
	while (isalpha(*bp))
	    *cp++ = *bp++;
	*cp = 0;
	for (comp = ProgramCommands; comp->name != NULL; comp++) {
	    if (!strcmp(comp->name, vbuf)) {
		if (!(comp->func(++bp, f, file_position)))
		    return;
		break;
	    }
	}
	if (!comp->name && (!stack || (stack->type != TYPE_SKIP &&
				       stack->type != TYPE_NULL))) {
	    exec_command(buf);
	}
    }
    fclose(f);
}

int comparison(p)
    char *p;
{
    char buf1[50];
    char buf2[50];
    char op, *ptr;
    int i;

    if (sscanf(p, " \"%[^\"]\" %c= \"%[^\"]\"", buf1, &op, buf2) != 3 ||
	(op != '!' && op != '=')) {
	return -1;
    }
    for (ptr = buf1; *ptr; ptr++)
	if (isalpha(*ptr) && islower(*ptr))
	    *ptr = toupper(*ptr);
    for (ptr = buf2; *ptr; ptr++)
	if (isalpha(*ptr) && islower(*ptr))
	    *ptr = toupper(*ptr);

    i = strcmp(buf1, buf2);
    if ((!i && op == '!') || (i && op == '='))	/* if is false */
	return 0;
    else
	return 1;
}


static int whilecmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    int i;

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL))
	return 1;

    i = comparison(p);

    if (i == -1) {
	fprintf(stderr, "program: syntax error during 'while'\n");
	return 0;
    }
    if (i) {
	push(TYPE_WHILE, fp, 0);
    } else
	push(TYPE_NULL, 0, 0);
    return 1;
}

static int promptcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    VAR *v;
    char buf1[50], buf2[100], buf3[100];

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL))
	return 1;

    if (sscanf(p, "%s \"%[^\"]\"", buf1, buf2) != 2) {
	fprintf(stderr, "program: syntax error during 'prompt'\n");
	return 0;
    }
    prompt(buf2, buf3);

    v = (VAR *) malloc(sizeof(VAR));
    v->next = temps;
    v->name = (char *)malloc(strlen(buf1) + 1);
    strcpy(v->name, buf1);
    v->value = (char *)malloc(strlen(buf3) + 1);
    strcpy(v->value, buf3);
    temps = v;

    return 1;
}

static int docmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL)) {
	push(TYPE_NULL, 0, 0);
	return 1;
    }
    push(TYPE_DO, ftell(f), 0);
    return 1;
}

static int untilcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    int i;

    if (!stack || stack->type != TYPE_DO) {
        if (stack->type == TYPE_NULL) {
	    pop();
	    return 1;
	} else {
	    fprintf(stderr, "program: 'until' without matching 'do'\n");
	    return 0;
	}
    }
    i = comparison(p);
    if (i == -1) {
	fprintf(stderr, "program: syntax error during 'until'\n");
	return 0;
    }
    if (i) {
	pop();
    } else {
	fseek(f, stack->position, 0);
    }
    return 1;
}

static int letcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    int i, j;
    VAR *v;
    int status = 1;
    char buf1[50], buf2[50], buf3[50], buf4[50];

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL))
	return 1;

    if (sscanf(p, " %s = \"%[^\"]\" %s \"%[^\"]\"",
	       buf1, buf2, buf3, buf4) != 4) {
	if (sscanf(p, " %s = \"%[^\"]\"", buf1, buf3) != 2) {
	    fprintf(stderr, "program: syntax error during 'let'\n");
	    return 0;
	}
    } else {
	if (sscanf(buf2, "%d", &i) != 1)
	    status = 0;
	if (sscanf(buf4, "%d", &j) != 1)
	    status = 0;
	if (!strcmp(buf3, "+") && status) {
	    sprintf(buf3, "%d", i + j);
	} else if (!strcmp(buf3, "-") && status) {
	    sprintf(buf3, "%d", i - j);
	} else if (!strcmp(buf3, "*") && status) {
	    sprintf(buf3, "%d", i * j);
	} else if (!strcmp(buf3, "/") && status) {
	    sprintf(buf3, "%d", i / j);
	} else if (!strcmp(buf3, "%") && status) {
	    sprintf(buf3, "%d", i % j);
	} else if (!strcmp(buf3, "cat")) {
	    sprintf(buf3, "%s%s", buf2, buf4);
	} else {
	    fprintf(stderr, "program: unknown keyword '%s' in 'let'.\n", buf3);
	    return 0;
	}
    }

    for (v = temps; v && strcmp(v->name, buf1); v = v->next);
    if (!v) {
	for (v = globals; v && strcmp(v->name, buf1); v = v->next);
    }
    if (v) {
	free(v->value);
    } else {
	v = (VAR *) malloc(sizeof(VAR));
	v->name = (char *)malloc(strlen(buf1));
	strcpy(v->name, buf1);
	v->next = temps;
	temps = v;
    }
    v->value = (char *)malloc(strlen(buf3));
    strcpy(v->value, buf3);

    return 1;
}

static int globalcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    char buf1[50];
    VAR *v;

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL))
	return 1;

    sscanf(p, " %s", buf1);
    v = (VAR *) malloc(sizeof(VAR));
    v->next = globals;
    v->value = (char *)malloc(1);
    *v->value = 0;
    v->name = (char *)malloc(strlen(buf1));
    strcpy(v->name, buf1);
    globals = v;
    return 1;
}

static int returncmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    char buf[50];

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL)) {
	return 1;
    }
    if (1 != sscanf(p, " \"%[^\"]\"", buf))
	return 0;

    if (current_run) {
	VAR *v = (VAR *) malloc(sizeof(VAR));

	v->name = (char *)malloc(strlen(params[0]) + 1);
	v->value = (char *)malloc(strlen(buf) + 1);
	v->next = current_run->temps;
	current_run->temps = v;
	strcpy(v->name, params[0]);
	strcpy(v->value, buf);
    }
    return 0;
}

static int ifcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    int i;

    if (stack && (stack->type == TYPE_SKIP || stack->type == TYPE_NULL)) {
	push(TYPE_NULL, 0, 0);
	return 1;
    }
    i = comparison(p);
    if (i == -1) {
	fprintf(stderr, "program: syntax error during 'if'\n");
	return 0;
    }
    if (i)
	push(TYPE_RUN, 0, 0);
    else
	push(TYPE_SKIP, 0, 0);
    return 1;
}

static int elsecmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    if (stack && stack->type == TYPE_NULL)
	return 1;

    if (!stack || stack->position || (stack->type != TYPE_SKIP &&
				      stack->type != TYPE_RUN)) {
	fprintf(stderr, "program: 'else' without matching 'if'\n");
	return 0;
    }
    if (stack->type == TYPE_RUN)
	stack->type = TYPE_SKIP;
    else
	stack->type = TYPE_RUN;
    stack->position = 1;
    return 1;
}

static int ficmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    if (stack && (stack->type == TYPE_RUN || stack->type == TYPE_SKIP ||
		  stack->type == TYPE_NULL)) {
	pop();
	return 1;
    }
    fprintf(stderr, "program: 'fi' without matching 'if'\n");
    return 0;
}

static int forcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    char vbuf[30], buf[50], *pc;
    VAR *v;

    if (stack && (stack->type == TYPE_NULL || stack->type == TYPE_SKIP)) {
	push(TYPE_NULL, 0, 0);
	return 1;
    }
    if (sscanf(p, " %s in %s", vbuf, buf) != 2) {
	fprintf(stderr, "program: syntax error, format is: for <var> in <list>\n");
	return 0;
    }
    push(TYPE_FOR, ftell(f), (void *)malloc(sizeof(FORDATA)));
    stack->data->number = 0;
    v = (VAR *) malloc(sizeof(VAR));
    v->value = (char *)malloc(30);
    v->name = (char *)malloc(strlen(vbuf) + 1);
    strcpy(v->name, vbuf);
    stack->data->var = v;
    if (!strcmp(buf, "rooms")) {
	stack->data->type = FORDATA_ROOMS;
	stack->data->room_list = room_directory;
	strcpy(v->value, ((directory_info *)
			  stack->data->room_list->list[0])->name);
	if (stack->data->room_list->count == 0)
	    stack->type = TYPE_NULL;
    } else if (!strcmp(buf, "messages")) {
	stack->data->type = FORDATA_MESSAGES;
	strcpy(v->value, "1");
	if (Lastmessage == 0)
	    stack->type = TYPE_NULL;
    } else if (!strcmp(buf, "new")) {
	stack->data->type = FORDATA_NEW;
	stack->data->number = MAX(CurrentMessage + 1, 1) - 1;
	sprintf(v->value, "%d", MAX(CurrentMessage + 1, 1));
	if (CurrentMessage >= Lastmessage)
	    stack->type = TYPE_NULL;
    } else if (*buf == '(') {
	while (*p != '(')
	    p++;
	stack->data->room_list = (result_list *)malloc(sizeof(result_list));
	stack->data->room_list->list = (void **)malloc(sizeof(void *) * 20);
	stack->data->room_list->count = 0;
	while (*p != ')' && *p != '\0') {
	    stack->data->room_list->list[stack->data->room_list->count] =
	      (char *)malloc(25);
	    pc = stack->data->room_list->list[stack->data->room_list->count];
	    p++;
	    while (*p != ')' && *p != ' ')
		*pc++ = *p++;
	    *pc = 0;
	    stack->data->room_list->count++;
	}
	stack->data->type = FORDATA_LIST;
	if (!*p) {
	    pop();
	    fprintf(stderr, "program: syntax error: missing ')'\n");
	    return 0;
	}
	strcpy(v->value, *stack->data->room_list->list);
	if (stack->data->room_list->count == 0)
	    stack->type = TYPE_NULL;
    } else {
	pop();
	free(v->name);
	free(v->value);
	free(v);
	fprintf(stderr, "Unrecognized type of for loop\n");
	return 0;
    }
    v->next = temps;
    temps = v;
    return 1;
}

static int endcmd(p, f, fp)
    char *p;
    FILE *f;
    long fp;
{
    if (!stack)
	return 0;
    if (stack->type == TYPE_NULL) {
	pop();
	return 1;
    }
    if (stack->type != TYPE_FOR && stack->type != TYPE_WHILE) {
	fprintf(stderr,
		"program: 'end' does not have matching 'for' or 'while'\n");
	return 0;
    }
    if (stack->type == TYPE_WHILE) {
	fseek(f, stack->position, 0);
	pop();
	return 1;
    }
    stack->data->number++;

    if (stack->data->type == FORDATA_ROOMS) {
	if (stack->data->number >= stack->data->room_list->count ||
	    ((directory_info *)
	     stack->data->room_list->list[stack->data->number])
	    ->priority < Cutoff_priority) {
	    pop();
	    return 1;
	}
	strcpy(stack->data->var->value, ((directory_info *)
		  stack->data->room_list->list[stack->data->number])->name);
    } else if (stack->data->type == FORDATA_LIST) {
	if (stack->data->number >= stack->data->room_list->count) {
	    pop();
	    return 1;
	}
	strcpy(stack->data->var->value, (char *)
	       stack->data->room_list->list[stack->data->number]);
    } else {
	if (stack->data->number >= Lastmessage) {
	    pop();
	    return 1;
	}
	sprintf(stack->data->var->value, "%d", stack->data->number + 1);
    }
    fseek(f, stack->position, 0);
    return 1;
}

char ivarbuf[20];

static char *bbvariables(name, message, rinfo)
    char *name;
    int message;
    directory_info *rinfo;
{
 /*
  * The following selectors are valid for rooms 
  */
    if (!strcmp(name, "room")) {
	if (!rinfo)
	    return "";
	return rinfo->name;
    }
    if (!strcmp(name, "current")) {
	if (strcmp(rinfo->name, Roomname) != 0) {
	    fprintf(stderr, "The %s selector only works for the current room",
		    name);
	    return 0;
	}
	sprintf(ivarbuf, "%d", CurrentMessage);
	return ivarbuf;
    }
    if (!strcmp(name, "groups")) {
	if (!rinfo) {
	    fprintf(stderr, "no such room\n");
	    return 0;
	}
	return group_names(&rinfo->groups, group_directory);
    }
    if (!strcmp(name, "messages") || !strcmp(name, "last")) {
	if (strcmp(rinfo->name, Roomname) != 0) {
	    fprintf(stderr, "The %s selector only works for the current room",
		    name);
	    return 0;
	}
	sprintf(ivarbuf, "%d", Lastmessage);
	return ivarbuf;
    }
    if (!strcmp(name, "new")) {
	if (!rinfo) {
	    fprintf(stderr, "no such room\n");
	    return 0;
	}
	sprintf(ivarbuf, "%d", rinfo->new_messages);
	return ivarbuf;
    }
    if (!strcmp(name, "owner")) {
	if (!rinfo) {
	    fprintf(stderr, "no such room\n");
	    return 0;
	}
	return rinfo->owner;
    }
    if (!strcmp(name, "priority")) {
	if (!rinfo) {
	    fprintf(stderr, "no such room\n");
	    return 0;
	}
	sprintf(ivarbuf, "%d", rinfo->priority);
	return ivarbuf;
    }
    if (!strcmp(name, "summary")) {
	if (!rinfo) {
	    fprintf(stderr, "no such room\n");
	    return 0;
	}
	return rinfo->summary;
    }
 /*
  * The following selectors are valid for messages 
  */
    if (!strcmp(name, "permanent") || !strcmp(name, "perm")) {
	sprintf(ivarbuf, "#%d", SUMMARY_NUMBER(message));
	return ivarbuf;
    }
    if (!strcmp(name, "poster"))
	return SUMMARY(message)->poster;
    if (!strcmp(name, "subject"))
	return SUMMARY(message)->subject;
    if (!strcmp(name, "date"))
	return Date(SUMMARY(message)->date)+4;
    if (!strcmp(name, "idate")) {
	sprintf(ivarbuf, "%d", SUMMARY(message)->date);
	return ivarbuf;
    }
    return 0;
}

static char *setvariables(name)
    char *name;
{
    if (!strcmp(name, "abbreviate"))
	return Abbrev ? "true" : "false";
    if (!strcmp(name, "cutoff")) {
	sprintf(ivarbuf, "%d", Cutoff_priority);
	return ivarbuf;
    }
    if (!strcmp(name, "date"))
	return Date(time(0))+4;
    if (!strcmp(name, "editmode"))
	return Editmode ? "true" : "false";
    if (!strcmp(name, "editor"))
	return Editor;
    if (!strcmp(name, "environment"))
	return Environment ? "true" : "false";
    if (!strcmp(name, "file_header_format") || !strcmp(name, "fheader"))
	return File_header_format;
    if (!strcmp(name, "header_format") || !strcmp(name, "header"))
	return Header_format;
    if (!strcmp(name, "helpfile"))
	return Helpfile;
    if (!strcmp(name, "hooks"))
	return Hooks ? "true" : "false";
    if (!strcmp(name, "id"))
	return Id;
    if (!strcmp(name, "idate")) {
	sprintf(ivarbuf, "%d", time(0));
	return ivarbuf;
    }
    if (!strcmp(name, "mail_address") || !strcmp(name, "mail"))
	return Mail_address;
    if (!strcmp(name, "mailer"))
	return Mailer;
    if (!strcmp(name, "nickname") || !strcmp(name, "nick"))
	return Nickname;
    if (!strcmp(name, "no_clobber"))
	return No_clobber ? "true" : "false";
    if (!strcmp(name, "pager"))
	return Pager;
    if (!strcmp(name, "path"))
	return Path;
    if (!strcmp(name, "port")) {
	sprintf(ivarbuf, "%d", Cutoff_priority);
	return ivarbuf;
    }
    if (!strcmp(name, "server"))
	return Server;
    if (!strcmp(name, "snide"))
	return Snide ? "true" : "false";
    return 0;
}


static int expand_macro(macro, copyto, active, rinfo)
    char *macro;
    char **copyto;
    int active;
    directory_info *rinfo;
{
    int tmp;
    char *cp;
    VAR *p;

    for (p = temps; p && strcmp(macro, p->name); p = p->next);
    if (!p) {
	for (p = globals; p && strcmp(macro, p->name); p = p->next);
    }
    if (!p) {
	if (!strcmp(macro, "*")) {
	    cp = all_params;
	    while (*cp)
		*(*copyto)++ = *cp++;
	} else if (isdigit(*macro)) {
	    sscanf(macro, "%d", &tmp);
	    if (tmp > 10 || !params[tmp]) {
		fprintf(stderr, "program: no parameter number %s\n", macro);
		return 0;
	    }
	    cp = params[tmp];
	    while (*cp)
		*(*copyto)++ = *cp++;
	} else {
	    if ((int)(cp = bbvariables(macro, active, rinfo)) == -1)
		return 0;
	    if (!cp)
		cp = setvariables(macro);
	    if (cp) {
		while (*cp)
		    *((*copyto)++) = *cp++;
	    } else {
		fprintf(stderr, "program: variable ${%s} is undefined.\n",
			macro);
		return 0;
	    }
	}
    } else {
	for (cp = p->value; *cp; cp++, (*copyto)++)
	    **copyto = *cp;
    }
    return 1;
}
