#include <stdio.h>
#include <sys/types.h>
#ifdef PTX
#include <dirent.h>		/* MAXNAMLEN */
#else
#include <sys/dir.h>		/* MAXNAMLEN */
#endif PTX
#include <errno.h>
#include <pwd.h>
#include <ctype.h>
#include <unistd.h> // For getuid
#include <stdlib.h> // For malloc, realloc

#include "../bb/config.h"
#include "../lib/client.h"

#define BUFSIZE (2048)
#define BLOCK_SIZE (4096)
#define QUEUEFILE "~/.m2bb-queue"

extern char *getenv(char *name);

void read_buf(char **buf, int *buf_l);
int poster(char *room, char *subject, char *nick, char *buf, int buf_l);
int runqueue(void);
int requeue(FILE *fp, char *room, char *subject, char *from, char *buf,
	int buf_l);
int enqueue(char *room, char *subject, char *from, char *buf, int buf_l);
int qhead(char *file, char *room, char *subject, char *from,
	      char **buf, int *buf_l);
char *tilde(char *file, struct passwd * user);
int signon(char *server);
void signoff(void);
void panicsave(char *room, char *sub, char *from, char *buf, int buf_l);
void usage(char *prog);

#define writemessage(xx, xy, xz, yy, yz, zz) \
  (fprintf(xx, "room: %s\nsubject: %s\nfrom: %s\nsize: %d\n%s", \
	   xy, xz, yy, zz, yz) >= 0)

static char *Queuefile = QUEUEFILE;
static FILE *Logfile = 0;

int main(int argc, char *argv[])
{
    char subject[BUFSIZE] = "", from[BUFSIZE] = "", date[BUFSIZE] = "";
    char line[BUFSIZE] = "";
    char *buf;
    char *room = 0;
    char *log = 0;
    int i, tmp;
    int buf_l;
    struct passwd *user;
    char *server = SERVER;
    int port_number = PORT_NUMBER;

    if (!(user = getpwuid(getuid()))) {
	fprintf(stderr, "The system doesn't recognize you.\n  How odd.\n");
	exit(1);
    }
    if ((buf = getenv("BBSERVER")))
	server = buf;
    if ((buf = getenv("BBSERVERPORT"))) {
	if (sscanf(buf, "%d", &tmp) != 1) {
	    fprintf(stderr, "Server port must be an integer value, not %s\n",
		    buf);
	    return 1;
	}
	else
	    port_number = tmp;
    }
    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-')
	    switch (argv[i][1]) {
	    case 'f':
		if (argv[i][2])
		    Queuefile = &argv[i][2];
		else
		    Queuefile = argv[++i];
		break;
	    case 'l':
		if (argv[i][2])
		    log = &argv[i][2];
		else
		    log = argv[++i];
		break;
	    case 'Q':
		runqueue();
		return 0;
	    case 'S':
	    case 's':
		if (argv[i][2])
		    server = &argv[i][2];
		else
		    server = argv[++i];
		break;
	    default:
		fprintf(stderr, "Illegal flag '-%c'\n", argv[i][1]);
		usage(argv[0]);
		break;
	    }
	else {
	    if (room)
		usage(argv[0]);
	    room = argv[i];
	}
    }
    if (!room)
	usage(argv[0]);

    if ((buf = tilde(Queuefile, user)))
	Queuefile = buf;

    if (log) {
	if ((buf = tilde(log, user)))
	    log = buf;
	Logfile = fopen(log, "a");
    }
    fgets(line, sizeof(line), stdin);
    while (*line != '\n') {
	if (sscanf(line, "Date: %[^\n]", date) == 1) {
	    fgets(line, sizeof(line), stdin);
	    while (*line == '\t' || *line == ' ') {
		int lineLength;

		*line = ' ';
		lineLength = strlen(line);
		if (line[lineLength - 1] == '\n')
		    line[lineLength - 1] = '\0';
		strcat(date, line);
		fgets(line, sizeof(line), stdin);
	    }
	} else if (sscanf(line, "From: %[^\n]", from) == 1) {
	    fgets(line, sizeof(line), stdin);
	    while (*line == '\t' || *line == ' ') {
		int lineLength;

		*line = ' ';
		lineLength = strlen(line);
		if (line[lineLength - 1] == '\n')
		    line[lineLength - 1] = '\0';
		strcat(from, line);
		fgets(line, sizeof(line), stdin);
	    }
	} else if (sscanf(line, "Subject: %[^\n]", subject) == 1) {
	    fgets(line, sizeof(line), stdin);
	    while (*line == '\t' || *line == ' ') {
		int lineLength;

		*line = ' ';
		lineLength = strlen(line);
		if (line[lineLength - 1] == '\n')
		    line[lineLength - 1] = '\0';
		strcat(subject, line);
		fgets(line, sizeof(line), stdin);
	    }
	} else
	    fgets(line, sizeof(line), stdin);
    }

    read_buf(&buf, &buf_l);
    if (!signon(server)) {
	if (Logfile)
	    fprintf(Logfile, "Error signing on to bb . . .");
	if (!enqueue(room, subject, from, buf, buf_l)) {
	    if (Logfile)
		fprintf(Logfile, " can't enqueue file!\n");
	    panicsave(room, subject, from, buf, buf_l);
	    return 1;
	}
	if (Logfile)
	    fprintf(Logfile, " message saved to %s\n", Queuefile);
	return 0;
    }
    runqueue();
    if (!poster(room, subject, from, buf, buf_l)) {
	if (Logfile)
	    fprintf(Logfile, "Error posting message to %s . . .", room);
	if (!enqueue(room, subject, from, buf, buf_l)) {
	    if (Logfile)
		fprintf(Logfile, " can't enqueue file!\n");
	    panicsave(room, subject, from, buf, buf_l);
	    return 1;
	}
	if (Logfile)
	    fprintf(Logfile, " message saved to %s\n", Queuefile);
	signoff();
	return 1;
    }
    signoff();
    if (Logfile)
	fclose(Logfile);
    return 0;
}

void read_buf(char **buf, int *buf_l)
{
    int space_left, newly_read;

    space_left = BLOCK_SIZE;
    *buf = (char *)malloc(space_left);
    *buf_l = 0;
    while ((newly_read = fread(*buf + *buf_l, 1, space_left, stdin)) > 0) {
	*buf_l += newly_read;
	space_left -= newly_read;
	if (space_left == 0) {
	    space_left = BLOCK_SIZE;
	    *buf = (char *)realloc(*buf, *buf_l + space_left);
	}
    }
}

int signon(char *server)
{
    bb_priority hidden_priority;

    if (!server_open(server, PORT_NUMBER)) {
	printf("error connecting to server: %s\n", client_error);
	return 0;
    }
    if (!client_signon("daemon", "blegga;", &hidden_priority, FALSE)) {
	printf("error signing on: %s\n", client_error);
	return 0;
    }
    return 1;
}

void signoff(void)
{
    if (!client_signoff())
	fprintf(stderr, "error signing off: %s\n", client_error);
}

int poster(char *room, char *subject, char *nick, char *buf, int buf_l)
{
    int message_number, rid;
    int return_code = 1;
    char *p;

    if (!client_select(room, &message_number, &rid)) {
	printf("error selecting %s: %s\n", room, client_error);
	return 0;
    }
    if (!client_nick(nick)) {
	fprintf(stderr, "warning: error setting nickname %s: %s\n", nick,
		client_error);
    }
    for (p = nick; *p && *p != '<'; p++);
    if (*p)
	nick = p + 1;
    for (p = nick; *p && *p != '>' && *p != '%' && *p != '@'; p++)
	if (isupper(*p))
	    *p = tolower(*p);
    if (*p)
	*p = '\0';
    client_setname(nick);
    if (!client_post(subject, buf, buf_l)) {
	printf("error posting message: %s\n", client_error);
	return_code = 0;
    }
    return return_code;
}

int runqueue(void)
{
    char room[BUFSIZE] = "", subject[BUFSIZE] = "", from[BUFSIZE] = "";
    char date[BUFSIZE] = "";
    char *buf = "";
    int buf_l;
    char tmpfile[MAXNAMLEN];
    extern int errno;
    FILE *rqfp = 0;

    strncpy(tmpfile, Queuefile, MAXNAMLEN);
    strncat(tmpfile, "~", MAXNAMLEN);
    if (rename(Queuefile, tmpfile) < 0) {
	if (errno != ENOENT)
	    perror("rename");
	return 0;
    }
    while (qhead(tmpfile, room, subject, from, &buf, &buf_l)) {
	if (!poster(room, subject, from, buf, buf_l)) {
	    if (Logfile)
		fprintf(Logfile, "Error posting from queue to room %s . . . ", room);
	    if (!requeue(rqfp, room, subject, from, date, buf, buf_l)) {
		if (Logfile)
		    fprintf(Logfile, "error requeuing message!\n");
		panicsave(room, subject, from, buf, buf_l);
	    } else if (Logfile)
		fprintf(Logfile, "message requeued\n");
	}
    }
    if (rqfp)
	fclose(rqfp);
    if (unlink(tmpfile) < 0)
	perror("unlink");
    return 0;
}

int requeue(FILE *fp, char *room, char *subject, char *from, char *buf,
	int buf_l)
{
    if (!(fp || fp == fopen(Queuefile, "w")))
	return 0;

    if (!writemessage(fp, room, subject, from, buf, buf_l))
	return 0;

    return 1;
}

int enqueue(char *room, char *subject, char *from, char *buf, int buf_l)
{
    FILE *fp;

    if ((fp = fopen(Queuefile, "a")) == NULL)
	return 0;

    if (!writemessage(fp, room, subject, from, buf, buf_l)) {
	fclose(fp);
	return 0;
    }
    if (fclose(fp) == EOF)
	return 0;

    return 1;
}

int qhead(char *file, char *room, char *subject, char *from, char **buf, int *buf_l)
{
    static FILE *fp = 0;
    char buffer[BUFSIZE];

    if (!fp && (fp = fopen(file, "r")) == NULL)
	return 0;
    if (fgets(buffer, sizeof(buffer), fp) == NULL)	/* Empty file */
	return 0;
    *buf_l = -1;
    do {
	sscanf(buffer, "room: %[^\n]", room);
	sscanf(buffer, "subject: %[^\n]", subject);
	sscanf(buffer, "from: %[^\n]", from);
	if (sscanf(buffer, "size: %d\n", buf_l) > 0)
	    break;
    } while (fgets(buffer, sizeof(buffer), fp) != NULL);
    if (*buf_l < 0) {
	fprintf(stderr, "Error reading from %s, file corrupted.\n", Queuefile);
	fclose(fp);
	return 0;
    }
    *buf = (char *)malloc(*buf_l * sizeof(char));
    if (!fread(*buf, sizeof(char), *buf_l, fp)) {
	fprintf(stderr, "Error reading from %s, file corrupted.\n", Queuefile);
	fclose(fp);
	return 0;
    }
    return 1;
}

void panicsave(char *room, char *sub, char *from, char *buf, int buf_l)
{
}

void usage(char *prog)
{
    fprintf(stderr, "usage:\t%s [-f queuefile] [-l logfile] <room>\n", prog);
    fprintf(stderr, "\t%s [-f queuefile] [-l logfile] -Q\n", prog);
    exit(1);
}
