#include <ar.h>
#include "ar.h"
#include "macros.h"

#ifndef SEEK_SET
#define SEEK_SET (0)
#define SEEK_CUR (1)
#define SEEK_END (2)
#endif /* SEEK_SET */

AR *ar_open(const char *filename, const char *mode)
{
    FILE *file;
    AR *ar;
    char buf[] = ARMAG;
    char l_mode[10];

    strcpy(l_mode, mode);
    if (*l_mode == 'a')
	*l_mode = 'w';
    file = fopen(filename, l_mode);
    if (!file)
	return NULL;
    if (*l_mode == 'r') {
	bzero(buf, sizeof(buf));
	fread(buf, strlen(ARMAG), 1, file);
	if (strcmp(buf, ARMAG) != 0)
	    return NULL;
    } else {
	fwrite(ARMAG, strlen(ARMAG), 1, file);
    }
    ar = (AR *)malloc(sizeof(*ar));
    ar->file = file;
    ar->index = (struct index **) malloc(sizeof(*ar->index));
    *ar->index = NULL;
    ar->index_size = 1;
    ar->last_pos = ftell(ar->file);
    ar->indexed = 0;
    ar->buf = NULL;
    return ar;
}

void ar_close(AR *ar)
{
    struct index **p;

    if (!ar)
	return;
    fclose(ar->file);
    for (p = ar->index; *p; p++)
	free(p);
    free(ar->index);
    if (ar->buf)
	free(ar->buf);
    free(ar);
}

struct index *ar_find(AR *ar, const char *name)
{
    struct ar_hdr hdr;
    int found = 0;
    struct index **p;

    for (p = ar->index; *p; p++) {
	found = strcmp(name, (*p)->name) == 0;
    }
    if (!found && !ar->indexed)
	fseek(ar->file, ar->last_pos, SEEK_SET);
    for (; !found && !ar->indexed; p++) {
	int size, count, i;

	fread((char *)&hdr, sizeof(hdr), 1, ar->file);
	if (sscanf(hdr.ar_size, "%d", &size) != 1)
	    return NULL;
	count = p - ar->index;
	if (count >= ar->index_size - 1) {
	    struct index **q;

	    ar->index_size += 20;
	    ar->index = (struct index **)
	      Realloc(ar->index, ar->index_size * sizeof(*ar->index));
	    p = ar->index + count;
	    for (q = p; q < ar->index + ar->index_size; q++)
		*q = NULL;
	}
	*p = (struct index *) malloc(sizeof(**p));
	bcopy(hdr.ar_name, (*p)->name, sizeof(hdr.ar_name));
	for (i = strlen((*p)->name) - 1;
	     i >= 0 && (*p)->name[i] == ' ';
	     i--)
	    (*p)->name[i] = '\0';
	(*p)->size = size;
	(*p)->pos = ftell(ar->file);
	fseek(ar->file, (*p)->size, SEEK_CUR);
	ar->last_pos = ftell(ar->file);
	ar->indexed = feof(ar->file);
	found = strcmp(name, (*p)->name) == 0;
    }
    return found ? *(--p) : NULL;
}

char *ar_get(AR *ar, const char *name)
{
    struct index *p;

    p = ar_find(ar, name);
    if(!p) return NULL;
    fseek(ar->file, p->pos, SEEK_SET);
    ar->buf = (char *)Realloc(ar->buf, p->size + 1);
    fread(ar->buf, p->size, 1, ar->file);
    ar->buf[p->size] = '\0';
    return ar->buf;
}

int ar_put(AR *ar, const char *name, const int uid, const int date,
	       const char *buf, const int length)
{
       ;

    fseek(ar->file, 0, SEEK_END);
    fprintf(ar->file, "%-16s", name);
    fprintf(ar->file, "%-12d", date);
    fprintf(ar->file, "%-6d", uid);
    fprintf(ar->file, "%-6d", 0);
    fprintf(ar->file, "%-8o", 0100660);
    fprintf(ar->file, "%-10d", length);
    fprintf(ar->file, "%-2s", ARFMAG);
    fwrite(buf, length, 1, ar->file);
    if (length % 2 != 0)
	fprintf(ar->file, "\n");
    ar->indexed = 0;
    return 1;
}

#define BLOCK_SIZE (8192)

int ar_delete(AR *ar, const char *name)
{
    char *buf;
    int buf_l;
    int space_left, newly_read;
    struct index *p;

    p = ar_find(ar, name);
    if(!p) return 0;
    fseek(ar->file, p->pos + p->size, SEEK_SET);
    space_left = BLOCK_SIZE;
    buf = (char *)malloc(space_left);
    if (!buf) {
	perror("malloc");
	return 0;
    }
    buf_l = 0;
    bzero(buf, space_left);
    while (!feof(ar->file) &&
	   (newly_read = fread(buf + buf_l, 1, space_left, ar->file)) > 0) {
	buf_l += newly_read;
	space_left -= newly_read;
	if (space_left == 0) {
	    space_left = BLOCK_SIZE;
	    buf = (char *)realloc(buf, buf_l + space_left);
	    if (!buf) {
		perror("realloc");
		return 0;
	    }
	    bzero(buf + buf_l, space_left);
	}
    }
    fseek(ar->file, p->pos - sizeof(struct ar_hdr), SEEK_SET);
    fwrite(buf, buf_l, 1, ar->file);
    ftruncate(ar->file, ftell(ar->file));
    return 1;
}
