#include "../rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/lib/client/client.c,v 1.7 1999/08/07 00:25:54 kc Exp $")

#include "../header.h"
#ifndef VMS
#include <sys/time.h>
#endif
#include <netinet/in.h>
#include <unistd.h> // For close(), select(), read()

int server_socket = -1, server_major = -1, server_minor = -1;
char client_error[BUFFER_SIZE];

int read_server_code(int *code)
{
    int success;
    char c;

    success = read_socket_byte(server_socket, &c) != -1;
    if (!success) {
	strcpy(client_error, "connection failed");
	server_close();
    } else {
	switch (*code = (int)c) {
	case SERVER_ERR:
	    read_socket(server_socket, client_error, sizeof(client_error));
	    break;
	default:
	    break;
	}
    }
    return success;
}

int write_server_code(int code)
{
    char c = (char)code;

    return write_socket_byte(server_socket, c);
}

int server_open(const char *server_hostname, int server_port)
{
    struct sockaddr_in server;
    struct hostent *host;
    int code;
    int port_number;
    int connected = 0;

    for (port_number = server_port;
	 port_number < server_port + 3 && !connected;
	 port_number++) {
	if (server_socket >= 0)
	    close(server_socket);
	server_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (server_socket == -1) {
	    sprintf(client_error, "Unable to create a socket: %s",
		    strerror(errno));
	    return 0;
	}
#ifdef SO_DONTLINGER
	setsockopt(server_socket, SOL_SOCKET, SO_DONTLINGER, (char *)0, 0);
#endif SO_DONTLINGER
	host = gethostbyname((char *)server_hostname);
	if (!host) {
#ifdef HOST_NOT_FOUND
	    extern int h_errno;

	    if (h_errno == HOST_NOT_FOUND)
		sprintf(client_error, "No such host %s",
			server_hostname);
	    else
#endif HOST_NOT_FOUND
		sprintf(client_error, "Error looking up host %s",
			server_hostname);
	    close(server_socket);
	    return 0;
	}
	server.sin_family = host->h_addrtype;
	bcopy(host->h_addr, &server.sin_addr, host->h_length);
	server.sin_port = htons(port_number);
	if (!(connected =
	      connect(server_socket, (struct sockaddr *) & server,
		      sizeof(server)) == 0)) {
	    switch (errno) {
	    case ECONNREFUSED:
		sprintf(client_error, "the bb server on %s is down",
			server_hostname);
		break;
	    default:
		sprintf(client_error, "connecting to server: %s",
			strerror(errno));
	    }
	}
    }
    if (!connected) {
	close(server_socket);
	server_socket = -1;
	return 0;
    }
    read_server_code(&code);
    if (code != SERVER_ACK) {
        if (code == SERVER_FORWARD) {
	    char new_server[64];
	    int new_server_port;
	    static forwarded = 0;

	    if(++forwarded > 3) {
		strcpy(client_error, "server_open forwarded too many times");
		return 0;
	    }
	    read_socket(server_socket, new_server, sizeof(new_server));
	    read_socket_int(server_socket, &new_server_port);
	    return server_open(new_server, new_server_port);
	}
	if (code != SERVER_ERR)
	    strcpy(client_error, "server handshake failed");
	close(server_socket);
	server_socket = -1;
	return 0;
    }
    return client_protocol();
}

void server_close()
{
    if   (server_socket != -1)
	     close(server_socket);
    server_socket = -1;
}

void server_flush()
{
#ifndef VMS
    struct timeval timeout = {0 /* sec */ , 0 /* usec */ };
    fd_set ready;
    char flush[BUFFER_SIZE];

    for (;;) {
	FD_ZERO(&ready);
	FD_SET(server_socket, &ready);
	if (select(server_socket + 1, &ready, 0, 0, &timeout) < 0)
	    return;
	if (!FD_ISSET(server_socket, &ready))
	    return;
	read(server_socket, flush, sizeof(flush));
    }
#endif VMS
}

int client_connected()
{
    if  (server_socket < 0)
	    return 0;
    return test_socket(server_socket);
}

int client_test()
{
    int code;

    write_server_code(CLIENT_TEST);
    read_server_code(&code);
    return code == SERVER_ACK;
}

int valid(const char *string)
{
    const char *s;

    for (s = string; *s; s++)
	if (!isprint(*s) && *s != '\t')
	    return 0;
    return 1;
}

int valid_id(const char *string)
{
    const char *s;

    if (!*string)
	return 0;
    for (s = string; *s; s++)
	if ((!isprint(*s)) || (*s == ' ') || (*s == '/'))
	    return 0;
    return 1;
}

int valid_nick(char *string)
{
    char *s;
    int length = 0;

    for (s = string; *s && length <= 70; s++) {
	if (!isprint(*s) && *s != '\t')
	    return 0;
	if (*s == '\t')
	    length += 8;
	else
	    length++;
    }
    if (length > 70)
	*s = '\0';
    return 1;
}
