#include "../rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/lib/client/directory.c,v 1.6 1999/08/07 00:25:57 kc Exp $")

#include <stdlib.h> // For malloc, free, realloc, qsort
#include "../header.h"

#define EXTEND_COUNT 20

static int directory_order(const void *, const void *);

static directory_info **room_list = NULL;
static directory_info **room_index = NULL;
static int room_list_size = 0, room_index_size = 0;

result_list *client_directory()
{
    static result_list result;
    directory_info *entry;
    int code, i, rid;

    block();
    write_server_code(CLIENT_DIRECTORY);
    if (!read_server_code(&code)) {
	unblock();
	return NULL;
    }
    if (code == SERVER_NAK) {
	unblock();
	return &result;
    }
    i = 0;
    while (code != SERVER_EOF) {
	switch (code) {
	case SERVER_ACK:
	    if (server_minor >= 3) {
		read_socket_int(server_socket, &rid);
		while (rid >= room_index_size) {
		    room_index_size += EXTEND_COUNT;
		    room_index = Realloc(room_index, directory_info *,
					 room_index_size);
		    bzero(room_index + room_index_size - EXTEND_COUNT,
			  EXTEND_COUNT * sizeof(*room_index));
		}
		if (!room_index[rid])
		    room_index[rid] = (directory_info *)
		      malloc(sizeof(directory_info));
		entry = room_index[rid];
	    } else {
		if (i >= room_list_size) {
		    room_list_size += EXTEND_COUNT;
		    room_list = Realloc(room_list, directory_info *,
					room_list_size);
		    bzero(room_list + room_list_size - EXTEND_COUNT,
			  EXTEND_COUNT * sizeof(*room_list));
		}
		if (!room_list[i])
		    room_list[i] = (directory_info *)
		      malloc(sizeof(directory_info));
		entry = room_list[i++];
	    }
	    read_socket(server_socket, entry->name, sizeof(entry->name));
	    read_socket(server_socket, entry->summary,
			sizeof(entry->summary));
	    read_socket(server_socket, entry->owner, sizeof(entry->owner));
	    read_socket_int(server_socket, (int *)&entry->priority);
	    read_socket_int(server_socket, (int *)&(entry->messagenumber));
	    read_socket_int(server_socket, (int *)&entry->new_messages);
	    read_socket_int(server_socket, (int *)&entry->class);
	    read_socket_int(server_socket, (int *)&entry->default_expiration);
	    read_socket_int(server_socket, (int *)&entry->default_class);
	    read_socket_group_set(server_socket, &entry->groups);
	    read_socket_int(server_socket, (int *)&entry->archive);
	    break;
	case SERVER_DELETE_ROOM:
	    read_socket_int(server_socket, &rid);
	    if (room_index[rid]) {
		free(room_index[rid]);
		room_index[rid] = NULL;
	    }
	    break;
	default:
	    unblock();
	    return NULL;
	    break;
	}
	if (!read_server_code(&code)) {
	    unblock();
	    return NULL;
	}
    }
    unblock();
    if (server_minor >= 3)
	for (rid = 0; rid < room_index_size; rid++) {
	    if (i >= room_list_size) {
		room_list_size += EXTEND_COUNT;
		room_list = Realloc(room_list, directory_info *,
				    room_list_size);
		bzero(room_list + room_list_size - EXTEND_COUNT,
		      EXTEND_COUNT * sizeof(*room_list));
	    }
	    if (room_index[rid])
		room_list[i++] = room_index[rid];
	}
    qsort(room_list, i, sizeof(directory_info *), directory_order);
    result.list = (void **)room_list;
    result.count = i;
    return &result;
}

static int directory_order(const void *p1, const void *p2)
{
    int order;
    const directory_info * const *di1 = p1;
    const directory_info * const *di2 = p2;

    order = (*di2)->priority - (*di1)->priority;
    if (!order)
	order = istrcmp((*di1)->name, (*di2)->name);
    return order;
}

void client_directory_flush()
{
    int rid;

    for(rid=0; rid < room_index_size; rid++) {
	if(room_index[rid]) {
	    free(room_index[rid]);
	    room_index[rid] = NULL;
	}
    }
}
