#include "../rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/lib/client/message_info.c,v 1.5 2001/01/26 22:04:00 kc Exp $")

#include <stdlib.h> // For malloc, realloc, free
#include "header.h"

#define EXTEND_COUNT 100

struct message_cache_entry {
    int size;
    header_info **list;
};

static struct message_cache_entry *message_cache = NULL;
static int message_cache_size = 0;

header_info *get_message_entry(int rid, int message)
{
    if (!load_message_cache(rid, message, message))
	return NULL;
    return message_cache[rid].list[message];
}

boolean load_message_cache(int rid, int start_message, int end_message)
{
    struct message_cache_entry *entry;
    int start_run, end_run;

    if (rid >= message_cache_size)
	return read_message_cache(rid, start_message, end_message);
    entry = &message_cache[rid];
    if (start_message >= entry->size)
	return read_message_cache(rid, start_message, end_message);
    if (end_message >= entry->size) {
	if (!read_message_cache(rid, entry->size, end_message))
	    return FALSE;
	end_message = entry->size;
    }
    start_run = start_message;
    while (start_run <= end_message) {
	while (start_run <= end_message && entry->list[start_run])
	    start_run++;
	if (start_run <= end_message) {
	    for (end_run = start_run + 1;
		 end_run <= end_message && !entry->list[end_run];
		 end_run++);
	    if (entry->list[end_run])
		end_run--;
	    if (!read_message_cache(rid, start_run, end_run))
		return FALSE;
	    start_run = end_run + 1;
	}
    }
    return TRUE;
}

boolean read_message_cache(int rid, int start_message, int end_message)
{
    int code;
    int messagenumber;
    header_info *entry;

    while (rid >= message_cache_size) {
	message_cache_size += EXTEND_COUNT;
	message_cache = Realloc(message_cache, struct message_cache_entry,
				message_cache_size);
	bzero(message_cache + message_cache_size - EXTEND_COUNT,
	      EXTEND_COUNT * sizeof(struct message_cache_entry));
    }
    block();
    write_server_code(CLIENT_MESSAGE_INFO);
    write_socket_int(server_socket, rid);
    if (start_message >= end_message) {
	write_socket_int(server_socket, SERVER_MESSAGE_SINGLE);
	write_socket_int(server_socket, start_message);
    } else {
	write_socket_int(server_socket, SERVER_MESSAGE_RANGE);
	write_socket_int(server_socket, start_message);
	write_socket_int(server_socket, end_message);
    }
    if (!read_server_code(&code)) {
	unblock();
	return 0;
    }
    if (code == SERVER_NAK) {
	unblock();
	return 1;
    }
    while (code == SERVER_ACK) {
	read_socket_int(server_socket, (int *)&messagenumber);
	while (messagenumber >= message_cache[rid].size) {
	    message_cache[rid].size += EXTEND_COUNT;
	    message_cache[rid].list = Realloc(message_cache[rid].list,
					      header_info *,
					      message_cache[rid].size);
	    bzero(message_cache[rid].list +
		  message_cache[rid].size - EXTEND_COUNT,
		  EXTEND_COUNT * sizeof(header_info *));
	}
	entry = message_cache[rid].list[messagenumber];
	if (!entry)
	    entry = (header_info *) malloc(sizeof(header_info));
	entry->number = messagenumber;
	read_socket_int(server_socket, (int *)&entry->date);
	read_socket_int(server_socket, (int *)&entry->lines);
	read_socket(server_socket, entry->poster, sizeof(entry->poster));
	read_socket(server_socket, entry->id, sizeof(entry->id));
	read_socket(server_socket, entry->nick, sizeof(entry->nick));
	read_socket(server_socket, entry->subject, sizeof(entry->subject));
	read_socket_int(server_socket, (int *)&entry->expiration);
	message_cache[rid].list[messagenumber] = entry;
	if (!read_server_code(&code)) {
	    unblock();
	    return 0;
	}
    }
    unblock();
    return code == SERVER_EOF && message_cache[rid].size > start_message;
}
