#include "rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/lib/transport.c,v 1.8 1999/08/07 00:25:53 kc Exp $")

#ifndef VMS
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#else
#include <types.h>
#include <time.h>
#include <socket.h>
#include <unistd.h> // For select, close, read, write
#include <stdlib.h> // For malloc, free
#endif VMS
#ifdef PTX
#include <sys/select.h>
#endif PTX

#include <stdio.h>

#define DEBUG_TRANSPORT (1 << 0)

#ifndef READ_TIMEOUT
#define READ_TIMEOUT (15)
#endif
#ifndef WRITE_TIMEOUT
#define WRITE_TIMEOUT (60)
#endif

int Debug = 0;
FILE *Debug_file = stderr;
int Read_timeout = READ_TIMEOUT;
int Write_timeout = WRITE_TIMEOUT;

#include "transport.h"

#ifndef VMS
#define socket_read(socket, buf, length) read(socket, buf, length)
#define socket_write(socket, buf, length) write(socket, buf, length)
#define socket_close(socket) close(socket)
#endif VMS

int test_socket(int socket)
{
#ifndef VMS
    static struct timeval timeout = {0 /* sec */ , 0 /* usec */ };
    fd_set ready;

    FD_ZERO(&ready);
    FD_SET(socket, &ready);
    if (select(socket + 1, &ready, 0, 0, &timeout) == -1)
	return 0;
    if (FD_ISSET(socket, &ready)) {
	close(socket);
	return 0;
    }
#endif VMS
    return 1;
}

static int read_bytes(int socket, char *bytes, int length)
{
    int l, status;

    l = length;
    while (l > 0) {
	status = socket_read(socket, bytes, l);
	if (status == -1 && errno == EWOULDBLOCK) {
	    fd_set ready;
	    struct timeval timeout = {Read_timeout, 0};

	    FD_ZERO(&ready);
	    FD_SET(socket, &ready);
	    if (select(socket + 1, &ready, 0, 0, &timeout) == -1) {
		socket_close(socket);
		return -1;
	    }
	    status = socket_read(socket, bytes, l);
	}
	if (status < 1) {
#ifdef DEBUG
	    if (Debug)
	  	fprintf(Debug_file, "read_bytes: error: %s\n",
			strerror(errno));
#endif DEBUG
	    socket_close(socket);
	    return -1;
	}
	l -= status;
	bytes += status;
#ifdef DEBUG
	if (Debug & DEBUG_TRANSPORT && l > 0)
	    fprintf(Debug_file, "read_bytes: %d / %d\n", length - l, length);
#endif DEBUG
    }
    return length;
}

int read_socket_byte(int socket, char *c)
{
    int status = socket_read(socket, c, 1);

    if (status == -1 && errno == EWOULDBLOCK) {
	fd_set ready;
	struct timeval timeout = {Read_timeout, 0};

	FD_ZERO(&ready);
	FD_SET(socket, &ready);
	if (select(socket + 1, &ready, 0, 0, &timeout) == -1) {
#ifdef DEBUG
	    if (Debug)
	  	fprintf(Debug_file, "read_socket_byte: error: %s\n",
			strerror(errno));
#endif DEBUG
	    socket_close(socket);
	    return -1;
	}
	status = socket_read(socket, c, 1);
    }
    if (status < 1) {
	socket_close(socket);
	return -1;
    }
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_byte %d\n", (int)*c);
#endif DEBUG
    return status;
}

static int write_bytes(int socket, const char *bytes, int length)
{
    int l, status;

    l = length;
    while (l > 0) {
#ifdef WRITE_PACKET_SIZE
	status = socket_write(socket, bytes, MIN(l, WRITE_PACKET_SIZE));
#else
	status = socket_write(socket, bytes, l);
#endif WRITE_PACKET_SIZE
        if (status == -1 && errno == EAGAIN) {
            fd_set ready;
            struct timeval timeout = {Write_timeout, 0};

            FD_ZERO(&ready);
            FD_SET(socket, &ready);
            if (select(socket + 1, 0, &ready, 0, &timeout) == -1) {
#ifdef DEBUG
                if (Debug)
                  fprintf(Debug_file, "write_bytes: error: %s\n",
                          strerror(errno));
#endif DEBUG
                socket_close(socket);
                return -1;
            }
#ifdef WRITE_PACKET_SIZE
	    status = socket_write(socket, bytes, MIN(l, WRITE_PACKET_SIZE));
#else
	    status = socket_write(socket, bytes, l);
#endif WRITE_PACKET_SIZE
        }
	if (status < 1) {
	    socket_close(socket);
	    return -1;
	}
	l -= status;
	bytes += status;
    }
    return length;
}

int write_socket_byte(int socket, char c)
{
    int l;

#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "write_byte %d\n", (int)c);
#endif DEBUG
    l = socket_write(socket, &c, 1);
    if (l == -1 && errno == EAGAIN) {
	fd_set ready;
	struct timeval timeout = {Write_timeout, 0};

	FD_ZERO(&ready);
	FD_SET(socket, &ready);
	if (select(socket + 1, 0, &ready, 0, &timeout) == -1) {
#ifdef DEBUG
	    if (Debug)
		fprintf(Debug_file, "write_socket_byte: error: %s\n",
			strerror(errno));
#endif DEBUG
	    socket_close(socket);
	    return -1;
	}
	l = socket_write(socket, &c, 1);
    }
    if (l < 1) {
	socket_close(socket);
	return -1;
    }
    return 1;
}

int read_socket_int(int socket, int *number)
{
    if (read_bytes(socket, (char *)number, sizeof(long)) != sizeof(long))
	return -1;
    *number = ntohl(*number);
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_int %d\n", *number);
#endif DEBUG
    return sizeof(long);
}

int write_socket_int(int socket, int number)
{
    long l;
    long n;

#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "write_socket_int %d\n", number);
#endif DEBUG
    n = htonl(number);
    l = write_bytes(socket, (char *)&n, sizeof(n));
    if (l != sizeof(n))
	return -1;
    return l;
}

int read_socket(int socket, char *msg, int length)
{
    long l, status;
    char *junk;
    int discard;

    bzero(msg, length);
    if (read_bytes(socket, (char *)&l, sizeof(l)) != sizeof(l))
	return -1;
    l = ntohl(l);
    if (l > length) {
	discard = l - length;
	l = length;
    } else
	discard = 0;
    length = l;
    if ((status = read_bytes(socket, msg, length)) != length)
	return status;
    if (discard) {
	junk = (char *)malloc(discard);
	status = read_bytes(socket, junk, discard);
	free(junk);
	if (status < discard)
	    return -1;
    }
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket %s\n", msg);
#endif DEBUG
    return length;
}

int read_socket_s(int socket, char *msg, int length)
{
    long l, status;
    char *junk;
    int discard;

    bzero(msg, length--);
    if (read_bytes(socket, (char *)&l, sizeof(l)) != sizeof(l))
	return -1;
    l = ntohl(l);
    if (l > length) {
	discard = l - length;
	l = length;
    } else
	discard = 0;
    if ((status = read_bytes(socket, msg, l)) != l)
	return status;
    if (discard) {
	junk = (char *)malloc(discard);
	status = read_bytes(socket, junk, discard);
	free(junk);
	if (status < discard)
	    return status;
    }
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_s(%d, '%s') = %d\n",
		length + 1, msg, (int)l);
#endif DEBUG
    return l;
}

int read_socket_buf(int socket, char **msg)
{
    long l, status;
    char *buf;

    if (read_bytes(socket, (char *)&l, sizeof(l)) != sizeof(l))
	return -1;
    l = ntohl(l);
    buf = *msg = (char *)malloc(l + 1);
    bzero(buf, l + 1);
    if ((status = read_bytes(socket, buf, l)) != l) {
	free(*msg);
	return -1;
    }
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_buf %s\n", *msg);
#endif DEBUG
    return l;
}

int write_socket(int socket, const char *msg, int length)
{
    long l;

#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "write_socket %s\n", msg);
#endif DEBUG
    l = htonl(length);
    if (write_bytes(socket, (const char *)&l, sizeof(l)) != sizeof(l))
	return -1;
    if (length == 0)
	return 0;
    return write_bytes(socket, msg, length);
}

int write_socket_string(int socket, const char *msg)
{
    return write_socket(socket, msg, strlen(msg));
}

int read_socket_group_set(int socket, group_set * groups)
{
    int i, l = 0, status;

#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_group_set: begin\n");
#endif DEBUG
    for (i = 0; i < howmany(MAX_GROUPS, NGROUPBITS); i++) {
	status = read_socket_int(socket, (int *)&groups->group_bits[i]);
	if (status == -1) {
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_group_set: error: %s\n",
		strerror(errno));
#endif DEBUG
	    return -1;
	}
	l += status;
    }
#ifdef DEBUG
    if (Debug & DEBUG_TRANSPORT)
	fprintf(Debug_file, "read_socket_group_set: 0x%04x\n",
		(int)*groups->group_bits);
#endif DEBUG
    return l;
}

int write_socket_group_set(int socket, const group_set *groups)
{
    int i, l = 0, status;

    for (i = 0; i < howmany(MAX_GROUPS, NGROUPBITS); i++) {
	if ((status = write_socket_int(socket, groups->group_bits[i])) == -1)
	    return -1;
	l += status;
    }
    return l;
}
