#ifndef _BB_TYPES_H
#define _BB_TYPES_H

#include <sys/types.h>
#include "config.h"

#include <errno.h>
extern int errno;

#ifndef VMS
#if defined(NeXT) || defined(__APPLE__)
#include <string.h>
#else
extern char *sys_errlist[];

#define strerror(errno) (sys_errlist[(errno)])
#endif
#endif				/* VMS */

#ifndef FALSE
#define FALSE (0)
#endif
#ifndef TRUE
#define TRUE (1)
#endif

#define ROOM_SIZE 20
#define LINE_SIZE 80
#define ID_SIZE 20
#define GROUP_SIZE 20

typedef long bb_priority;
typedef enum {
    no_access, read_only, read_write, moderator, no_class
} class_t;

typedef enum {
    common_archive, mail_archive, no_archive
} bb_archive_t;
typedef char boolean;

/* groups -- adapted from FD_SET definitions */
#ifndef NBBY
#define NBBY 8			/* number of bits in a byte (char) */
#endif

typedef long group_mask;

#define NGROUPBITS (sizeof(group_mask) * NBBY)	/* bits per mask */
#ifndef howmany
#define howmany(x, y) (((x)+((y)-1))/(y))
#endif

typedef struct group_set {
    group_mask group_bits[howmany(MAX_GROUPS, NGROUPBITS)];
} group_set;

#define	GROUP_SET(n, p)	((p)->group_bits[(n)/NGROUPBITS] |= (1 << ((n) % NGROUPBITS)))
#define	GROUP_CLR(n, p)	((p)->group_bits[(n)/NGROUPBITS] &= ~(1 << ((n) % NGROUPBITS)))
#define	GROUP_ISSET(n, p)	((p)->group_bits[(n)/NGROUPBITS] & (1 << ((n) % NGROUPBITS)))
#define GROUP_ZERO(p)	bzero((char *)(p), sizeof(*(p)))

typedef struct {
    char name[ROOM_SIZE];
    char summary[LINE_SIZE];
    char owner[ID_SIZE];
    bb_priority priority;
    long messagenumber;
    long new_messages;
    class_t class;
    time_t default_expiration;
    class_t default_class;
    group_set groups;
    bb_archive_t archive;
} directory_info;

typedef struct {
    char id[ID_SIZE];
    bb_priority priority;
    long last_message_read;
    class_t class;
} users_info;

typedef struct {
    char id[ID_SIZE];
    char nick[LINE_SIZE];
    char last_access_source[LINE_SIZE];
    time_t last_access_time;
    char mail_address[LINE_SIZE];
    int uses;
    int posts;
    time_t connect_time;
} user_info;

typedef struct {
    long number;
    time_t date;
    long lines;
    char poster[ID_SIZE];
    char id[ID_SIZE];
    char nick[LINE_SIZE];
    char subject[LINE_SIZE];
    time_t expiration;
} header_info;

typedef struct {
    char address[LINE_SIZE];
    char nick[LINE_SIZE];
    char id[ID_SIZE];
    char room[ROOM_SIZE];
    time_t idle;
    int slot;
    int major_version;
    int minor_version;
} user_list_info;

typedef struct {
    char name[GROUP_SIZE];
    char owner[ID_SIZE];
    class_t class;
    class_t default_class;
    bb_priority priority;
    time_t max_expiration;
    int gid;
} group_info;

#endif				/* _BB_TYPES_H */
