#include "../server/server.h"
#if defined(XA370) || defined(PTX)
#include <dirent.h>
#else
#include <sys/dir.h>
#define dirent direct
#endif XA370 || PTX

static void format_message(const char *roomname, bb_rid room_id, int messagenumber);
static int ascending_order(const void *, const void *);

#define EXTEND_COUNT 100

int main(int argc, char *argv[])
{
    bb_rid room_id;
    const char *roomname;
    int messagenumber;
    int *posts = NULL;
    int posts_size = 0;

    argc--;
    argv++;
    if (chdir(MESSAGEBASE) == -1) {
	perror("accessing messagebase");
	return 1;
    }
    database_init(TRUE, FALSE);
    switch (argc) {
    default:
	roomname = argv[0];
	break;
    case 0:
	fprintf(stderr, "format: archive <room> [<messagenumber>]\n");
	return 1;
    }
    room_id = get_rid(roomname);
    if (room_id == -1) {
	if (sscanf(roomname, "%d", &room_id) != 1) {
	    fprintf(stderr, "No such room '%s'\n", roomname);
	    return -1;
	}
    }
    roomname = get_roomname(room_id);
    if (!*roomname) {
	FILE *room_list;

	room_list = fopen(EXPIRED_ROOMS, "r");
	if (room_list) {
	    int rid;
	    static char name[BUFFER_SIZE];

	    while (fscanf(room_list, "%d %[^\n]\n", &rid, name) == 2) {
		if (room_id == rid) {
		    roomname = name;
		    break;
		}
	    }
	    fclose(room_list);
	}
    }
    if (!*roomname)
	roomname = argv[0];

    if (argc > 1) {
	int i;

	for (i = 1; i < argc; i++) {
	    sscanf(argv[i], "%d", &messagenumber);
	    format_message(roomname, room_id, messagenumber);
	}
    } else {
	char fn[BUFFER_SIZE];
	DIR *dirp;
	struct dirent *dp;
	int i, count = 0;;

	sprintf(fn, EXPIRED_ROOM, room_id);
	if (dirp = opendir(fn)) {
	    for (dp = readdir(dirp); dp; dp = readdir(dirp)) {
		if (strncmp(dp->d_name, "post", 4) == 0) {
		    if (count == posts_size) {
			posts_size += EXTEND_COUNT;
			posts = (int *)Realloc(posts, int, posts_size);
			bzero(posts + posts_size - EXTEND_COUNT,
			      EXTEND_COUNT * sizeof(*posts));
		    }
		    sscanf(dp->d_name, "post%d", &posts[count++]);
		}
	    }
	    closedir(dirp);
	    qsort(posts, count, sizeof(int), ascending_order);
	    for (i = 0; i < count; i++)
		format_message(roomname, room_id, posts[i]);
	} else {
	    fprintf(stderr, "There are no archives for %s\n", roomname);
	}
    }
    return 0;
}

static void format_message(const char *roomname, bb_rid room_id, int messagenumber)
{
    char message_file[BUFFER_SIZE];
    FILE *file;
    char buf[BUFFER_SIZE];

    sprintf(message_file, EXPIRED_MESSAGE, room_id, messagenumber);
    if (file = fopen(message_file, "r")) {
	fprintf(stdout, "--------------- %s #%d ---------------\n\n",
		roomname, messagenumber);
	while (fgets(buf, sizeof(buf), file))
	    fputs(buf, stdout);
	fclose(file);
	fputs("\n", stdout);
    }
}

static int ascending_order(const void *p1, const void *p2)
{
    const int *i1 = p1, *i2 = p2;
    int order;

    order = *i1 - *i2;
    return order;
}

