#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/misc/loaddbm.c,v 1.2 2000/01/30 06:13:35 kc Exp $")

#include <ndbm.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>

datum readdatum(void);
void readdata(void *ptr, size_t size);
void usage(const char *name);

int main(int argc, char *argv[])
{
    DBM *db;

    if (argc != 2) {
        usage(argv[0]);
    }
    db = dbm_open(argv[1], O_RDWR | O_CREAT, 0666);
    if (!db) {
	perror("dbm_open");
	exit(1);
    }
    while (!feof(stdin)) {
	datum key;
	datum content;

	key = readdatum();
	if (key.dsize == 0) {
	    break; // End of file
	}
	content = readdatum();
	switch (dbm_store(db, key, content, DBM_INSERT)) {
	    datum oldContent;

	    case 0:
	        break;
	    case 1:
		oldContent = dbm_fetch(db, key);
		if (oldContent.dptr == NULL) {
		    perror("dbm_fetch");
		    exit(1);
		}
		if (oldContent.dsize == content.dsize && memcmp(oldContent.dptr, content.dptr, content.dsize) == 0) {
		    break; // old and new values are identical, we don't care that it didn't want to store the same value again
		}
		fprintf(stderr, "dbm_store failed: key already exists: '");
		fwrite(key.dptr, key.dsize, 1, stderr);
		fprintf(stderr, "'\n");
		fprintf(stderr, "\told value: '");
		fwrite(oldContent.dptr, oldContent.dsize, 1, stderr);
		fprintf(stderr, "'\n");
		fprintf(stderr, "\tnew value: '");
		fwrite(content.dptr, content.dsize, 1, stderr);
		fprintf(stderr, "'\n");
		break;
	    default:
	        perror("dbm_store");
	        exit(1);
	}
	free(key.dptr);
	free(content.dptr);
    }
    dbm_close(db);
    return 0;
}

datum readdatum(void)
{
    datum someDatum;

    readdata(&someDatum.dsize, sizeof(someDatum.dsize));
    if (someDatum.dsize == 0)
        return someDatum;
    someDatum.dptr = malloc(someDatum.dsize);
    readdata(someDatum.dptr, someDatum.dsize);
    return someDatum;
}

void readdata(void *ptr, size_t size)
{
    if (fread(ptr, size, 1, stdin) != 1) {
	if (feof(stdin)) {
	    bzero(ptr, size);
	} else {
	    perror("fread");
	    exit(1);
	}
    }
}

void usage(const char *name)
{
    fprintf(stderr, "usage: %s dbname\n", name);
    exit(1);
}
