#include "../server/server.h"

extern int optind;

int verify = 0;
char *id;

static void removeuser(bb_uid uid);

main(int argc, char *argv[])
{
    int ch;
    int match = 0;
    int numeric_id = 0;
    bb_uid uid = -1, last_uid;

    chdir(MESSAGEBASE);
    if(!database_init(FALSE, TRUE)) {
	perror("database init failed");
	return 1;
    }
    while ((ch = getopt(argc, argv, "mnq")) != EOF) {
	switch (ch) {
	case 'm':
	    match = 1;
	    verify = 1;
	    break;
	case 'n':
	    numeric_id = 1;
	    break;
	case 'q':
	    verify = 1;
	    break;
	}
    }
    if (optind >= argc) {
	printf("usage: %s [-mnq] id-list\n", argv[0]);
	return 1;
    }
    if (match && numeric_id) {
	fprintf(stderr, "-m and -n are incompatible options\n");
	return 1;
    }
    for (id = argv[optind]; optind < argc; id = argv[++optind]){
	if (match) {
	    char *p;
	    char *flag = "< last uid >";

	    if (p = (char *)re_comp(id))
		fprintf(stderr, "%s: %s", id, p);
	    last_uid = get_uid(flag);
	    for (uid = 0; uid <= last_uid; uid++)
		if (re_exec(get_id(uid)))
		    removeuser(uid);
	} else {
	    if (!numeric_id) {
		if ((uid = get_uid(id)) != -1)
		    removeuser(uid);
		else
		    printf("%s: no such user\n", id);
	    } else {
		if (sscanf(id, "%d", &uid)== 1) {
		    removeuser(uid);
		} else {
		    printf("'%s' is not an integer\n", id);
		}
	    }
	}
    }
    database_shutdown();
    return 0;
}

static void removeuser(bb_uid uid)
{
    user_entry userinfo;

    if (!get_user_entry(&userinfo, uid)) {
	printf("error accessing user entry for %d\n", uid);
    } else {
	bb_rid rid;
	int room_count = 0;
	room_user_entry room_userinfo;

	for (rid = 0; rid <= Last_rid; rid++)
	    if (get_room_user_entry(&room_userinfo, rid, uid))
		room_count++;
	printf("user <%s> (%d) (%s)\n\tuses: %d\n\trooms accessed: %d\n\tlast accessed on %s",
	       userinfo.id,
	       uid,
	       userinfo.last_access_source,
	       userinfo.uses,
	       room_count,
	       ctime(&(userinfo.last_access_time)));
	if (verify) {
	    char response[80];

	    printf("Do you really want to nuke this guy? ");
	    fflush(stdout);
	    gets(response);
	    if (*response != 'y')
		return;
	}
	if (!(remove_user_entry(uid) && remove_uid(userinfo.id))) {
	    printf("server database error\n");
	} else {
	    int zotched_room_count = 0;

	    for (rid = 0; rid <= Last_rid; rid++)
		if (remove_room_user_entry(rid, uid))
		    zotched_room_count++;
	    printf("user %s removed from %d rooms\n", id, zotched_room_count);
	}
    }
}
