#include "../server/server.h"
#include <sys/time.h>

char *message_summary(int number, message_entry * entry, boolean full,
		           int width);

main(int argc, char *argv[])
{
    int message;
    message_entry *message_info;
    char room[ROOM_SIZE];
    time_t now;
    bb_rid rid;
    summary_list *summary;
    summary_entry *entry;
    char message_base[80] = MESSAGEBASE;
    int ch;
    extern char *optarg;
    extern int optind;

    while ((ch = getopt(argc, argv, "M:t")) != EOF)
	switch (ch) {
	case 'M':
	    strcpy(message_base, optarg);
	    break;
	}
    if (chdir(message_base) != 0) {
	perror(message_base);
	return 1;
    }
    if (!database_init(TRUE, FALSE)) {
	perror("database init failed");
	return 1;
    }
    openlog("summary", LOG_PID, LOG_LOCAL2);
    bzero(room, sizeof(room));
    if (optind >= argc) {
	printf("usage: %s [-M messagebase] room\n", argv[0]);
	return 1;
    }
    strcpy(room, argv[optind]);
    if ((rid = get_rid(room)) == -1) {
	fprintf(stderr, "%s: no such room\n", room);
	return 1;
    }
    setlinebuf(stdout);
    now = time(0);
    if (!(summary = get_summary(rid))) {
	fprintf(stderr, "Error retrieving summary for %s (%d)\n",
		room, rid);
	return 1;
    }
    for (entry = summary->list;
	 entry < summary->list + summary->count;
	 entry++) {
	message = entry->number;
	message_info = &(entry->header);
	printf("%s\n", message_summary(message, message_info, TRUE, 80));
    }
    database_shutdown();
    return 0;
}

char *Date(time_t date)
{
    char *d;

    d = ctime(&date);
    *(d + 24) = '\0';
    return d;
}

char *message_summary(int number, message_entry * entry, boolean full,
		           int width)
{
    static char line[BUFFER_SIZE];

    if (full) {
	sprintf(line, "%c %4d %12.12s %4d [%10.10s] %s",
		entry->deleted ? '*' : ' ',
		number,
		Date(entry->date) + 4,
		entry->lines,
		entry->poster_id,
		entry->subject);
    } else {
	sprintf(line, "(%12.12s) %4d [%10.10s] %s",
		Date(entry->date) + 4,
		entry->lines,
		entry->poster_id,
		entry->subject);
    }
    line[width - 1] = '\0';
    return line;
}
