#include "../server/server.h"

char *Classes[] =
{"no_access", "read_only", "read_write", "moderator", "no_class"};

extern int optind;

void list_user(bb_uid uid);

boolean display_uid = FALSE, display_nick = FALSE, display_mail_only = FALSE;
boolean display_groups = FALSE, display_access_time = FALSE;
boolean display_full = FALSE;

int main(int argc, char *argv[])
{
    int ch;
    extern char *optarg;
    char message_base[80] = MESSAGEBASE;

    while ((ch = getopt(argc, argv, "M:Nafgmn")) != EOF)
	switch (ch) {
	case 'M':
	    strcpy(message_base, optarg);
	    break;
	case 'N':
	    display_nick = !display_nick;
	    break;
	case 'a':
	    display_access_time = !display_access_time;
	    break;
	case 'f':
	    display_full = !display_full;
	    break;
	case 'g':
	    display_groups = !display_groups;
	    break;
	case 'm':
	    display_mail_only = !display_mail_only;
	    break;
	case 'n':
	    display_uid = !display_uid;
	    break;
	}

    if (optind > argc) {
	printf("usage: %s [-M messagebase] [-Ngmn] id-list\n", argv[0]);
	return 1;
    }
    if (chdir(message_base) != 0) {
	perror(message_base);
	return 1;
    }
    if (!database_init(TRUE, FALSE)) {
	perror("database init failed");
	return 1;
    }
    if (optind < argc) {
	char *id;

	for (id = argv[optind]; optind < argc; id = argv[++optind])
	    list_user(get_uid(id));
    } else {
	char *flag = "< last uid >";
	bb_uid uid, last_uid = get_uid(flag);

	for (uid = 0; uid <= last_uid; uid++)
	    list_user(uid);
    }

    database_shutdown();
    return 0;
}

void list_user(bb_uid uid)
{
    DbUser user_info;
    bb_gid gid;
    char id_string[120];
    char *mail_address;

    if (!get_db_user(&user_info, uid))
	return;
    mail_address = *user_info.mail_address || display_mail_only ?
      user_info.mail_address : user_info.last_access_source;
    if (!*mail_address)
	return;
    if (display_full) {
	printf("%d\t%d\t%d\t%d %2d:%2.2d:%2.2d\t%s\t%s\t%s\t%s\t%s",
	       (int)uid,
	       user_info.uses,
	       user_info.posts,
	       (int)user_info.connect_time / 60 / 60 / 24,
	       (int)user_info.connect_time / 60 / 60 % 24,
	       (int)user_info.connect_time / 60 % 60,
	       (int)user_info.connect_time % 60,
	       user_info.id,
	       user_info.nick,
	       mail_address,
	       user_info.last_access_source,
	       ctime(&user_info.last_access_time));
    } else if (display_uid) {
	if (display_nick) {
	    if (*user_info.nick)
		printf("%4d %s<%s>\n", (int)uid, user_info.nick, user_info.id);
	    else
		printf("%4d <%s>\n", (int)uid, user_info.id);
	} else {
	    sprintf(id_string, "<%s>", user_info.id);
	    printf("%4d %-22s %s\n", (int)uid, id_string, mail_address);
	}
    } else {
	if (display_nick) {
	    if (*user_info.nick)
		printf("%2d %2d:%2.2d:%2.2d %s<%s>\n",
		       (int)user_info.connect_time / 60 / 60 / 24,
		       (int)user_info.connect_time / 60 / 60 % 24,
		       (int)user_info.connect_time / 60 % 60,
		       (int)user_info.connect_time % 60,
		       user_info.nick,
		       user_info.id);
	    else
		printf("%2d %2d:%2.2d:%2.2d <%s>\n",
		       (int)user_info.connect_time / 60 / 60 / 24,
		       (int)user_info.connect_time / 60 / 60 % 24,
		       (int)user_info.connect_time / 60 % 60,
		       (int)user_info.connect_time % 60,
		       user_info.id);
	} else if (display_access_time) {
	    sprintf(id_string, "<%s> (%d)", user_info.id, (int)uid);
	    printf("%2d %2d:%2.2d:%2.2d %-22s %s",
		   (int)user_info.connect_time / 60 / 60 / 24,
		   (int)user_info.connect_time / 60 / 60 % 24,
		   (int)user_info.connect_time / 60 % 60,
		   (int)user_info.connect_time % 60,
		   id_string,
		   ctime(&user_info.last_access_time));
	} else {
	    sprintf(id_string, "<%s> (%d)", user_info.id, (int)uid);
	    printf("%2d %2d:%2.2d:%2.2d %-22s %s\n",
		   (int)user_info.connect_time / 60 / 60 / 24,
		   (int)user_info.connect_time / 60 / 60 % 24,
		   (int)user_info.connect_time / 60 % 60,
		   (int)user_info.connect_time % 60,
		   id_string,
		   mail_address);
	}
    }
    if (display_groups) {
	for (gid = 0; gid <= Last_gid; gid++) {
	    DbGroup group_entry;

	    get_db_group(&group_entry, gid);
	    printf("\t%10d %-20s %s\n",
		   (int)user_info.group_priority[gid],
		   group_entry.name,
		   Classes[user_info.group_class[gid]]);
	}
    }
}
