#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/server/database.c,v 1.9 1999/08/07 00:36:29 kc Exp $")

#include <fcntl.h>
#include <ndbm.h>
#include <sys/file.h>
#include "server.h"
#include <stdarg.h> /* dmy */

#ifndef COPY
#ifdef X
#define COPY(src, dest, dtype) *dest = *((dtype *) src);
#else
#define COPY(src, dest, dtype) bcopy(src, dest, sizeof(dtype));
#endif
#endif

static FILE *old_dir_f, *dir_f; /* backwards compatiility */
static FILE *user_f;
static FILE *old_summary_f, *summary_f; /* backwards compatibility */
static DBM *room_dbm;

static DbRoom *Room_cache;

#define PadMalloc(type, num) \
    ((num) ? (type *) malloc(pad_up_to_25_pct(num) * sizeof(type)) : 0)

#define PadRealloc(ptr,type,num) \
  ((ptr) ? \
      ((num) ? \
          (type *) realloc(ptr, pad_up_to_25_pct(num)*sizeof(type)) : \
          (type *) (free(ptr), 0) ) : \
      PadMalloc(type, num))


int Directory_stamp, Group_directory_stamp;
bb_rid Last_rid = -1;
bb_gid Last_gid = -1;
bb_uid Last_uid = -1;
bb_rid Last_arid = -1;

boolean Read_only;		/* whether messagebase is read-only */
int Max_active_slot = -1;	/* the maximum slot number currently in use */
struct client_struct Client[MAX_CLIENTS];

 /* an array containing information on each client connection */
time_t Now;
bb_uid Owner_uid = OWNER_UID, Daemon_uid = DAEMON_UID;

#define DATA_ROOM_USER_ID 'i'
#define DATA_UID 'u'
#define DATA_RID 'r'
#define DATA_GID 'g'
#define DATA_GROUP 'G'
#define DATA_SUMMARY 's' /* 9/95, dmy */
#define DATA_ARID 'a' /* 9/95, dmy */



static char *rid_flag = "< last rid >", *arid_flag = "< last arid >",
*gid_flag = "< last gid >", *uid_flag = "< last uid >";


/* summary-entry database functions. */
static int convert_rooms();
static void startup_dump();
static int get_old_db_directory(old_DbDirectory *old_db_directory,
			    bb_rid rid);
static int get_db_directory(DbDirectory *db_directory, bb_rid rid);
static int put_db_directory(bb_rid rid, const DbDirectory *db_directory);


/* summary functions */
static bb_sid get_last_sid();
static int put_last_sid(bb_sid last_sid);
static int add_summary_block(summary_block *sum_block_info);
static int get_summary_block(summary_block *sum_block_info, bb_sid sid);
static int put_summary_block(bb_sid sid, summary_block *sum_block_info);
static int remove_summary_block(bb_sid sid);



/* for debugging -dmy */
void Bleh(const char *fmt, ...)
{
#ifdef dmy_debugging
  FILE *bleh_f;
  va_list args;

  if (bleh_f = fopen("server.log", "a")) {
    va_start(args, fmt);

    vfprintf(bleh_f, fmt, args);
    fprintf(bleh_f, ".\n");

    va_end(args);

    fclose(bleh_f);
  }
#endif
}

char *stop_me()
{
  static char *stop_me_string = "Debugger Halting Execution";
  return stop_me_string;
}


int database_init(boolean read_only, boolean override_locks)
{
    int convert_summary = 0, convert_directory = 0;

    Directory_stamp = Group_directory_stamp = 1;
    Read_only = read_only;

    if (!Read_only) {

      
	if (!(dir_f = fopen(DIRECTORY, "r+"))) {
	  convert_directory = 1;
	  dir_f = fopen(DIRECTORY, "w+");
	}
	if (dir_f && flock(fileno(dir_f), LOCK_EX | LOCK_NB) == -1 &&
	    !override_locks) {
	    fclose(dir_f);
	    return 0;
	}
	if (!(user_f = fopen(USER, "r+"))) {
	    user_f = fopen(USER, "w+");
	}
	if (!(summary_f = fopen(SUMMARY, "r+"))) {
	  convert_summary = 1;
	  summary_f = fopen(SUMMARY, "w+");
	}

	room_dbm = dbm_open(ROOM, O_RDWR | O_CREAT, 0666);
    } else {
	dir_f = fopen(DIRECTORY, "r");
	user_f = fopen(USER, "r");
	summary_f = fopen(SUMMARY, "r");
	room_dbm = dbm_open(ROOM, O_RDONLY, 0666);
    }
    if (room_dbm) {
	Last_rid = get_rid(rid_flag);
	Last_gid = get_gid(gid_flag);
	Last_uid = get_uid(uid_flag);
	Last_arid = get_rid(arid_flag);
    }
    init_cache();
    if (!Read_only && Last_gid < 0) {
    /* This is a new database--perform some initialization */

	DbGroup db_group;
	bb_gid gid;

	syslog(LOG_NOTICE, "Creating default group");
	strcpy(db_group.name, "default");
	db_group.owner = Owner_uid;
	db_group.default_class = DEFAULT_GROUP_CLASS;
	db_group.group_offset = DEFAULT_GROUP_OFFSET;
	db_group.max_expiration = MAX_EXPIRATION;
	gid = add_gid(db_group.name);
	if (!put_db_group(gid, &db_group)) {
	    syslog(LOG_ERR, "database_init: group initialization failed\n");
	    database_shutdown();
	    return FALSE;
	}
	put_last_sid(-1);
    }

    /* 9/95, dmy.  conversion -- dmy */
    if (convert_summary || convert_directory)
      /* dumb, temporary conversion -- dmy */
      if ((old_dir_f = fopen(OLD_DIRECTORY, "r")))
	convert_rooms();

#ifdef debugged_version_to_come
    compress_database();
#endif

    return (dir_f && user_f && room_dbm);
}

void database_shutdown()
{
    dump_cache();
    flock(fileno(dir_f), LOCK_UN);
    fclose(dir_f);
    fclose(user_f);
    dbm_close(room_dbm);
    if (Room_cache) {
	free(Room_cache);
	Room_cache = NULL;
    }
}

/*
 * TOPIC: DbUser -- database user storage.

 Users are stored in a packed user file, arrayed by bb-uid.

 The dbm file holds the username->bb-uid mapping.


 */

/* uid key */
typedef struct {
    char type;
    char name[ID_SIZE];
} uid_key;


static void make_uid_key(datum *key, const char *id)
{
  static uid_key auid_key;
  
  bzero(&auid_key, sizeof(auid_key));
  auid_key.type = DATA_UID;
  downcase(auid_key.name, id);
  key->dptr = (char *)&auid_key;
  key->dsize = sizeof(auid_key);
}

/* basic operations */


bb_uid get_uid(const char *id)
{
    datum key, content;
    bb_uid uid;

    make_uid_key(&key, id);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &uid, bb_uid);
    return uid;
}

/* currently backwards (should be put(key, content)) */
int put_uid(const char *id, bb_uid uid)
{
    datum key, content;
 
    make_uid_key(&key, id);
    content.dptr = (char *)&uid;
    content.dsize = sizeof(uid);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

int remove_uid(const char *id)
{
    datum key;

    make_uid_key(&key,id);
    return dbm_delete(room_dbm, key) >= 0;
}

const char *get_id(bb_uid uid)
{
    static DbUser db_user;

    return get_db_user(&db_user, uid) ? db_user.id : "******";
}

int get_db_user(DbUser *db_user, bb_uid uid)
{
    if  (uid < 0) return FALSE;
    fseek(user_f, uid * sizeof(DbUser), 0);
    if (fread(db_user, sizeof(DbUser), 1, user_f) != 1)
	return FALSE;
    return (db_user->uses != 0);
}

int put_db_user(bb_uid uid, const DbUser *db_user)
{
    if (uid < 0) return FALSE;
    fseek(user_f, uid * sizeof(DbUser), 0);
    if (fwrite(db_user, sizeof(DbUser), 1, user_f) != 1)
	return FALSE;
    return TRUE;
}

int remove_db_user(bb_uid uid)
{
    DbUser db_user;

    bzero(&db_user, sizeof(db_user));
    return put_db_user(uid, &db_user);
}

bb_uid add_uid(const char *id)
{
  Last_uid++;
  put_uid(uid_flag, Last_uid);
  put_uid(id, Last_uid);

  return Last_uid;
}

/*
 * TOPIC: DbDirectory
 *
 * rid -> DbDirectory, from directory file.
 * roomname -> rid, from dbm file.
 */



/* rid key */
typedef struct {
    char type;
    char roomname[ROOM_SIZE];
} rid_key;


static void make_rid_key(datum *key, const char *roomname)
{
  static rid_key arid_key;

  bzero(&arid_key, sizeof(arid_key));
  arid_key.type = DATA_RID;
  downcase(arid_key.roomname, roomname);
  key->dptr = (char *)&arid_key;
  key->dsize = sizeof(arid_key);
}


bb_rid add_rid(const char *roomname)
{
    int i;
    int num_rooms;

    Last_rid++;
    Last_arid++;
    put_rid(rid_flag, Last_rid);
    put_rid(roomname, Last_rid);
    put_rid(arid_flag, Last_arid); /* see also create.c */
    num_rooms = Last_rid + 1;
    Bleh("add_rid: Last_arid %d", Last_arid);
    Room_cache = PadRealloc(Room_cache, DbRoom, num_rooms);
    bzero((char *)&Room_cache[Last_rid], sizeof(DbRoom));

    /* i want to move this out of database.c'add_rid.  it
     * is kind of silly in here.  it belongs in lib.
     */
    
    for (i = 0; i <= Max_active_slot; i++) {
      Client[i].dir_info = PadRealloc(Client[i].dir_info,
				      directory_info *, num_rooms);
      Client[i].dir_info[Last_rid] = NULL;

      /* 9/95, dmy. big bug fix */
      Client[i].stamp = PadRealloc(Client[i].stamp,
				struct client_stamp, num_rooms);
      Client[i].stamp[Last_rid].update = 1;
      Client[i].stamp[Last_rid].sent = 0;

    }

    return Last_rid;
}

bb_rid get_rid(const char *roomname)
{
    datum key, content;
    bb_rid rid;

    make_rid_key(&key, roomname);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &rid, bb_rid);
    return rid;
}

int put_rid(const char *roomname, bb_rid rid)
{
    datum key, content;

    make_rid_key(&key, roomname);
    content.dptr = (char *)&rid;
    content.dsize = sizeof(rid);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

int remove_rid(const char *roomname)
{
    datum key;

    make_rid_key(&key, roomname);
    return (dbm_delete(room_dbm, key) >= 0);
}


const char *get_roomname(bb_rid rid)
{
    static DbDirectory db_dir;

    return get_db_directory(&db_dir, rid) ? db_dir.name : "";
}

/* for backwards compatibility */
static int get_old_db_directory(old_DbDirectory *old_db_directory,
				bb_rid rid)
{
  int fread_val;
  if (rid < 0) return FALSE;
  fseek(old_dir_f, rid * sizeof(old_DbDirectory), 0);
  fread_val = fread(old_db_directory,
		    sizeof(old_DbDirectory), 1, old_dir_f)==1;
  if (fread_val) {
    return (old_db_directory->next_message_number > 0);
  }
  else
    return FALSE;
}


static int get_db_directory(DbDirectory *db_dir, bb_rid rid)
{
  int fread_val;

  if (rid < 0) return FALSE;
  fseek(dir_f, rid * sizeof(DbDirectory), 0);
  fread_val = fread(db_dir, sizeof(DbDirectory), 1, dir_f);
    
  return (fread_val ==1 && db_dir->next_message_number > 0);
}

static int put_db_directory(bb_rid rid,
			    const DbDirectory *db_dir)
{
    if  (rid < 0) return FALSE;
    Bleh("put_db_directory (%d): %d num_msgs, first block %d",
	 rid, db_dir->num_msgs, db_dir->msg_sid[0]);
    fseek(dir_f, rid * sizeof(DbDirectory), 0);
    if (fwrite(db_dir, sizeof(DbDirectory), 1, dir_f)!=1)
	return FALSE;
    Directory_stamp++;
    update_cached_room_user_info(rid, -1);
    return TRUE;
}

static int remove_db_directory(bb_rid rid)
{
  DbDirectory db_dir;

  bzero(&db_dir, sizeof(db_dir));
  return put_db_directory(rid, &db_dir);
}

/*
 * TOPIC: DbGroup
 *
 * groupname -> gid, thru dbm file.
 * gid -> db_group, thru dbm file.
 */

/* gid key */
typedef struct {
    char type;
    char groupname[GROUP_SIZE];
} gid_key;

/* db_group key */
typedef struct {
    char type;
    bb_gid gid;
} db_group_key;



static void make_gid_key(datum *key, const char *groupname)
{
  static gid_key agid_key;

  bzero(&agid_key, sizeof(agid_key));
  agid_key.type = DATA_GID;
  downcase(agid_key.groupname, groupname);
  key->dptr = (char *)&agid_key;
  key->dsize = sizeof(agid_key);
}

bb_gid add_gid(const char *name)
{
    Last_gid++;
    put_gid(gid_flag, Last_gid);
    put_gid(name, Last_gid);
    return Last_gid;
}

bb_gid get_gid(const char *groupname)
{
    datum key, content;
    bb_gid gid;

    make_gid_key(&key, groupname);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return -1;
    COPY(content.dptr, &gid, bb_gid);
    return gid;
}

int put_gid(const char *groupname, bb_gid gid)
{
    datum key, content;

    make_gid_key(&key, groupname);
    content.dptr = (char *)&gid;
    content.dsize = sizeof(gid);
    return dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
}

int remove_gid(const char *groupname)
{
    datum key;

    make_gid_key(&key, groupname);
    return dbm_delete(room_dbm, key) >= 0;
}

/*
 * groupname->gid, from dbm file; gid -> groupname from get_db_group.
 */
const char *get_groupname(bb_gid gid)
{
    static DbGroup db_group;

    return get_db_group(&db_group, gid) ? db_group.name : "";
}


/* DbGroup: gid -> db_group */

static void make_db_group_key(datum *key, bb_gid gid)
{
  static db_group_key adb_group_key;

  bzero(&adb_group_key, sizeof(adb_group_key));
  adb_group_key.type = DATA_GROUP;
  adb_group_key.gid = gid;
  key->dptr = (char *)&adb_group_key;
  key->dsize = sizeof(adb_group_key);
}

int get_db_group(DbGroup *db_group, bb_gid gid)
{
    datum key, content;

    make_db_group_key(&key, gid);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return FALSE;
    COPY(content.dptr, db_group, DbGroup);

    return TRUE;
}

int put_db_group(bb_gid gid, const DbGroup *db_group)
{
    datum key, content;
    int result;

    Group_directory_stamp++;
    make_db_group_key(&key, gid);
    content.dptr = (char *)db_group;
    content.dsize = sizeof(DbGroup);
    result = dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
    if (result)
	update_cached_group_info(gid);
    return result;
}


/*
 * TOPIC: DbRoomUser
 *
 * rid x uid -> DbRoomUser, in dbm file.
 */

/* room_user key */
typedef struct {
    char type;
    bb_rid rid;
    bb_uid uid;
} room_user_key;


static void make_room_user_key(datum *key, bb_rid rid, bb_uid uid)
{
  static room_user_key aroom_user_key;

  bzero(&aroom_user_key, sizeof(aroom_user_key));
  aroom_user_key.type = DATA_ROOM_USER_ID;
  aroom_user_key.rid = rid;
  aroom_user_key.uid = uid;
  key->dptr = (char *)&aroom_user_key;
  key->dsize = sizeof(aroom_user_key);
}

int get_db_room_user(DbRoomUser *db_room_user,
			    bb_rid rid, bb_uid uid)
{
    datum key, content;

    if (rid < 0 || uid < 0) return FALSE;

    make_room_user_key(&key, rid, uid);
    content = dbm_fetch(room_dbm, key);
    if (content.dptr == NULL)
	return FALSE;
    COPY(content.dptr, db_room_user, DbRoomUser);
    return TRUE;
}

int put_db_room_user(bb_rid rid, bb_uid uid,
			const DbRoomUser *db_room_user)
{
    datum key, content;
    int result;

    if (rid < 0 || uid < 0) return FALSE;

    make_room_user_key(&key, rid, uid);
    content.dptr = (char *)db_room_user;
    content.dsize = sizeof(DbRoomUser);
    result = dbm_store(room_dbm, key, content, DBM_REPLACE) >= 0;
    if (result)
	update_cached_room_user_info(rid, uid);
    return result;
}

int remove_db_room_user(bb_rid rid, bb_uid uid)
{
    datum key;

    make_room_user_key(&key, rid, uid);
    return dbm_delete(room_dbm, key) >= 0;
}

/*
 * Message database functions.
 *
 * rid x msg # -> DbMessage, thru DbRoom & indexing.
 */

static int get_message_index(bb_rid rid, int message)
{
    static int index = -1;
    DbSummary **sum_info;

    /* (-1) makes it semantically correct --dmy */
    int last_message = Room_cache[rid].directory->num_msgs - 1;

    if (last_message < 0) return -1;

    sum_info = Room_cache[rid].summary;

    /* if not initialized, start from middle */
    if (index < 0 || index > last_message) index = last_message / 2;

    /* found match */
    if (sum_info[index]->number == message) return index;

    /* count forward */
    if (sum_info[index]->number < message)
      for (; index <= last_message && sum_info[index]->number < message;
	   index++) ;

    /* count backward */
    else
      for (; index >= 0 && sum_info[index]->number > message; index--);

    return (index >= 0 && index <= last_message &&
	    sum_info[index]->number == message) ? index : -1;
}


int db_get_db_message(DbMessage *db_message,
			 bb_rid rid, int message)
{
    int message_index;
    DbRoom *db_room;

    if (!(db_room = get_db_room(rid))) return FALSE;
    if ((message_index = get_message_index(rid, message)) < 0) return FALSE;
    COPY(&db_room->summary[message_index]->header, db_message, DbMessage);
    return TRUE;
}

/* 9/95 db_room->summary is now an array of pointers.  watch for leaks. -dmy */
int put_db_message(bb_rid rid, int message,
		      const DbMessage *db_message)
{
    int message_index;
    DbRoom *db_room;

    Bleh("put_message (%d) , message %d, %s", rid, message,
	 db_message->poster_id);

    if (!(db_room = get_db_room(rid))) return FALSE;

    if ((message_index = get_message_index(rid, message)) < 0) {
	int j;

    /* Find the message just prior to where we'll be inserting this one. */

	for (message_index = db_room->directory->num_msgs - 1;
	     message_index >= 0 && db_room->summary[message_index]->number > message;
	     message_index--);

    /* Move past it. */

	message_index++;

    /* Allocate the extra memory for this message. */

	db_room->directory->num_msgs++;
	db_room->summary = PadRealloc(db_room->summary, DbSummary *,
					db_room->directory->num_msgs);

    /* Shift messages from i through the last message. */

	for (j = db_room->directory->num_msgs - 1; j > message_index; j--)
	  db_room->summary[j] = db_room->summary[j - 1];

    /* Put the message here! */
	db_room->summary[message_index] =
	  (DbSummary *) malloc(sizeof(DbSummary));
	db_room->summary[message_index]->number = message;
    }

    COPY(db_message, &db_room->summary[message_index]->header, DbMessage);
    db_room->version++;
    return TRUE;
}

/* db_room->summary is now array of ptrs.  Watch for leaks. 9/95, dmy. */
int remove_db_message(bb_rid rid, int message)
{
    int index;
    DbRoom *db_room;
    DbDirectory *db_dir;

    if (!(db_room = get_db_room(rid))) return FALSE;
    db_dir = db_room->directory; /* brevity */

    if ((index = get_message_index(rid, message)) < 0) return FALSE;

    free(db_room->summary[index]);

    db_dir->num_msgs--;

    for (; index < db_dir->num_msgs; index++)
      db_room->summary[index] = db_room->summary[index + 1];
    db_room->summary =
      PadRealloc(db_room->summary, DbSummary *, db_dir->num_msgs);
    db_room->version++;

    return TRUE;
}

/*
 * Topic: SummaryBlock
 *
 * db_dir holds summary_count, block numbers.
 * rid x block_num -> db_summary, thru summary file.
 * --dmy
 */


/*
 * summary functions.
 */

static bb_sid get_last_sid()
{
  bb_sid last_sid;

  fseek(summary_f, 0, 0);
  if (fread(&last_sid, sizeof(last_sid), 1, summary_f) >= 0)
    return last_sid;
  else
    return -1;
}

static int put_last_sid(bb_sid last_sid)
{
  fseek(summary_f, 0, 0);
  return (fwrite(&last_sid, sizeof(last_sid), 1, summary_f) >= 0);
}
    

static int add_summary_block(summary_block *sum_block_info)
{
  int last_sid = get_last_sid() + 1;

  put_last_sid(last_sid);
  put_summary_block(last_sid, sum_block_info);
  return last_sid;
}

static int get_summary_block(summary_block *sum_block_info, bb_sid sid)
{
  int retval;

  if (sid < 0 || sid > get_last_sid()) return FALSE;
  fseek(summary_f, sizeof(long) + sid * sizeof(summary_block), 0);
  retval = fread(sum_block_info, sizeof(summary_block), 1, summary_f) >= 0;
  return retval;
}

static int put_summary_block(bb_sid sid, summary_block *sum_block_info)
{
  if (sid < 0 || sid > get_last_sid()) return FALSE;
  fseek(summary_f, sizeof(long) + sid * sizeof(summary_block), 0);
  return (fwrite(sum_block_info, sizeof(summary_block), 1, summary_f) >= 0);
}

/* save and replace. */
static int remove_summary_block(bb_sid sid)
{
  summary_block sum_block_info;
  int last_sid = get_last_sid();
  DbRoom *db_room;

  if (sid < 0 || sid > last_sid) return FALSE;
  if (sid != last_sid) {
    /* move last dummary block to removal position */
    get_summary_block(&sum_block_info, last_sid);
    put_summary_block(sid, &sum_block_info);
    if ((db_room = get_db_room(sum_block_info.rid))) {
      /* update x-ref */
      db_room->directory->msg_sid[sum_block_info.block_num] = sid;
      db_room->version++;
    }
    else {
      syslog(LOG_ERR, "remove_summary_block (%d): room %d no data",
	     sid, sum_block_info.rid);
      return FALSE;
    }
  }
  /* clean out last position */
  bzero(&sum_block_info, sizeof(sum_block_info));
  put_summary_block(last_sid, &sum_block_info);
  put_last_sid(last_sid-1);
  return TRUE;
}


/* constructors for directory entries */
static DbDirectory *create_db_directory()
{
  DbDirectory *db_dir = (DbDirectory *) malloc(sizeof(DbDirectory));

  if (db_dir) {
    int block_index;

    bzero(db_dir, sizeof(DbDirectory));
    db_dir->arid = -1;
    db_dir->num_msgs = 0;
    for (block_index=0; block_index<SUMMARY_BLOCKS; block_index++)
      db_dir->msg_sid[block_index] = -1;
  }
  return db_dir;
}

/* converts, setting new DbDirectory values to defaults */  
static DbDirectory *
create_db_directory_from_old(old_DbDirectory *old_db_directory)
{
  DbDirectory *db_dir = create_db_directory();

  if (db_dir)
    COPY(old_db_directory, db_dir, old_DbDirectory);

  return db_dir;
} 


/* get an old-style summary list.  Used for conversion. */
DbRoom *get_old_db_room(bb_rid rid)
{
  DbSummary **sum_info;
  DbDirectory *db_dir = NULL;
  old_DbDirectory old_db_directory;
  int count;

  if (rid > Last_rid) {
    syslog(LOG_ERR, "get_old_summary(%d): rid > Last_rid (%d)",
	   rid, Last_rid);
    return NULL;
  }

  /* 9/95, dmy.  Hey, wait. Something's wrong with this conditional. */
#ifdef silly_conditional
  if (!Room_cache)
#endif
    Room_cache[rid].last_access = Now;
  /* are we current? */
  if (Room_cache[rid].version > 0)
    return &Room_cache[rid];

  /* old, sparse search */
  fseek(old_summary_f,
	rid * (sizeof(int) + OLD_MAX_MESSAGES * sizeof(DbSummary)),
	0);
  if (fread(&count, sizeof(count), 1, old_summary_f) != 1)
    count = 0;
  
  if (count > OLD_MAX_MESSAGES) {
    syslog(LOG_ERR, "get_old_summary(%d): invalid count: %d\n", count);
    return NULL;
  }

  if (count > 0) {
    int message_index;

    sum_info = PadMalloc(DbSummary *, count);

    if (!sum_info) {
      syslog(LOG_ERR, "get_old_db_room (%d): malloc: %m\n", rid);
      return NULL;
    }
    for (message_index=0; message_index<count; message_index++) {
      sum_info[message_index] =
	(DbSummary *) malloc(sizeof(DbSummary));
      if (fread(sum_info[message_index], sizeof(DbSummary), 1, old_summary_f)
	  == -1) {
	syslog(LOG_ERR, "get_old_db_room (%d): fread: %m\n", rid);
	free(sum_info);
	return NULL;
      }
    }
  }
  else
    sum_info = NULL;

  if (get_old_db_directory(&old_db_directory, rid) &&
      (db_dir = create_db_directory_from_old(&old_db_directory))) {
    db_dir->arid = rid;
    db_dir->num_msgs = count;
  }
  else {
    syslog(LOG_ERR, "%s, get_old_db_room (%d): db_dir: %m\n",
	   stop_me(), rid);

    free(db_dir);
    free(sum_info);
    return NULL;
  }
				
  Room_cache[rid].summary = sum_info;
  Room_cache[rid].directory = db_dir;
  Room_cache[rid].saved_version = 0; /* data is unsaved */
  Room_cache[rid].version = 1;
  return &Room_cache[rid];
}



/* 9/95, dmy.  get room from database */
DbRoom *get_db_room(bb_rid rid)
{
  DbSummary **sum_info; /* 9/95, dmy */
  DbDirectory *db_dir;
  int count;
  int version_rollover_worry = 987654321;


  if (rid < 0 || rid > Last_rid) {
    syslog(LOG_ERR, "%s, get_db_room(%d): bad rid (%d)",
	   stop_me(), rid, Last_rid);
    return NULL;
  }

  if (!Room_cache) {
    syslog(LOG_ERR, "get_db_room (%d): no cache: %m\n", rid);
    return NULL;
  }

  /* if version number gets too high, stop updating last_access so
   * that cache eventually gets flushed. --dmy
   */
  if (Room_cache[rid].version < version_rollover_worry)
    Room_cache[rid].last_access = Now;


  /* if cache exists, return handle */
  if (Room_cache[rid].version > 0) {
    return &Room_cache[rid];
  }

  /* load directory info */
  db_dir = create_db_directory(); /* constructor */

  /* room does not exist on disk */
  if (!get_db_directory(db_dir, rid)) {
    free(db_dir);
    return NULL;
  }

  count = db_dir->num_msgs;

  if (count < 0 || count > MAX_MESSAGES) {
    syslog(LOG_ERR, "get_db_room (%d): count (%d) overflow: %m\n",
	   rid, count);
    free(db_dir);
    return NULL;
  }

  /* load summary info */
  if (count > 0) {
    int block_num;

    sum_info = PadMalloc(DbSummary *, count);

    if (!sum_info) {
      syslog(LOG_ERR, "get_db_room(%d): malloc: %m\n", rid);
      free(db_dir);
      return NULL;
    }

    /* read summary blocks from summary file */
    for (block_num=0; block_num*SUMMARY_BLOCK_SIZE<count; block_num++) {
      int block_start = block_num * SUMMARY_BLOCK_SIZE;
      int msgs_in_block = MIN(count - block_start, SUMMARY_BLOCK_SIZE);
      summary_block sum_block_info;
      int block_msg_index;

      if (!get_summary_block(&sum_block_info, db_dir->msg_sid[block_num])) {
	syslog(LOG_ERR, "get_db_room(%d): Bad block %d: %m\n",
	       rid, block_num);
	bzero(&sum_block_info, sizeof(summary_block));
      }

      for (block_msg_index = 0;
	   block_msg_index < msgs_in_block;
	   block_msg_index++) {
	int message_index = block_start + block_msg_index;
	sum_info[message_index] = (DbSummary *) malloc(sizeof(DbSummary));
	*sum_info[message_index] = sum_block_info.entry[block_msg_index];
      }
    }

  }
  else
    sum_info = NULL;

  if (Room_cache[rid].summary) free(Room_cache[rid].summary);
  Room_cache[rid].summary = sum_info; 
  Room_cache[rid].directory = db_dir;
  Room_cache[rid].version = Room_cache[rid].saved_version = 1;
  return &Room_cache[rid];
}

DbDirectory *get_db_room_directory(bb_rid rid)
{
  DbRoom *db_room;

  return (db_room = get_db_room(rid)) ? db_room->directory : NULL;
}



int get_db_room_arid(bb_rid rid)
{
  DbDirectory *db_dir;

  return (db_dir = get_db_room_directory(rid)) ? db_dir->arid : -1;
}

/* changed, 9/95, dmy.  write the Room_cache */
int put_db_room(bb_rid rid)
{
  int count;
  int block_num;
  DbRoom *db_room;
  DbDirectory *db_dir;

  if (Read_only) return FALSE;
  if (!(db_room = get_db_room(rid)) || !(db_dir = db_room->directory))
    return FALSE;

  count = db_dir->num_msgs;
  if (count < 0 || count > MAX_MESSAGES) {
    syslog(LOG_ERR, "put_db_room (%d): bad count (%d)\n", rid, count);
    return FALSE;
  }

  /* put new messages */
  for (block_num=0; block_num*SUMMARY_BLOCK_SIZE<count; block_num++) {
    static summary_block sum_block_info;
    int block_start = block_num * SUMMARY_BLOCK_SIZE;
    int msgs_in_block = MIN(count - block_start, SUMMARY_BLOCK_SIZE);
    int block_msg_index;

    /* x-refs */
    bzero(&sum_block_info, sizeof(summary_block));
      
    sum_block_info.rid = rid;
    sum_block_info.block_num = block_num;

    for (block_msg_index=0; block_msg_index < msgs_in_block; block_msg_index++)
      if (db_room->summary[block_start + block_msg_index])
	COPY(db_room->summary[block_start + block_msg_index],
	     &sum_block_info.entry[block_msg_index], DbSummary);

    if (db_dir->msg_sid[block_num] >= 0) {
      if (!put_summary_block(db_dir->msg_sid[block_num],
			    &sum_block_info)) {
	syslog(LOG_ERR, "put_db_room (%d): can't put block %d: %m\n",
	       rid, block_num);
      }
    }
    else {
      db_dir->msg_sid[block_num] =
	add_summary_block(&sum_block_info);
    }
  }	

  /* remove any excess messages */
  for (; block_num<SUMMARY_BLOCKS; block_num++)
    if (db_dir->msg_sid[block_num] >= 0) {
      remove_summary_block(db_dir->msg_sid[block_num]);
      db_dir->msg_sid[block_num] = -1;
    }

  /* put directory info */
  put_db_directory(rid, db_dir);

 /* update saved_version number */
  Room_cache[rid].saved_version = Room_cache[rid].version;
  return TRUE;
}

/* changed, 9/95, dmy.  remove the room's saved data entry */
int remove_db_room(bb_rid rid)
{
  int block_num;
  DbRoom *db_room;
  DbDirectory *db_dir;

  if (Read_only) return FALSE;

  if (!(db_room = get_db_room(rid))) {
    syslog(LOG_ERR, "remove_db_room (%d): no room info: %m\n", rid);
    return FALSE;
  }
  if (!(db_dir = db_room->directory)) {
    syslog(LOG_ERR, "remove_db_room (%d): no dir info: %m\n", rid);
    return FALSE;
  }

  /* remove directory data */
  remove_db_directory(rid);

  /* remove any excess messages */
  for (block_num=0; block_num<SUMMARY_BLOCKS; block_num++)
    if (db_dir->msg_sid[block_num] >= 0) {
      remove_summary_block(db_dir->msg_sid[block_num]);
      db_dir->msg_sid[block_num] = -1;
    }

  /* set cache room to dead */
  db_room->version = 0;
  *db_dir->name = '\0';
  db_dir->next_message_number = 0;

  return TRUE;
}

int add_db_room(char *roomname)
{
  bb_rid rid;
  DbRoom *db_room; /* alias */
  DbDirectory *db_dir; /* alias */
  int i;

  /* rid == Last_rid */
  rid = add_rid(roomname);

  Room_cache = PadRealloc(Room_cache, DbRoom, rid + 1);
  db_room = &Room_cache[rid];

  /* initialize DbRoom */
  bzero((char *)db_room, sizeof(DbRoom));
  db_room->directory = db_dir = create_db_directory();

  /* initialize directory */
  db_dir->arid = Last_arid;
  strcpy(db_dir->name, roomname);
  db_dir->next_message_number = 1;
  db_dir->first_message_number = 1;
  db_dir->creation_time = time(NULL);

  /* unsaved */
  db_room->saved_version = 0;
  db_room->version = 1;
  db_room->summary = NULL;

  /* initialize Client data */
  for (i = 0; i <= Max_active_slot; i++) {
    Client[i].dir_info = PadRealloc(Client[i].dir_info, /* UMR caused here */
				    directory_info *, rid + 1);
    Client[i].dir_info[rid] = NULL;

    /* 9/95, dmy. big bug fix */
    Client[i].stamp = PadRealloc(Client[i].stamp,
				 struct client_stamp, rid + 1);
    Client[i].stamp[rid].update = 1;
    Client[i].stamp[rid].sent = 0;

  }
  return rid;
}

/* clear, but do not destroy db_room */
void db_room_clear(DbRoom *db_room)
{
  int message_index, num_msgs;

  if (db_room) {
    num_msgs = db_room->directory->num_msgs;
    free(db_room->directory);
    db_room->directory = NULL;
    if (db_room->summary) {
      for (message_index=0; message_index<num_msgs; message_index++)
	free(db_room->summary[message_index]);
      free(db_room->summary);
    }
    db_room->summary = NULL;
    db_room->version = 0;
  }
}
    


/*
 * Cache operations.
 */
void init_cache()
{
    int rid;
    
    if (Room_cache) {
      free(Room_cache);
      Room_cache = NULL;
    }
    if (Last_rid >= 0) {
      Room_cache = PadRealloc(Room_cache, DbRoom, Last_rid+1);
      bzero((char *)Room_cache, (Last_rid + 1) * sizeof(DbRoom));
      for (rid=0; rid<=Last_rid; rid++)
	Room_cache[rid].version = Room_cache[rid].saved_version = 0;
    } else {
      Room_cache = NULL;
    }
}

void flush_cache()
{
  bb_rid rid;

  for (rid = 0; rid <= Last_rid; rid++)
    if (Room_cache[rid].version != Room_cache[rid].saved_version)
      put_db_room(rid);
}

void dump_cache()
{
  bb_rid rid;

  if (!Room_cache)
    return;
  for (rid = 0; rid <= Last_rid; rid++)
    if (Room_cache[rid].version) {
      if (Room_cache[rid].version > Room_cache[rid].saved_version)
	put_db_room(rid);

      db_room_clear(&Room_cache[rid]);
    }
}

void check_cache()
{
  bb_rid rid;

  for (rid = 0; rid <= Last_rid; rid++)
    if (Room_cache[rid].version) {
      if (Room_cache[rid].version > Room_cache[rid].saved_version)
	put_db_room(rid);
      if (Room_cache[rid].summary &&
	  (Now - Room_cache[rid].last_access > CACHE_TIMEOUT ||
	   Max_active_slot < 0)) {
	db_room_clear(&Room_cache[rid]);
      }
    }
}

void update_cached_room_user_info(bb_rid rid, bb_uid uid)
{
    int slot;

    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on &&
	    (Client[slot].uid == uid || uid == -1)) {
	    if (Client[slot].dir_info && Client[slot].dir_info[rid]) {
		free(Client[slot].dir_info[rid]);
		Client[slot].dir_info[rid] = NULL;
	    }
	    Client[slot].update_directory = 0;
	    if (Client[slot].stamp)
		Client[slot].stamp[rid].update++;
	}
}

boolean update_cached_group_info(bb_gid gid)
{
    int slot;
    DbGroup db_group;

    if (!get_db_group(&db_group, gid)) {
	ERROR("error accessing group information");
	return FALSE;
    }
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on) {
	    DbUser userinfo;
	    bb_rid rid;

	    if (!get_db_user(&userinfo, Client[slot].uid))
		continue;

	    Client[slot].group_priority[gid] =
	      userinfo.group_priority[gid] + db_group.group_offset;
	    Client[slot].group_class[gid] = userinfo.group_class[gid];

	    for (rid = 0; rid <= Last_rid; rid++) {
		if (Client[slot].dir_info[rid] &&
		    GROUP_ISSET(gid, &Client[slot].dir_info[rid]->groups)) {
		    free(Client[slot].dir_info[rid]);
		    Client[slot].dir_info[rid] = NULL;
		    Client[slot].stamp[rid].update++;
		}
	    }
	    Client[slot].update_directory = 0;
	}
    return TRUE;
}

void database_changelast(bb_gid lgid, bb_rid lrid, bb_uid luid)
{
         Last_gid = lgid;
    Last_rid = lrid;
    Last_uid = luid;
    put_gid(gid_flag, Last_gid);
    put_rid(rid_flag, Last_rid);
    put_uid(uid_flag, Last_uid);
    init_cache();
}


/*
 * startup routines`
 */

/* 9/95, dmy.  converts the old summary file to the new one
 * return 1 on success.
 */
static int convert_rooms()
{
  bb_rid rid;
  DbRoom *db_room;
  char old_sumname[128], old_dirname[128];

  syslog(LOG_NOTICE, "Converting Database from Older Format");

  put_last_sid(-1);

  if (!(old_summary_f = fopen(OLD_SUMMARY, "r"))) return FALSE;

  for (rid=0; rid <= Last_rid; rid++) {
    db_room = get_old_db_room(rid);
    put_db_room(rid);
  }
  /* new element */
  Last_arid = Last_rid;
  put_rid(arid_flag, Last_arid);
  fclose(old_dir_f);
  fclose(old_summary_f);
  sprintf(old_sumname, "%s.old", OLD_SUMMARY);
  sprintf(old_dirname, "%s.old", OLD_DIRECTORY);
  return (rename(OLD_SUMMARY, old_sumname) >= 0 &&
	  rename(OLD_DIRECTORY, old_dirname) >= 0);
}



/* 9/95, dmy.  This does a whole bunch of things.
 *  needs to handle client data.
 */
void compress_database()
{
  bb_rid *new_rid, old_rid, num_rid;
  bb_uid *new_uid, old_uid, num_uid, max_uid;
  DbUser db_user;

  /* don't run when people are on -dmy */
  if (Max_active_slot >= 0)
    return;

  syslog(LOG_INFO, "Compressing Database");

  /* renumber non-null users */
  new_uid = (bb_uid *) malloc((Last_uid+1) * sizeof(bb_uid));
  for (old_uid = 0, num_uid = 0, max_uid = -1;
       old_uid <= Last_uid;
       old_uid++) {
    /* hold OWNER, DAEMON uid's fixed */
    if (DAEMON(old_uid) || OWNER(old_uid)) {
      new_uid[old_uid] = old_uid;
      if (max_uid < old_uid)
	max_uid = old_uid;
    }
    else {
      while (DAEMON(num_uid) || OWNER(num_uid))
	num_uid++;
      new_uid[old_uid] = get_db_user(&db_user, old_uid) ? num_uid++ : -1;
    }
  }
  if (max_uid < num_uid - 1)
    max_uid = num_uid - 1;

  /* compress users and uid's */
  for (old_uid = 0; old_uid <= Last_uid; old_uid++) {
    bb_uid anew_uid = new_uid[old_uid]; /*  brevity */
    if (anew_uid >= 0) {
      if (old_uid != anew_uid) {
	get_db_user(&db_user, old_uid);
	put_db_user(anew_uid, &db_user);
	remove_db_user(old_uid);
	put_uid(db_user.id, anew_uid);
      }
    }
    else
      remove_db_user(old_uid);
  }


  /* renumber non-null rooms  */
  new_rid = (bb_rid * ) malloc((Last_rid+1) * sizeof(bb_rid));
  for (old_rid = 0, num_rid = 0; old_rid <= Last_rid; old_rid++) {
    new_rid[old_rid] = get_db_room(old_rid) ? num_rid++ : -1;
  }

  /* compress rooms, rid's, summaries, user_rooms, and messages */
  for (old_rid = 0; old_rid <= Last_rid; old_rid++) {
    DbRoom *db_room;
    DbDirectory *db_dir;
    DbSummary *db_summary;
    DbRoomUser db_room_user;
    int sum_count, message_index;
    int block_num;
    bb_rid anew_rid = new_rid[old_rid]; /*  brevity */

    if (anew_rid >= 0) {
      if (!(db_room = get_db_room(old_rid))) {
	syslog(LOG_ERR, "compress: can't get room (%d) data", old_rid);
      }
      db_dir = db_room->directory;
      sum_count = db_dir ? db_dir->num_msgs : -1;

      /* update summaries:  update poster_id's */
      if (sum_count > 0) {
	/* update posters' id's */
	for (message_index = 0; message_index < sum_count; message_index++) {
	  db_summary = db_room->summary[message_index];
	  old_uid = db_summary->header.poster_uid;
	  db_summary->header.poster_uid =
	    old_uid >= 0 && old_uid <= Last_uid ? new_uid[old_uid] : -1;
	}
      }

      /* update summary blocks for this room */
      for (block_num = 0; block_num < SUMMARY_BLOCKS; block_num++) {
	if (db_dir->msg_sid[block_num] >= 0) {
	  summary_block sum_block_info;
	  if (get_summary_block(&sum_block_info, db_dir->msg_sid[block_num])) {
	    if (sum_block_info.rid == old_rid) {
	      sum_block_info.rid = anew_rid;
	      put_summary_block(db_dir->msg_sid[block_num], &sum_block_info);
	    }
	    else {
	      syslog(LOG_ERR, "compress: update summary_block %d: %d != %d",
		     db_dir->msg_sid[block_num], sum_block_info.rid, old_rid);
	      db_dir->msg_sid[block_num] = -1;
	    }
	  }
	  else
	    syslog(LOG_ERR, "compress: cannot load summary_block %d",
		   db_dir->msg_sid[block_num]);
	}
      }

      /* compress room and rid */
      old_uid = db_dir->owner_uid;
      db_dir->owner_uid = 
	old_uid >= 0 && old_uid <= Last_uid ? new_uid[old_uid] : -1;
      if (old_rid != anew_rid) {
	/* vacate old directory spot */
	if (!remove_db_directory(old_rid))
	  syslog(LOG_ERR, "compress: can't remove dir_entry %d", old_rid);
	if (!put_rid(db_dir->name, anew_rid))
	  syslog(LOG_ERR, "compress: can't put rid %d", anew_rid);

	COPY(db_room, &Room_cache[anew_rid], DbRoom);

	bzero(Room_cache+old_rid, sizeof(DbRoom));
      }
      if (!put_db_directory(anew_rid, db_dir))
        syslog(LOG_ERR, "compress: can't put dir_entry %d", anew_rid);

      /* compress DbRoomUser entries */
      for (old_uid=0; old_uid <= Last_uid; old_uid++)
	if (get_db_room_user(&db_room_user, old_rid, old_uid)) {
	  bb_uid anew_uid = new_uid[old_uid]; /* brevity */

	  if (old_rid != anew_rid || old_uid != anew_uid) {
	    if (!remove_db_room_user(old_rid, old_uid))
	      syslog(LOG_ERR, "compress: can't remove DbRoomUser %d, %d",
		     old_rid, old_uid);
	    if (!put_db_room_user(anew_rid, anew_uid, &db_room_user))
	      syslog(LOG_ERR, "compress: can't put DbRoomUser %d, %d",
		     anew_rid, anew_uid);
	  }
	}
    }
  }

  /* write stats */
  syslog(LOG_ERR,
	 "Compress: (%d->%d) Users, (%d->%d) Rooms, %d Summary Blocks",
	 Last_uid+1, max_uid+1, Last_rid+1, num_rid, get_last_sid()+1);

  /* store new sizes, resize, dump rooms */
  Last_rid = num_rid-1;
  Last_uid = max_uid;
  put_rid(rid_flag, Last_rid);
  put_uid(uid_flag, Last_uid);

  Room_cache = PadRealloc(Room_cache, DbRoom, Last_rid + 1);

  {
    bb_rid rid;
    for (rid=0; rid <= Last_rid; rid++)
      put_db_room(rid);
  }

  free(new_rid);
  free(new_uid);
  startup_dump();
  syslog(LOG_INFO, "Database Compression Complete");

}

/* display values at startup time to server.log file -dmy */
static void startup_dump()
{
  bb_rid rid;
  bb_uid uid;
  DbUser db_user;
  DbDirectory db_dir;
  DbRoom *db_room;
  DbSummary *db_summary;
  DbRoomUser db_room_user;
  FILE *dump_f;
  char temp_buf1[32], temp_buf2[32];

  syslog(LOG_ERR, "Writing Startup Room and Users Lists");

  if ((dump_f = fopen(STARTUP_USERS, "w"))) {
    fprintf(dump_f, "Users in database since last reboot:\n");
    fprintf(dump_f, "id (name)\n\n");
    for (uid=0; uid<=Last_uid; uid++)
      if (get_db_user(&db_user, uid))
	fprintf(dump_f, "%4d (%s)\n", (int)uid, get_id(uid));
    fclose(dump_f);
  }

  if ((dump_f = fopen(STARTUP_ROOMS, "w"))) {
    fprintf(dump_f, "Rooms in database since last reboot:\n");
    fprintf(dump_f, "id [abs-id] (name) owner-id (owner-name)\n\n");
    for (rid=0; rid <= Last_rid; rid++)
      if (get_db_directory(&db_dir, rid)) {
	sprintf(temp_buf1, "[%d]", (int)db_dir.arid);
	sprintf(temp_buf2, "(%s)", db_dir.name);
	fprintf(dump_f, "%4d %6s %-22s %4d (%s)\n",
		(int)rid, temp_buf1, temp_buf2, (int)db_dir.owner_uid,
		get_id(db_dir.owner_uid));
      }
    fclose(dump_f);
  }
  

  return;

  /* old verbose message garbage */
  Bleh("\n*USER DUMP*");
  for (uid=0; uid<=Last_uid; uid++)
    Bleh("  User %3d, ID %s", uid, get_id(uid));

  Bleh("\n*ROOM DUMP*");
  for (rid=0; rid <= Last_rid; rid++) {
    int count, index;
    db_room = get_db_room(rid);
    count = db_room ? db_room->directory->num_msgs : -1;
    if (get_db_directory(&db_dir, rid)) {
      Bleh("  Room %3d, ID:\t%d, Name:\t%s",
	   rid, db_dir.arid, db_dir.name);
      Bleh("       %3d messages, Owner:\t%d (%s)",
	   count, db_dir.owner_uid, get_id(db_dir.owner_uid));
    }
    else
      Bleh("Room %3d has %d messages, no data", rid, count);
    Bleh("\n  ROOM_USER_DATA (room %d %s):",
	 rid, db_dir.name);
    for (uid=0; uid<=Last_uid; uid++)
      if (get_db_room_user(&db_room_user, rid, uid))
	Bleh("    user %2d, last_read %2d, offset %d, username %s",
	     uid, db_room_user.last_message_read,
	     db_room_user.user_offset, get_id(uid));
    Bleh("\n  SUMMARY_FILE_DATA (room %d %s, count %d):",
	 rid, db_dir.name, count);
    for (index=0; index<count+10; index++) {
      if (index < count) {
	db_summary = db_room->summary[index];
	Bleh("    #%2d, poster %s, deleted %c, subj %s",
	     db_summary->number, db_summary->header.poster_id,
	     db_summary->header.deleted ? 'Y' : 'N',
	     db_summary->header.subject);
      }
    }
  }
}
