#ifndef _BB_SERVER_DATABASE_H
#define _BB_SERVER_DATABASE_H

#include "../lib/types.h"

typedef long bb_uid;
typedef long bb_rid;
typedef long bb_gid;
typedef long bb_arid; /* room id, absolute */
typedef long bb_sid; /* summary block */

#define SUMMARY_BLOCK_SIZE 32 /*  9/95, dmy; max # summaries in a block. */
#define SUMMARY_BLOCKS 128 /* bigger version */
#define MAX_MESSAGES (SUMMARY_BLOCK_SIZE*SUMMARY_BLOCKS)
#define OLD_MAX_MESSAGES 2048

typedef struct {
    char name[ROOM_SIZE];
    bb_uid owner_uid;
    time_t creation_time;
    time_t default_expiration;
    class_t default_class;
    long next_message_number;
    char summary_line[LINE_SIZE];
    bb_archive_t archive;
    long first_message_number;
    group_set groups;
    bb_priority room_offset;
} old_DbDirectory;


typedef struct {
    char name[ROOM_SIZE];
    bb_uid owner_uid;
    time_t creation_time;
    time_t default_expiration;
    class_t default_class;
    long next_message_number;
    char summary_line[LINE_SIZE];
    bb_archive_t archive;
    long first_message_number;
    group_set groups;
    bb_priority room_offset;
    bb_arid arid; /* new data must go at the end */
    int num_msgs;
    bb_sid msg_sid[SUMMARY_BLOCKS];
} DbDirectory;

typedef struct {
    char name[GROUP_SIZE];
    bb_uid owner;
    class_t default_class;
    bb_priority group_offset;
    time_t max_expiration;
} DbGroup;

typedef struct {
    long last_message_read;
    bb_priority user_offset;
    class_t class;
} DbRoomUser;

typedef struct {
    time_t date;
    time_t expiration;
    int size;
    int lines;
    boolean deleted;
    bb_uid poster_uid;
    char poster_nick[LINE_SIZE];
    char poster_id[ID_SIZE];
    char subject[LINE_SIZE];
} DbMessage;

typedef struct {
    DbMessage header;
    int number;
} DbSummary;

/* summary file is composed of summary_blocks */
typedef struct {
  bb_rid rid;
  bb_sid block_num;
  DbSummary entry[SUMMARY_BLOCK_SIZE];
} summary_block;
  
/* this is a cache-data struct for both the summary and dir_entry */
typedef struct {
    int version; /* 0 == unread */
    int saved_version;
    time_t last_access;
    DbDirectory *directory;
    /* NIL if unread or directory is empty -dmy.
       9/95 Array of pointers for faster msg expire. -dmy
       */
    DbSummary **summary;
} DbRoom;

typedef struct {
    time_t last_access_time;
    char id[ID_SIZE];
    char passwd[32];
    char nick[LINE_SIZE];
    char last_access_source[LINE_SIZE];
    char mail_address[LINE_SIZE];
    int uses;
    int posts;
    time_t connect_time;
    long UNUSED;
    bb_priority cutoff_priority;
    bb_priority group_priority[MAX_GROUPS];
    class_t group_class[MAX_GROUPS];
} DbUser;

/* Database variables, found in database.c */

extern bb_rid Last_rid;
extern bb_gid Last_gid;
extern bb_uid Last_uid;

extern boolean Read_only;
extern int Directory_stamp, Group_directory_stamp;
extern bb_uid Owner_uid, Daemon_uid;
extern time_t Now;

/* Database routines, found in database.c */

int database_init(boolean read_only, boolean override_locks);
void database_shutdown();

/* DbUser database functions */
bb_uid get_uid(const char *id);
int put_uid(const char *id, bb_uid uid);
int remove_uid(const char *id);
bb_uid add_uid(const char *id);
const char *get_id(bb_uid uid);
int get_db_user(DbUser *db_user, bb_uid uid);
int put_db_user(bb_uid uid, const DbUser *db_user);
int remove_db_user(bb_uid uid);

/* DbDirectory database functions */
bb_rid get_rid(const char *roomname);
bb_rid add_rid(const char *roomname);
int put_rid(const char *roomname, bb_rid rid);
int remove_rid(const char *roomname);
bb_rid get_last_arid();
const char *get_roomname(bb_rid rid);

/* DbGroup database functions */
bb_gid get_gid(const char *groupname);
bb_gid add_gid(const char *groupname);
int put_gid(const char *groupname, bb_gid gid);
int remove_gid(const char *groupname);
const char *get_groupname(bb_gid gid);
int get_db_group(DbGroup *db_group, bb_gid gid);
int put_db_group(bb_gid gid, const DbGroup *db_group);
int remove_db_group(bb_gid gid);

/* DbRoomUser database functions */
int get_db_room_user(DbRoomUser *db_room_user,
			bb_rid rid, bb_uid uid);
int put_db_room_user(bb_rid rid, bb_uid uid,
			const DbRoomUser *db_room_user);
int remove_db_room_user(bb_rid rid, bb_uid uid);

/* message-header database functions */
int db_get_db_message(DbMessage *db_message,
			 bb_rid rid, int message);
int put_db_message(bb_rid rid, int message,
		      const DbMessage *db_message);
int remove_db_message(bb_rid rid, int message);

/* DbRoom (directory + summary) database functions */
DbRoom *get_old_db_room(bb_rid rid);
DbRoom *get_db_room(bb_rid rid);
DbDirectory *get_db_room_directory(bb_rid rid);
int get_db_room_arid(bb_rid rid);
int put_db_room(bb_rid rid);
int add_db_room(char *roomname);
int remove_db_room(bb_rid rid);

void init_cache();
void flush_cache();
void dump_cache();
void check_cache();

/* used during startup */
void compress_database();

/* this is not for general use! */
void database_changelast(bb_gid lgid, bb_rid lrid, bb_uid luid);
void Bleh(const char *fmt, ...);

/* Internal */
int pad_up_to_25_pct(int num);
#endif _BB_SERVER_DATABASE_H
