#include "../lib/rcsid.h"

NAMED_RCS_ID(ForwardMain, "$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/server/forward_main.c,v 1.7 1999/08/07 00:26:25 kc Exp $")

#ifndef FORWARD_PORT
#define FORWARD_PORT (Port_number)
#endif FORWARD_PORT

int main(int argc, char *argv[], char *envp[])
{
    struct sockaddr_in server, from;
    char hostname[64];
    int Port_number = PORT_NUMBER, port_number, bound = FALSE;

#ifdef Status
    char status[LINE_SIZE], *status_line = argv[0];
    int status_size = strlen(status_line);

#endif Status

    int dont_fork = FALSE, leave_segv = FALSE;
    int ch;
    extern char *optarg;
    char message_base[LINE_SIZE] = MESSAGEBASE;
    boolean readonly = FALSE;	/* open messagebase read-only */
    struct linger linger_options;

 /*
  * Read in flags 
  */

    Argv = argv;

    while ((ch = getopt(argc, argv, "D:EO:FM:RSdo:p:")) != EOF)
	switch (ch) {
	case 'D':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid uid number %s.\n",
			    optarg);
		else
		    Daemon_uid = tmp;
		break;
	    }
	case 'E':
	    Disable_expire = TRUE;
	    break;
	case 'F':
	    dont_fork = TRUE;
	    break;
	case 'M':
	    strcpy(message_base, optarg);
	    break;
	case 'O':
	case 'o':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid uid number %s.\n",
			    optarg);
		else
		    Owner_uid = tmp;
		break;
	    }
	case 'R':
	    readonly = TRUE;
	    break;
	case 'S':
	    leave_segv = TRUE;
	    break;
#ifdef DEBUG
	case 'd':
	    Debug = (1 << 7);
	    break;
#endif DEBUG
	case 'p':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid port number %s.\n",
			    optarg);
		else
		    Port_number = tmp;
		break;
	    }
	}

 /* First, make sure there isn't already a server running. */

    gethostname(hostname, sizeof(hostname));
    if (server_open(hostname, Port_number)) {
	fprintf(stderr, "There's already a server running.\n");
	return 1;
    }
 /* Next, make sure we can actually access the message_base directory. */

    setgid(getegid());
    setuid(geteuid());

    switch (getuid()) {
    case 0:
    case USER_UID:
	break;
    default:
	fprintf(stderr, "Sorry.\n");
	return 1;
    }

 /* Fork off a separate process, and write its process id to stdout. */
    if (!dont_fork) {
	int pid;

	switch (pid = fork()) {
	case -1:
	    perror("fork");
	    return 1;
	case 0:
	    break;
	default:
	    printf("%d\n", pid);
	    return 0;
	}
    }
    if (!readonly) {
	FILE *pidfile = fopen(PIDFILE, "w");

	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
    }
#ifdef Status
    Status("Initializing (bb)");
#endif Status

 /*
  * Create a pristine environment by closing standard files and detaching
  * from the terminal.  Of course, if we're debugging, leave them alone. 
  */

#ifdef DEBUG
    if (!Debug)
#endif DEBUG
    {
	int i;

    /* Close stdin, stdout, and stderr. */
	for (i = 0; i < 3; ++i)
	    close(i);

    /* Detach from the terminal. */
	i = open("/dev/tty", O_RDWR);
	if (i >= 0) {
#ifdef TIOCNOTTY
	    ioctl(i, TIOCNOTTY, 0);
#endif
	    close(i);
	}
#ifdef LOG_LOCAL2
	openlog("bb", LOG_PID, LOG_LOCAL2);
#else
	openlog("bb", LOG_PID);
#endif
    }

 /* Ignore broken pipes (they'll be detected later). */
#ifdef DMY_NODEBUG

    signal(SIGPIPE, SIG_IGN);

 /* Reboot on hangups */

    signal(SIGHUP, server_reboot);

 /* Trap segmentation violations */
    if (!leave_segv)
	signal(SIGSEGV, server_segv);
#endif

 /* Create socket to listen on. */

    master_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (master_socket < 0) {
	syslog(LOG_ERR, "opening master socket: %m");
	return 1;
    }
 /*
  * Automatically close on exec() 
  */
    fcntl(master_socket, F_SETFD, 1);

 /*
  * set REUSEADDR and DONTLINGER socket options.  DONTLINGER isn't
  * implemented in some environments (notably ULTRIX). 
  */

#ifndef AIX
    if (setsockopt(master_socket, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0)
	== -1) {
	syslog(LOG_ERR, "server REUSEADDR: %m");
	errno = 0;
    }
#endif
#ifdef SO_LINGER
    linger_options.l_onoff = 0;
    linger_options.l_linger = 0;
    if (setsockopt(master_socket, SOL_SOCKET, SO_LINGER,
		   (char *)&linger_options, sizeof(linger_options)) == -1) {
	syslog(LOG_ERR, "server DONTLINGER: %m");
	errno = 0;
    }
#endif SO_LINGER

 /*
  * Bind the socket to port Port_number in the INET (internet) family, and
  * listen for any incoming connections. 
  */

    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    for (port_number = Port_number;
	 port_number < Port_number + 3 && !bound;
	 port_number++) {
	server.sin_port = htons(port_number);
	if (!(bound = bind(master_socket, (struct sockaddr *) & server,
			   sizeof(server)) == 0)) {
#ifdef PTX
	    if (errno == EADDRNOTAVAIL)
		errno = EADDRINUSE;
#endif PTX
	    if (errno != EADDRINUSE) {
		syslog(LOG_ERR, "binding %d: %m", port_number);
		return 1;
	    }
	    errno = 0;
	}
    }

    if (!bound)
	return 1;
    port_number--;
    syslog(LOG_INFO, "bound to port %d", port_number - 1);
#ifdef DEBUG
    if (Debug)
	printf("Server bound to port %d.\n", port_number);
#endif DEBUG

 /* Initialize width to the size of our descriptor table. */

    syslog(LOG_INFO, "forwarding connections to %s:%d",
	   FORWARD_SERVER, FORWARD_PORT);

 /* Trap signals */
#ifdef DMY_NO_DEBUG
    signal(SIGHUP, hangup);
    signal(SIGTERM, terminate);
#endif

 /* Start accepting connections. */
    listen(master_socket, 5);

    while (!Server_Shutdown) {
	int len = sizeof(from);
	int socket;

#ifdef Status
	sprintf(status, "(bb) Forwarding connections to %s:%d",
		FORWARD_SERVER, FORWARD_PORT);
	Status(status);
#endif Status


	socket =
	  accept(master_socket, (struct sockaddr *) & from, &len);
	if (socket == -1)
	    syslog(LOG_DEBUG, "accept: %m");
	else {
	/*
	 * Let data linger for 60 seconds after a close().
	 */
#ifdef SO_LINGER
		linger_options.l_onoff = 1;
		linger_options.l_linger = 60;
		if (setsockopt(socket,
			       SOL_SOCKET, SO_LINGER, (char *)&linger_options,
			       sizeof(linger_options)) == -1)
		    syslog(LOG_WARNING, "client LINGER: %m");
#endif SO_LINGER

	/*
	 * Automatically close on exec() 
	 */
	    fcntl(socket, F_SETFD, 1);

	/*
	 * Forward the connection. 
	 */

	    write_socket_code(socket, SERVER_FORWARD);
#ifdef PTX
	/* The first one gets ignored in PTX.  WHY!? ARGH! */
	    write_socket_code(socket, SERVER_FORWARD);
#endif PTX
	    write_socket_string(socket, FORWARD_SERVER);
	    write_socket_int(socket, FORWARD_PORT);
	    close(socket);

	}
    }


    if (Server_Reboot) {
	server_reboot();
	return 1;
    } else {
	syslog(LOG_INFO, "shutting down");
	return 0;
    }
}
