#include "../lib/rcsid.h"

RCS_ID("$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/server/lib.c,v 1.6 1999/03/14 23:14:36 kc Exp $")

#include "server.h"
#include <errno.h>

char error_text[LINE_SIZE];
boolean Disable_expire = FALSE;

static int db_room_new_messages(DbRoom *db_room, int last_message_read);
static boolean get_directory_info(directory_info *dir_info, int i, bb_rid rid);
static int archive(bb_rid rid, int message);

void update_dir_info(int i, bb_rid rid)
{
    if(!Client[i].dir_info[rid]) {
      Client[i].dir_info[rid] = (directory_info *)
	malloc(sizeof(directory_info));

      /* get rid of UMR --dmy, 11/95 */
      bzero((char *) Client[i].dir_info[rid], sizeof(directory_info));
    }
    
    if (!get_directory_info(Client[i].dir_info[rid], i, rid))
	*Client[i].dir_info[rid]->name = '\0';
}

static int db_room_new_messages(DbRoom *db_room, int last_message_read)
{
  int messages, sum_index, new;
  DbSummary *entry;

  messages = db_room->directory ? db_room->directory->num_msgs : -1;
  if (messages > 0) {
    for (sum_index = messages-1, new=0;
	 sum_index >= 0;
	 sum_index--) {
      entry = db_room->summary[sum_index];
      if (entry->number <= last_message_read) return new;
      if (!entry->header.deleted) new++;
    }
  } else
    return 0;
  return new;
}


void update_room_pointer(int i, bb_rid rid, int last_message_read)
{
    if(rid == Client[i].rid)
	Client[i].last_message_read = last_message_read;
    Client[i].update_directory = 0;
    Client[i].stamp[rid].update++;
    if (Client[i].dir_info[rid]) {
      DbRoom *db_room;

      if (!(db_room = get_db_room(rid))) return;
      Client[i].dir_info[rid]->new_messages =
	db_room_new_messages(db_room, last_message_read);

    } else {
      update_dir_info(i, rid);
    }
}

boolean save_room_pointer(int i)
{

  DbRoomUser db_room_user;
  bb_rid rid = Client[i].rid;
  
  if (rid == -1)
    return TRUE;
  if (!get_db_room_user(&db_room_user, rid, Client[i].uid)) {
    db_room_user.user_offset = DEFAULT_USER_OFFSET;
    db_room_user.class = no_class;
  }
  
  db_room_user.last_message_read = Client[i].last_message_read;
  if (!put_db_room_user(rid, Client[i].uid, &db_room_user)) {
    ERROR("error storing room pointers");
    return FALSE;
  }
  return TRUE;
}

directory_info *get_dir_info(int i, bb_rid rid)
{
    if(!Client[i].dir_info) {
	static directory_info dir_info;

	return get_directory_info(&dir_info, i, rid) ? &dir_info : NULL;
    }
    if (!Client[i].dir_info[rid])
	update_dir_info(i, rid);
    return *Client[i].dir_info[rid]->name ? Client[i].dir_info[rid] : NULL;
}

static boolean get_directory_info(directory_info *dir_info, int i, bb_rid rid)
{
    DbDirectory *db_directory;
    DbRoomUser db_room_user;
    DbRoom *db_room;
    int last_message_read;

    if (!(db_directory = get_db_room_directory(rid))) {
	ERROR("error accessing room information");
	return FALSE;
    }
    bcopy(db_directory->name, dir_info->name, sizeof(dir_info->name));
    bcopy(db_directory->summary_line, dir_info->summary,
	  sizeof(dir_info->summary));
    strcpy(dir_info->owner, get_id(db_directory->owner_uid));
    if (!(db_room = get_db_room(rid))) {
	ERROR("reading summary");
	return FALSE;
    }
    dir_info->messagenumber = db_directory->num_msgs > 0 ?
      db_room->summary[db_directory->num_msgs - 1]->number : 0;
    dir_info->default_expiration = db_directory->default_expiration;
    dir_info->default_class = db_directory->default_class;
    if (!get_db_room_user(&db_room_user, rid, Client[i].uid)) {
	last_message_read = 0;
	dir_info->priority = DEFAULT_USER_OFFSET;
	dir_info->class = no_class;
    } else {
	int j;

	last_message_read = rid == Client[i].rid ?
	  Client[i].last_message_read : db_room_user.last_message_read;
	for (j = 0;
	     j < db_directory->num_msgs &&
	     db_room->summary[j]->number <= db_room_user.last_message_read;
	     j++);
	dir_info->new_messages = db_directory->num_msgs - j;
	dir_info->priority = db_room_user.user_offset;
	dir_info->class = db_room_user.class;
    }

    dir_info->new_messages = db_room_new_messages(db_room, last_message_read);

    if (dir_info->class == no_class &&
	!DAEMON(Client[i].uid) &&
	!NO_ACCESS(group_class(i, db_directory->groups)))
	dir_info->class = db_directory->default_class;
    if (dir_info->priority == DEFAULT_USER_OFFSET)
	dir_info->priority = db_directory->room_offset;
    dir_info->priority += delta_priority(i, db_directory->groups);
    if (dir_info->new_messages < 0)
	dir_info->new_messages = 0;
    if (Client[i].uid == db_directory->owner_uid)
	dir_info->class = moderator;
    bcopy((char *)&db_directory->groups,
	  (char *)&dir_info->groups,
	  sizeof(dir_info->groups));
    dir_info->archive = db_directory->archive;
    return TRUE;
}

boolean update_cached_user_group_info(bb_uid uid, bb_gid gid)
{
    int slot;
    DbUser db_user;
    DbGroup db_group;

    if (!get_db_user(&db_user, uid)) {
	ERROR("error accessing user information");
	return FALSE;
    }
    if (!get_db_group(&db_group, gid)) {
	ERROR("error accessing group information");
	return FALSE;
    }
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on && Client[slot].uid == uid) {
	    bb_rid rid;

	    Client[slot].group_priority[gid] =
	      db_user.group_priority[gid] + db_group.group_offset;
	    Client[slot].group_class[gid] = db_user.group_class[gid];

	    for (rid = 0; rid <= Last_rid; rid++) {
		if (Client[slot].dir_info[rid] &&
		    GROUP_ISSET(gid, &Client[slot].dir_info[rid]->groups)) {
		    free(Client[slot].dir_info[rid]);
		    Client[slot].dir_info[rid] = NULL;
		    Client[slot].stamp[rid].update++;
		}
	    }
	    Client[slot].update_directory = 0;
	}
    return TRUE;
}

/* group information for a particular client */
group_info *get_group_info(int i, bb_gid gid)
{
    DbGroup db_group;
    static group_info groupinfo;

    if (!get_db_group(&db_group, gid)) {
	ERROR("error accessing group information");
	return NULL;
    }
    groupinfo.gid = gid;
    bcopy(db_group.name, groupinfo.name, sizeof(groupinfo.name));
    strcpy(groupinfo.owner, get_id(db_group.owner));
    groupinfo.default_class = db_group.default_class;
    groupinfo.max_expiration = db_group.max_expiration;
    groupinfo.priority = Client[i].group_priority[gid];
    groupinfo.class = Client[i].group_class[gid];
    if (groupinfo.class == no_class)
	groupinfo.class = db_group.default_class;
    if (Client[i].uid == db_group.owner)
	groupinfo.class = moderator;
    return &groupinfo;
}

bb_priority delta_priority(int i, group_set groups)
{
    bb_gid gid;
    bb_priority priority = SYSTEM_CUTOFF;

    for (gid = 0; gid <= Last_gid; gid++)
	if (GROUP_ISSET(gid, &groups) &&
	    Client[i].group_priority[gid] > priority)
	    priority = Client[i].group_priority[gid];
    return priority;
}

bb_priority group_class(int i, group_set groups)
{
    bb_gid gid;
    class_t class = no_access, group_class;
    DbGroup db_group;

    for (gid = 0; gid <= Last_gid; gid++) {
	if (GROUP_ISSET(gid, &groups)) {
	    group_class = Client[i].group_class[gid];
	    if (group_class == no_class && !DAEMON(Client[i].uid))
		group_class = get_db_group(&db_group, gid) ?
		  db_group.default_class : no_access;
	    if (group_class > class)
		class = group_class;
	}
    }
    return class;
}

time_t expiration_limit(group_set groups)
{
    bb_gid gid;
    time_t max_expiration = 0;
    DbGroup db_group;

    for (gid = 0; gid <= Last_gid; gid++) {
	if (GROUP_ISSET(gid, &groups))
	    if (get_db_group(&db_group, gid) &&
		db_group.max_expiration > max_expiration)
		max_expiration = db_group.max_expiration;
    }
    return max_expiration;
}

int remove_room(bb_rid rid)
{
    DbDirectory *db_directory;
    DbRoom *db_room;
    char filename[BUFFER_SIZE];
    bb_uid uid;
    int *message_number, message_count, summary_count;

    if (Read_only) {
	ERROR("Messagebase is readonly.");
	return 0;
    }
    if (!(db_directory = get_db_room_directory(rid))) {
	ERROR("error accessing room information");
	return 0;
    }
    if (!(db_room = get_db_room(rid))) {
	ERROR("error accessing room summary");
	return 0;
    }

    /* expire all remaining messages */
    summary_count = db_directory->num_msgs;
    message_number = (int *) malloc(summary_count * sizeof(int));
    for (message_count=0; message_count<summary_count; message_count++)
      message_number[message_count] = db_room->summary[message_count]->number;
    for (message_count=0; message_count<summary_count; message_count++)
      expire_message(rid, message_number[message_count]);
    free(message_number);

    for (uid = 0; uid <= Last_uid; uid++)
	remove_db_room_user(rid, uid);

    if (db_directory->archive == common_archive) {
	FILE *room_log, *room_list;
	char *classes[] =
	{"no_access", "read_only", "read_write", "moderator", "no_class"};

	/* record expiration in expired/Rooms file. */
	room_list = fopen(EXPIRED_ROOMS, "a");
	if (!room_list) {
	    syslog(LOG_ERR, "%s: %m", EXPIRED_ROOMS);
	} else {
	    fprintf(room_list, "%d %s\n",
		    (int)db_directory->arid, db_directory->name);
	    fclose(room_list);
	}
	/* write Log file in expired room's directory */
	sprintf(filename, EXPIRED_ROOM, (int)db_directory->arid);
	strcat(filename, "/Log");
	room_log = fopen(filename, "a");
	if (!room_log) {
	    syslog(LOG_ERR, "%s: %m", filename);
	} else {
	  fprintf(room_log, "Room ID:\t%d\nRoom Name:\t%s\n",
		  (int)db_directory->arid, db_directory->name);
	  fprintf(room_log, "Owner:\t\t%d (%s)\n",
		  (int)db_directory->owner_uid, get_id(db_directory->owner_uid));
	  fprintf(room_log, "Created:\t%sDefault Class:\t%s\n",
		  ctime(&db_directory->creation_time),
		  classes[db_directory->default_class]);
	  fprintf(room_log, "Expiration:\t%d (%d day%s)\nSummary:\t%s\n",
		  (int)db_directory->default_expiration,
		  (int)db_directory->default_expiration / 86400,
		  (int)db_directory->default_expiration / 86400 == 1 ? "" : "s",
		  db_directory->summary_line);
	  fprintf(room_log, "Groups:\t\t0x%04x\nPriority:\t%d\n",
		  (unsigned int)*db_directory->groups.group_bits,
		  (int)db_directory->room_offset);
	  fclose(room_log);
	}
    }

    /* room's directory should now be empty.  try to remove it -- dmy */
    sprintf(filename, ACTIVE_ROOM, (int)db_directory->arid);
    rmdir(filename);

    if (!remove_rid(db_directory->name)) {
	ERROR("error removing room identification");
	return 0;
    }
    if (!remove_db_room(rid)) {
	ERROR("error removing room information");
	return 0;
    }

#ifdef does_not_work
    compress_database();
#endif
    return 1;
}

int expire_message(bb_rid rid, int message)
{
    char filename[BUFFER_SIZE];

    if (Disable_expire) {
	ERROR("Expiration is disabled.");
	return 0;
    }
    if (Read_only) {
	ERROR("Messagebase is readonly.");
	return 0;
    }
    if (!archive(rid, message))
      return 0;

    remove_db_message(rid, message);
    sprintf(filename, ACTIVE_MESSAGE, get_db_room_arid(rid), message);
    if (unlink(filename) == -1) {
      syslog(LOG_ERR, "error deleting %s: %m\n", filename);
      return 0;
    }
    return 1;
}

static int archive(bb_rid rid, int message)
{
    DbDirectory *db_directory;
    DbMessage db_message;

    if (!db_get_db_message(&db_message, rid, message))
	return 0;
    if(db_message.deleted) return 1;
    if (!(db_directory = get_db_room_directory(rid))) {
	ERROR("error accessing room information");
	return 0;
    }
    switch (db_directory->archive) {
    case no_archive:
	break;
    case mail_archive:
	{
	    char subject[BUFFER_SIZE];

	    sprintf(subject, "archive of %s (%d): #%d",
		    db_directory->name, (int)rid, message);
	    if (!mail_message(rid, message, Owner_uid, db_directory->owner_uid,
			      subject)) {
		syslog(LOG_ERR, "error archiving %d/%d to mail", rid, message);
		return 0;
	    }
	    break;
	}
    case common_archive:
    default:
	{
	    char filename[BUFFER_SIZE], newname[BUFFER_SIZE];
	    FILE *active_msg, *expired_msg;
	    int c, arid;
	    

	    arid = get_db_room_arid(rid);

	    sprintf(filename, ACTIVE_MESSAGE, arid, message);
	    sprintf(newname, EXPIRED_MESSAGE, arid, message);
	    expired_msg = fopen(newname, "w");
	    if (!expired_msg && errno == ENOENT) {
		char dirname[BUFFER_SIZE];

		sprintf(dirname, EXPIRED_ROOM, arid);
		mkdir(dirname, 0777);
		expired_msg = fopen(newname, "w");
	    }
	    if (!expired_msg) {
		syslog(LOG_ERR, "error writing %s: %m\n", newname);
		return 0;
	    }
	    fprintf(expired_msg,
		    "Date: %sFrom: %s<%s> (%d)\nSubject: %s\nLines: %d\n\n",
		    ctime(&db_message.date),
		    db_message.poster_nick,
		    db_message.poster_id,
		    (int)db_message.poster_uid,
		    db_message.subject,
		    db_message.lines);
	    if (!(active_msg = fopen(filename, "r"))) {
		if (errno != ENOENT) {
		    syslog(LOG_ERR, "error reading %s: %m\n", filename);
		    return 0;
		}
	    } else {
		for (c = getc(active_msg); c != EOF; c = getc(active_msg))
		    putc((char)c, expired_msg);
		fclose(active_msg);
	    }
	    fclose(expired_msg);
	    break;
	}
    }
    return 1;
}

char *sh_lit(char *string)
{
    static char lit[BUFFER_SIZE];
    char *p, *q;

    q = lit;
    *q++ = '\'';
    for (p = string; *p; p++, q++) {
	switch (*p) {
	case '\'':
	    *q++ = '\'';
	    *q++ = '\\';
	    *q++ = *p;
	    *q = '\'';
	    break;
	case '\\':
	    *q++ = *p;
	default:
	    *q = *p;
	}
    }
    *q++ = '\'';
    *q = '\0';
    return lit;
}

FILE *mail(bb_uid from_uid, bb_uid to_uid, char *subject)
{
    char command[BUFFER_SIZE];
    FILE *message = NULL;
    char from_mail_address[BUFFER_SIZE], to_mail_address[BUFFER_SIZE];
    DbUser db_user;

    if (!get_db_user(&db_user, from_uid) &&
	!get_db_user(&db_user, Owner_uid)) {
	ERROR("error accessing sender user information");
	return NULL;
    }
    strcpy(from_mail_address, *db_user.mail_address ?
	   db_user.mail_address : db_user.last_access_source);
    if (!get_db_user(&db_user, to_uid)) {
	ERROR("error accessing recipient user information");
	return NULL;
    }
    strcpy(to_mail_address, *db_user.mail_address ?
	   db_user.mail_address : db_user.last_access_source);
    sprintf(command, "/usr/lib/sendmail %s", sh_lit(to_mail_address));
    if (!(message = popen(command, "w"))) {
	syslog(LOG_ERR, "popen(\"%s\", \"w\")", command);
	ERROR(sys_errlist[errno]);
	return NULL;
    }
    fprintf(message, "From: <%s>\nTo: <%s>\nSubject: %s\n\n",
	    from_mail_address, to_mail_address, subject);
    return message;
}

void server_shutdown(boolean reboot, boolean immediate)
{
    extern int master_socket;
    int slot;

    syslog(LOG_INFO, "server %s %s",
	   reboot ? "rebooting" : "shutting down",
	   immediate ? "immediately" : "eventually");
    Server_Shutdown = TRUE;
    Server_Reboot = reboot;

    if (immediate) {
    /* Stop accepting connections */
	close(master_socket);
    /* Close all clients. */
	for (slot = 0; slot <= Max_active_slot; slot++) {
	    if (Client[slot].signed_on)
		save_room_pointer(slot);
	    if (Client[slot].inuse)
		close_client(slot);
	}
    }
}

signal_t terminate()
{
    signal(SIGTERM, terminate);
    server_shutdown(FALSE, TRUE);
    Bleh("SERVER TERMINATE");
    return (signal_t)0;
}

signal_t interrupt()
{
    extern int Current_client;

    signal(SIGINT, interrupt);
    if (Client[Current_client].inuse) {
	syslog(LOG_INFO, "kicking <%s> (%d)",
	       Client[Current_client].id,
	       Client[Current_client].uid);
	if (Client[Current_client].signed_on)
	    save_room_pointer(Current_client);
	close_client(Current_client);
    }
    return (signal_t)0;
}

signal_t hangup()
{
    signal(SIGHUP, hangup);
    server_shutdown(TRUE, TRUE);
    Bleh("SERVER HANGUP");
    return (signal_t)0;
}

extern char **Argv;
char first_arg[] = "bb-server                                                                       ";

signal_t server_segv()
{
  *Argv = first_arg;
  syslog(LOG_INFO, "rebooting: segmentation violation - %m");
  execv(SERVER, Argv);
  syslog(LOG_INFO, "reboot failed - %s: %m", SERVER);
  Bleh("SEG VIO");
  return (signal_t)0;
}

signal_t server_reboot()
{
            *Argv = first_arg;
    syslog(LOG_INFO, "rebooting");
    execv(SERVER, Argv);
    syslog(LOG_INFO, "reboot failed - %s: %m", SERVER);
    return (signal_t)0;
}

void close_client(int slot)
{
    bb_rid rid;

 /* Close a client connection, marking its slot as available. */

    shutdown(Client[slot].socket, 2);
    close(Client[slot].socket);
    if (Client[slot].signed_on) {
	DbUser db_user;

	if (get_db_user(&db_user, Client[slot].uid)) {
	    db_user.connect_time += time(0) - Client[slot].signon_time;
	    db_user.posts += Client[slot].posts;
	    put_db_user(Client[slot].uid, &db_user);
	}
    }
    Client[slot].inuse = FALSE;
    Client[slot].signed_on = FALSE;
    Client[slot].uid = -1;
    Client[slot].major_version = 0;
    Client[slot].minor_version = 0;
    Client[slot].socket = 0;
    Client[slot].errors = 0;
    if (Client[slot].dir_info) {
	for (rid = 0; rid <= Last_rid; rid++) {
	    if (Client[slot].dir_info[rid]) {
		free(Client[slot].dir_info[rid]);
		Client[slot].dir_info[rid] = NULL;
	    }
	}
	free(Client[slot].dir_info);
	Client[slot].dir_info = NULL;
    }
    if (Client[slot].stamp) {
      free(Client[slot].stamp);
      Client[slot].stamp = NULL;
    }
    if (slot < Free_slot)
	Free_slot = slot;
    if (slot == Max_active_slot) {
	for (; Max_active_slot >= 0; Max_active_slot--)
	    if (Client[Max_active_slot].inuse)
		break;
	if (Server_Shutdown)
	    syslog(LOG_DEBUG,
		   "Shutdown requested -- max active slot: %d -> %d",
		   slot, Max_active_slot);
    }
}

int remove_user(bb_uid uid)
{
    DbUser db_user;
    bb_rid rid;

    if (!get_db_user(&db_user, uid)) {
	ERROR("error accessing user information");
	return 0;
    }
    for (rid = 0; rid <= Last_rid; rid++)
	remove_db_room_user(rid, uid);
    if (!remove_db_user(uid)) {
	syslog(LOG_ERR, "error removing user entry for %d <%s>",
	       uid, db_user.id);
	ERROR("removing user");
	return 0;
    }
    
    if (!remove_uid(db_user.id)) {
	syslog(LOG_ERR, "error subtracting uid %d <%s>", uid, db_user.id);
	ERROR("removing uid");
	return 0;
    }

    /* recycle */
#ifdef cannot_get_it_to_work
    compress_database();
#endif

    return 1;
}

int mail_message(bb_rid rid, int message,
		     bb_uid from_uid, bb_uid to_uid, char *subject)
{
    char filename[BUFFER_SIZE];
    DbMessage db_message;
    FILE *active_msg, *mail_message;
    int c;

    if (!db_get_db_message(&db_message, rid, message))
	return 0;
    sprintf(filename, ACTIVE_MESSAGE, get_db_room_arid(rid), message);
    mail_message = mail(from_uid, to_uid, subject);
    if (!mail_message)
	return 0;
    fprintf(mail_message,
	    "Date: %sFrom: %s<%s> (%d)\nSubject: %s\nLines: %d\n\n",
	    ctime(&db_message.date),
	    db_message.poster_nick,
	    db_message.poster_id,
	    (int)db_message.poster_uid,
	    db_message.subject,
	    db_message.lines);
    if (!(active_msg = fopen(filename, "r"))) {
	syslog(LOG_ERR, "error reading %s: %m\n", filename);
    } else {
	for (c = getc(active_msg); c != EOF; c = getc(active_msg))
	    putc((char)c, mail_message);
	fclose(active_msg);
    }
    pclose(mail_message);
    return active_msg != NULL;
}

static int summary_order(const void *p1, const void *p2)
{
    const DbSummary *se1 = *(const void **)p1, *se2 = *(const void **)p2;

    return se1->number - se2->number;
}

static void sort_summary(bb_rid rid)
{
  DbRoom *db_room;

    if (!(db_room = get_db_room(rid)))
	return;
    qsort(db_room->summary, db_room->directory->num_msgs,
	  sizeof(*db_room->summary), summary_order);
}

void expire_messages(bb_rid rid)
{
  boolean db_directory_changed = FALSE;
  DbDirectory *db_directory;
  DbRoom *db_room;
  DbSummary *entry;
  DbMessage db_message;
  int *message_number; /* array of message numbers prior to expirations */
  int message, last_message, summary_count;
  int sum_index;
  
  if (Disable_expire)	return;
  if (!(db_directory = get_db_room_directory(rid))) return;
  if (!(db_room = get_db_room(rid))) return;
  if (db_directory->num_msgs < 1) return;
  
  /* read message numbers first because directory changes. */
  summary_count = db_directory->num_msgs;
  message_number = (int *) malloc(summary_count * sizeof(int));
  for (sum_index=0; sum_index < summary_count; sum_index++) {
    entry = db_room->summary[sum_index];
    message_number[sum_index] = entry ? entry->number : -1;
  }
  for (sum_index=0, last_message = -1;
       sum_index < summary_count;
       sum_index++) {
    message = message_number[sum_index];
    if (message > 0) {
      if (message < last_message) {
	syslog(LOG_ERR,
	       "corrupted summary in %d (%s): #%d followed #%d",
	       rid, get_roomname(rid), message, last_message);
	sort_summary(rid);
	expire_messages(rid);
	free(message_number);
	return;
      }
      if (message == last_message) {
	syslog(LOG_ERR, "corrupted summary in %d (%s): duplicate #%d",
	       rid, get_roomname(rid), last_message);
	remove_db_message(rid, last_message);
      }
      last_message = message;
      if (db_get_db_message(&db_message, rid, message) &&
	  db_message.expiration < Now &&
	  expire_message(rid, message)) {
	/* make sure out-of-order messages don't screw us up. */
	if (message == db_directory->first_message_number) {
	  db_directory->first_message_number++;
	  db_directory_changed = TRUE;
	}
      }
    }
  }
  
  if (db_directory_changed)
    put_db_room(rid);
  free(message_number);
}

void expire_room(bb_rid rid)
{
    DbDirectory *db_directory;
    DbRoom *db_room;
    int slot, message;
    bb_gid gid;
    DbGroup db_group;

    if (!(db_room = get_db_room(rid)) ||
	!(db_directory = get_db_room_directory(rid)))
      return;
    
    for (message = 0; message < db_directory->num_msgs; message++)
      if (!db_room->summary[message]->header.deleted)
	return;

 /*
  * Old_Room has no (undeleted) messages left.  Make a few saving throws; if
  * they fail, remove the room. 
  */

 /* someone is still in the room */
    for (slot = 0; slot <= Max_active_slot; slot++)
	if (Client[slot].signed_on && Client[slot].rid == rid)
	    return;

 /* the room is a member of a supported group */
    for (gid = 0; gid <= Last_gid; gid++)
	if (GROUP_ISSET(gid, &db_directory->groups) &&
	    get_db_group(&db_group, gid) &&
	    db_group.group_offset >= 0)
	    return;

 /* the room is restricted:  send the owner mail */
    if (db_directory->default_class == no_access) {
	FILE *message_f;

	if ((message_f = mail(Owner_uid, db_directory->owner_uid,
			      "empty room on bb"))) {
	    fprintf(message_f, "The %s room is empty.  If you are no longer using this room,\nplease remove it.\n",
		    db_directory->name);
	    pclose(message_f);
	}
	return;
    }
    remove_room(rid);
}

/* 9/95, dmy.  (25-28) -> 28; (29-32) -> 32; (33-40) -> 40 */
int pad_up_to_25_pct(int num)
{
  int pwr, pad_num;

  pad_num = num - 1;
  for (pwr=0; pad_num>7; pwr++)
    pad_num /= 2;
  pad_num++;
  for (; pwr>0; pwr--)
    pad_num *= 2;

  return pad_num;
}

