#include <sys/types.h>
#include <stdio.h>
#include <strings.h>
#include <signal.h>
#include <syslog.h>

#include "../lib/libc.h"
#include "../lib/client.h"
#include "../lib/transport.h"
#include "../lib/protocol.h"
#include "../lib/macros.h"
#include "../lib/misc.h"

#include "config.h"
#include "database.h"

#define BUFFER_SIZE 512

#ifndef MESSAGEBASE
#define MESSAGEBASE "messages"
#endif

#define NO_ACCESS(class) ((class) == no_access || (class) == no_class)
#define READ_ACCESS(class) (!NO_ACCESS(class))
#define WRITE_ACCESS(class) ((class) == read_write || (class) == moderator)
#define MODERATOR_ACCESS(class) ((class) == moderator)

extern errno;
#if defined(NeXT) || defined(__APPLE__)
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
/* For inet_ntoa() */
#include <netinet/in.h>
#include <arpa/inet.h>
/* For shutdown() */
#include <sys/socket.h>
/* sys_errlist is already declared in the system headers */
#else
extern char *sys_errlist[];
#endif

#define read_socket_code(socket, code) read_socket_byte(socket, code)
#define write_socket_code(socket, code) write_socket_byte(socket, code)

extern char error_text[];

#define ERROR(s) strcpy(error_text, s)
#define SERVER_ERROR(s) \
  { \
    write_socket_code(Client[i].socket, SERVER_ERR); \
    write_socket_string(Client[i].socket, s); \
    return; \
  }

struct client_stamp {
    long update;
    long sent;
} *stamp;

extern struct client_struct {
    int socket;
    int errors;
    boolean inuse;
    boolean signed_on;
    boolean owner;
    long update_directory;
    long update_summary;
    long update_message_numbers;
    char nick[LINE_SIZE];
    bb_uid uid;
    char id[ID_SIZE];
    char address[LINE_SIZE];
    char user[32];
    char host[64];
    bb_rid rid;
    int posts;
    time_t last_action;
    time_t signon_time;
    bb_priority group_priority[MAX_GROUPS];
    class_t group_class[MAX_GROUPS];
    bb_priority cutoff_priority;
    directory_info **dir_info;
    long last_message_read;
    int major_version;
    int minor_version;
    struct client_stamp *stamp;
} Client[MAX_CLIENTS];

extern int Free_slot, Max_active_slot;
extern boolean Server_Shutdown, Server_Reboot;

/* The following routines can be found in lib.c */

directory_info *get_dir_info(int i, bb_rid rid);
void update_dir_info(int i, bb_rid rid);
void update_room_pointer(int i, bb_rid rid, int last_message_read);
boolean save_room_pointer(int i);
void update_cached_room_user_info(bb_rid rid, bb_uid uid);
boolean update_cached_group_info(bb_gid gid);
boolean update_cached_user_group_info(bb_gid gid, bb_uid uid);
group_info *get_group_info(int i, bb_gid gid);
bb_priority delta_priority(int i, group_set groups);
int remove_room(bb_rid rid);
void expire_room(bb_rid rid);
void expire_messages(bb_rid rid);
int expire_message(bb_rid rid, int message);
FILE *mail(bb_uid from_uid, bb_uid to_uid, char *subject);
void server_shutdown(boolean reboot, boolean immediate);
signal_t terminate();
signal_t interrupt();
signal_t hangup();
signal_t server_reboot();
signal_t server_segv();
void close_client(int slot);
int remove_user(bb_uid uid);
int mail_message(bb_rid rid, int message,
		     bb_uid from_uid, bb_uid to_uid, char *subject);
bb_priority group_class(int i, group_set groups);
time_t expiration_limit(group_set groups);

#define get_class(i, rid) (get_dir_info((i), (rid))->class)

/* These are the routines which implement each command. */

void server_catchup(int i);
void server_changeclass(int i);
void server_changeroom(int i);
void server_create(int i);
void server_cutoff(int i);
void server_db_message(int i);
void server_delete(int i);
void server_directory(int i);
void server_groupadd(int i);
void server_groupchange(int i);
void server_groupchangeclass(int i);
void server_groupcreate(int i);
void server_groupdirectory(int i);
void server_grouplist(int i);
void server_grouppriority(int i);
void server_groupremove(int i);
void server_groupsubtract(int i);
void server_kill(int i);
void server_list(int i);
void server_message_info(int i);
void server_message_numbers(int i);
void server_nick(int i);
void server_passwd(int i);
void server_post(int i);
void server_preserve(int i);
void server_priority(int i);
void server_protocol(int i);
void server_read(int i);
void server_remove(int i);
void server_rename(int i);
void server_setcurrent(int i);
void server_roominfo(int i);
void server_search(int i);
void server_select(int i);
void server_setmail(int i);
void server_setname(int i);
void server_shut(int i);
void server_signoff(int i);
void server_signon(int i);
void server_summary(int i);
void server_undelete(int i);
void server_users(int i);
void server_whois(int i);
