#include "../lib/rcsid.h"

NAMED_RCS_ID(StandardMain, "$Header: /NetworkDisk/Source/CVS/Staff/kc/bb/server/standard_main.c,v 1.13 2001/04/12 18:19:45 kc Exp $")

#ifdef STATS
struct timeval delta_time(struct timeval t1, struct timeval t2);
struct timeval add_times(struct timeval t1, struct timeval t2);
signal_t report_times();

static struct {
    int count;
    struct timeval accum;
    long max;
}      stats[256];

#endif STATS

int main(int argc, char *argv[], char *envp[])
{
    int width;
    struct sockaddr_in server, from;
    char hostname[64];
    int Port_number = PORT_NUMBER, port_number, bound = FALSE;

#ifdef Status
    char status[LINE_SIZE], *status_line = argv[0];
    int status_size = strlen(status_line);

#endif Status
    /* dont_fork == false, except for test versions --dmy */
    int dont_fork = FALSE, leave_segv = TRUE;
    int ch;
    extern char *optarg;
    char message_base[LINE_SIZE] = MESSAGEBASE;
    boolean readonly = FALSE;		/* open messagebase read-only */
    boolean override_locks = FALSE;	/* override messagebase file locking */
    struct linger linger_options;

 /*
  * Read in flags 
  */

    Argv = argv;

#ifdef NeXT
    malloc_singlethreaded();
#endif

    while ((ch = getopt(argc, argv, "dD:EFLM:O:o:p:RS")) != EOF)
	switch (ch) {
#ifdef DEBUG
	case 'd':
	    Debug = (1 << 7);
	    break;
#endif DEBUG
	case 'D':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid uid number %s.\n",
			    optarg);
		else
		    Daemon_uid = tmp;
		break;
	    }
	case 'E':
	    Disable_expire = TRUE;
	    break;
	case 'F':
	    dont_fork = TRUE;
	    break;
	case 'L':
	    override_locks = TRUE;
	    break;
	case 'M':
	    strcpy(message_base, optarg);
	    break;
	case 'O':
	case 'o':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid uid number %s.\n",
			    optarg);
		else
		    Owner_uid = tmp;
		break;
	    }
	case 'p':{
		int tmp;

		if (sscanf(optarg, "%d", &tmp) != 1)
		    fprintf(stderr, "Ignoring invalid port number %s.\n",
			    optarg);
		else
		    Port_number = tmp;
		break;
	    }
	case 'R':
	    readonly = TRUE;
	    break;
	case 'S':
	    leave_segv = !leave_segv;
	    break;
	}

 /* First, make sure there isn't already a server running. */

    gethostname(hostname, sizeof(hostname));
    if (server_open(hostname, Port_number)) {
	fprintf(stderr, "There's already a server running.\n");
	return 1;
    }
 /* Next, make sure we can actually access the message_base directory. */

    setgid(getegid());
    setuid(geteuid());

    switch (getuid()) {
    case 0:
    case USER_UID:
	break;
    default:
	fprintf(stderr, "Sorry.\n");
	return 1;
    }

    umask(007);
    if (chdir(message_base) == -1) {
	perror(message_base);
	return 1;
    }
 /* Fork off a separate process, and write its process id to stdout. */
    if (!dont_fork) {
	int pid;

	switch (pid = fork()) {
	case -1:
	    perror("fork");
	    return 1;
	case 0:
	    break;
	default:
	    printf("%d\n", pid);
	    return 0;
	}
    }
    if (!readonly) {
	FILE *pidfile = fopen(PIDFILE, "w");

	fprintf(pidfile, "%d\n", getpid());
	fclose(pidfile);
    }
#ifdef Status
    Status("Initializing (bb)");
#endif Status

 /*
  * Create a pristine environment by closing standard files and detaching
  * from the terminal.  Of course, if we're debugging, leave them alone. 
  */

#ifdef DEBUG
    if (!Debug)
#endif DEBUG
    {
	int i;

    /* Close stdin, stdout, and stderr. */
	for (i = 0; i < 3; ++i)
	    close(i);

    /* Detach from the terminal. */
	i = open("/dev/tty", O_RDWR, 0);
	if (i >= 0) {
#ifdef TIOCNOTTY
	    ioctl(i, TIOCNOTTY, 0);
#endif
	    close(i);
	}
#ifdef LOG_LOCAL2
	openlog("bb", LOG_PID, LOG_LOCAL2);
#else
	openlog("bb", LOG_PID);
#endif
    }
 /* Reboot on hangups */
    signal(SIGHUP, server_reboot);

 /* Trap segmentation violations */
    if (!leave_segv)
	signal(SIGSEGV, server_segv);

 /* Create socket to listen on. */

    master_socket = socket(AF_INET, SOCK_STREAM, 0);
    if (master_socket < 0) {
	syslog(LOG_ERR, "opening master socket: %m");
	return 1;
    }
    fcntl(master_socket, F_SETFD, 1);

 /*
  * set REUSEADDR and LINGER socket options.  DONTLINGER isn't
  * implemented in some environments (notably ULTRIX). 
  */

#ifndef AIX
    if (setsockopt(master_socket, SOL_SOCKET, SO_REUSEADDR, (char *)0, 0)
	== -1) {
#ifdef LOG_REUSEADDR_ERROR
	syslog(LOG_ERR, "server REUSEADDR: %m");
#endif
	errno = 0;
    }
#ifdef SO_REUSEPORT
    if (setsockopt(master_socket, SOL_SOCKET, SO_REUSEPORT, (char *)0, 0)
	== -1) {
#ifdef LOG_REUSEADDR_ERROR
	syslog(LOG_ERR, "server REUSEPORT: %m");
#endif
	errno = 0;
    }
#endif /* SO_REUSEPORT */
#endif /* AIX */
#ifdef SO_LINGER
    linger_options.l_onoff = 0;
    linger_options.l_linger = 0;
    if (setsockopt(master_socket, SOL_SOCKET, SO_LINGER,
		   (char *)&linger_options, sizeof(linger_options)) == -1) {
	syslog(LOG_ERR, "server DONTLINGER: %m");
	errno = 0;
    }
#endif SO_LINGER

 /*
  * Bind the socket to port Port_number in the INET (internet) family, and
  * listen for any incoming connections. 
  */

    server.sin_family = AF_INET;
    server.sin_addr.s_addr = INADDR_ANY;
    for (port_number = Port_number;
	 port_number < Port_number + 3 && !bound;
	 port_number++) {
	server.sin_port = htons(port_number);
	if (!(bound = bind(master_socket, (struct sockaddr *) & server,
			   sizeof(server)) == 0)) {
#ifdef PTX
	    if (errno == EADDRNOTAVAIL)
		errno = EADDRINUSE;
#endif PTX
	    if (errno != EADDRINUSE) {
		syslog(LOG_ERR, "binding %d: %m", port_number);
		return 1;
	    }
	    errno = 0;
	}
    }

    if (!bound) {
	syslog(LOG_ERR, "Ports %d - %d are in use", Port_number, Port_number + 2);
	return 1;
    }
    port_number--;
    syslog(LOG_INFO, "bound to port %d", port_number - 1);
#ifdef DEBUG
    if (Debug)
	printf("Server bound to port %d.\n", port_number);
#endif DEBUG

 /* Initialize width to the size of our descriptor table. */
#ifdef PTX
    setdtablesize(128);
#endif PTX
    width = getdtablesize();

 /* Initialize database routines. */
    if (!database_init(readonly, override_locks)) {
	syslog(LOG_ERR, "database init failed: %m");
	if (Debug)
	    perror("Database init failed");
	return 1;
    }
    syslog(LOG_INFO, "messagebase opened%s", Read_only ? " (read-only)" : "");

#ifdef STATS
 /* Initialize stats */
    {
	int i;

	for (i = 0; i < 256; i++) {
	    stats[i].count = 0;
	    stats[i].accum.tv_sec = 0;
	    stats[i].accum.tv_usec = 0;
	    stats[i].max = 0;
	}
    }
#endif STATS

 /* Ignore broken pipes (they'll be detected later). */
    signal(SIGPIPE, SIG_IGN);

 /* Trap signals */
    signal(SIGHUP, hangup);
    signal(SIGTERM, terminate);
    signal(SIGINT, interrupt);
#ifdef STATS
    signal(SIGPROF, report_times);
#endif STATS

 /* Start accepting connections. */
    listen(master_socket, 5);

    while (!Server_Shutdown || Max_active_slot >= 0) {
	int len = sizeof(from);
	int code, i = Current_client, l;
	fd_set ready;
	static int select_errors = 0;

#ifdef DUMP_CACHE
	static time_t cache_check = 0;

#endif DUMP_CACHE
#ifdef STATS
	struct timeval start_time, end_time, delta;

#endif STATS

#ifdef Status
	sprintf(status, "(bb) */%d", Max_active_slot + 1);
	Status(status);
#endif Status

	flush_cache();
	Now = time(0);
#ifdef DUMP_CACHE
	if (max_active_slot < 0) {
	    dump_cache();
	    Now = time(0);
	} else if (Now - cache_check > CACHE_TIMEOUT) {
	    check_cache();
	    cache_check = Now = time(0);
	}
#endif DUMP_CACHE
	Now = time(0);
	FD_ZERO(&ready);
	FD_SET(master_socket, &ready);
	for (i = 0; i <= Max_active_slot; i++)
	    if (Client[i].inuse) {
		if (Now - Client[i].last_action > TIMEOUT) {
		    char errmsg[80];

		    sprintf(errmsg, "timed out: inactive for %d seconds",
			    (int)(Now - Client[i].last_action));
		    write_socket_code(Client[i].socket, SERVER_ERR);
		    write_socket_string(Client[i].socket, errmsg);
		    if (Client[i].signed_on)
			save_room_pointer(i);
		    close_client(i);
		} else
		    FD_SET(Client[i].socket, &ready);
	    }
	if (select(width, &ready, 0, 0, 0) < 0) {
#ifdef DEBUG
	    if (Debug)
		syslog(LOG_DEBUG, "select: %m");
#endif DEBUG

	/*
	 * This code was written to isolate problem sockets and drop them,
	 * when necessary.  If select fails 10 times in a row, we
	 * individually select on each client, and drop any that return an
	 * error. 
	 */

	    select_errors++;
	    if (select_errors > 10) {
		for (i = 0; i <= Max_active_slot; i++)
		    if (Client[i].inuse) {
			struct timeval timeout;

			FD_ZERO(&ready);
			FD_SET(Client[i].socket, &ready);

		    /*
		     * select() with an empty timeout structure returns
		     * immediately. 
		     */

			timeout.tv_sec = 0;
			timeout.tv_usec = 0;

			if (select(width, &ready, 0, 0, &timeout) == -1) {
#ifdef DEBUG
			    if (Debug)
				syslog(LOG_DEBUG, "select (%d:%d): %m",
				       i, Client[i].socket);
#endif DEBUG
			    close_client(i);
			}
		    }
	    }
	    continue;
	}
	select_errors = 0;

    /* If a connection is waiting on the master socket, . . . */

	if (FD_ISSET(master_socket, &ready)) {

	/* . . . accept it. */

	    Client[Free_slot].socket =
	      accept(master_socket, (struct sockaddr *) & from, &len);
	    if (Client[Free_slot].socket == -1)
		syslog(LOG_DEBUG, "accept: %m");
	    else {
	    /*
	     * Let data linger for 60 seconds after a close(). 
	     */
#ifdef SO_LINGER
		linger_options.l_onoff = 1;
		linger_options.l_linger = 60;
		if (setsockopt(Client[Free_slot].socket,
			     SOL_SOCKET, SO_LINGER, (char *)&linger_options,
			       sizeof(linger_options)) == -1)
		    syslog(LOG_WARNING, "client LINGER: %m");
#endif SO_LINGER
	    /*
	     * Automatically close on exec() 
	     */
		fcntl(Client[Free_slot].socket, F_SETFD, 1);
#ifdef READ_TIMEOUT
		fcntl(Client[Free_slot].socket, F_SETFL, FNDELAY);
#endif /* READ_TIMEOUT */

		if (Free_slot == MAX_CLIENTS ||
		    Client[Free_slot].socket >= width - 2) {

		/*
		 * If the server can't handle any more connections, let the
		 * client know what's going on, then drop it. 
		 */

		    write_socket_code(Client[Free_slot].socket, SERVER_ERR);
		    write_socket_string(Client[Free_slot].socket,
					"server full");
		    close(Client[Free_slot].socket);

		} else {
		    static struct hostent *host;

		/* Get information on connection, store in client array. */
		    host = gethostbyaddr((char *)&from.sin_addr.s_addr,
			     sizeof(from.sin_addr.s_addr), from.sin_family);
		    if (host)
			strcpy(Client[Free_slot].host, host->h_name);
		    else
			sprintf(Client[Free_slot].host, "[%s]",
				inet_ntoa(from.sin_addr));

		/* Acknowledge the connection. */
		    write_socket_code(Client[Free_slot].socket, SERVER_ACK);

		/*
		 * Update maximum active slot index, record the current slot
		 * as being in use, and find the next free slot. 
		 */

		    if (Free_slot > Max_active_slot)
			Max_active_slot = Free_slot;
		    Client[Free_slot].inuse = TRUE;
		    Client[Free_slot].errors = 0;
		    Client[Free_slot].last_action = time(0);
		    while (Client[++Free_slot].inuse);
		}
	    }
	}
    /*
     * For all currently inuse clients, check whether they have any
     * outstanding requests, and, if so, service them. 
     */

	for (Current_client = i = 0;
	     i <= Max_active_slot;
	     Current_client = ++i)
	    if (Client[i].inuse)
		if (FD_ISSET(Client[i].socket, &ready)) {
		    char c;

		    l = read_socket_code(Client[i].socket, &c);
		    code = (int)c;
		    if (l == -1) {
			if (++Client[i].errors > 50) {
			    syslog(LOG_WARNING, "reading client %d (%d): %m",
						 i, Client[i].socket);
			    close_client(i);
			}
			continue;
		    }
		    Client[i].last_action = time(0);
		    Client[i].errors = 0;

		    if (l == 0) {
			close_client(i);
			continue;
		    }
		/*
		 * These events can be serviced regardless of whether or not
		 * the client has already signed on. 
		 */

#ifdef Status
		    sprintf(status, "(bb) %d/%d",
			    Current_client, Max_active_slot + 1);
		    Status(status);
#endif Status

		    switch (code) {
		    case CLIENT_PROTOCOL:
			server_protocol(i);
			continue;
		    case CLIENT_TEST:
			write_socket_code(Client[i].socket, SERVER_ACK);
			continue;
		    default:;
		    }

		    if (!Client[i].major_version) {
			switch (code) {
			case CLIENT_SIGNON:
			    {
				char err_text[BUFFER_SIZE];
				char client_protocol[BUFFER_SIZE];

				read_socket_s(Client[i].socket,
					      client_protocol,
					      sizeof(client_protocol));

				sprintf(err_text,
					"incompatible client/server protocols\n\tclient: %s\n\tserver: @(#)protocol.h\t%d.%d\t%s",
					client_protocol,
					PROTOCOL_MAJOR, PROTOCOL_MINOR,
					__DATE__);
				write_socket_code(Client[i].socket,
						  SERVER_ERR);
				write_socket_string(Client[i].socket,
						    err_text);
				close_client(i);
				continue;
			    }
			default:;
			    write_socket_code(Client[i].socket, SERVER_ERR);
			    write_socket_string(Client[i].socket,
				    "incompatible client/server protocols");
			    continue;
			}
		    }
		    switch (code) {
		    case CLIENT_SIGNON:
			server_signon(i);
			continue;
		    case CLIENT_WHOIS:
			server_whois(i);
			continue;
		    case CLIENT_USERS:
			server_users(i);
			continue;
		    default:;
		    }

		    if (!Client[i].signed_on) {
			write_socket_code(Client[i].socket, SERVER_ERR);
			write_socket_string(Client[i].socket,
					    "you are not signed on");
			continue;
		    }
#ifdef Status
		    sprintf(status, "(bb) %d/%d (%d:%d)",
			    Current_client, Max_active_slot + 1,
			    (int)Client[i].uid, code);
		    Status(status);
#endif Status

#ifdef STATS
		    gettimeofday(&start_time, 0);
#endif STATS

		    switch (code) {

		    case CLIENT_CATCHUP:
			server_catchup(i);
			break;
		    case CLIENT_CHANGECLASS:
			server_changeclass(i);
			break;
		    case CLIENT_CHANGEROOM:
			server_changeroom(i);
			break;
		    case CLIENT_CREATE:
			server_create(i);
			break;
		    case CLIENT_CUTOFF:
			server_cutoff(i);
			break;
		    case CLIENT_DELETE:
			server_delete(i);
			break;
		    case CLIENT_DIRECTORY:
			server_directory(i);
			break;
		    case CLIENT_ROOMINFO:
			server_roominfo(i);
			break;
		    case CLIENT_PRIORITY:
			server_priority(i);
			break;
		    case CLIENT_GROUPADD:
			server_groupadd(i);
			break;
		    case CLIENT_GROUPCHANGE:
			server_groupchange(i);
			break;
		    case CLIENT_GROUPCHANGECLASS:
			server_groupchangeclass(i);
			break;
		    case CLIENT_GROUPCREATE:
			server_groupcreate(i);
			break;
		    case CLIENT_GROUPDIRECTORY:
			server_groupdirectory(i);
			break;
		    case CLIENT_GROUPLIST:
			server_grouplist(i);
			break;
		    case CLIENT_GROUPPRIORITY:
			server_grouppriority(i);
			break;
		    case CLIENT_GROUPREMOVE:
			server_groupremove(i);
			break;
		    case CLIENT_GROUPSUBTRACT:
			server_groupsubtract(i);
			break;
		    case CLIENT_KILL:
			server_kill(i);
			break;
		    case CLIENT_LIST:
			server_list(i);
			break;
		    case CLIENT_MESSAGE_INFO:
			server_db_message(i);
			break;
		    case CLIENT_MESSAGE_NUMBERS:
			server_message_numbers(i);
			break;
		    case CLIENT_NICK:
			server_nick(i);
			break;
		    case CLIENT_PASSWORD:
			server_passwd(i);
			break;
		    case CLIENT_POST:
			server_post(i);
			break;
		    case CLIENT_PRESERVE:
			server_preserve(i);
			break;
		    case CLIENT_READ:
			server_read(i);
			break;
		    case CLIENT_REMOVE:
			server_remove(i);
			break;
		    case CLIENT_RENAME:
			server_rename(i);
			break;
		    case CLIENT_SEARCH:
			server_search(i);
			break;
		    case CLIENT_SELECT:
			server_select(i);
			break;
		    case CLIENT_SETCURRENT:
			server_setcurrent(i);
			break;
		    case CLIENT_SETMAIL:
			server_setmail(i);
			break;
		    case CLIENT_SETNAME:
			server_setname(i);
			break;
		    case CLIENT_SHUTDOWN:
			server_shut(i);
			break;
		    case CLIENT_SIGNOFF:
			server_signoff(i);
			break;
		    case CLIENT_SUMMARY:
			server_summary(i);
			break;
		    case CLIENT_UNDELETE:
			server_undelete(i);
			break;
		    default:
			write_socket_code(Client[i].socket, SERVER_ERR);
			write_socket_string(Client[i].socket,
					    "unrecognized client code");
			close_client(i);
		    }

#ifdef STATS
		    gettimeofday(&end_time, 0);
		    delta = delta_time(start_time, end_time);
		    stats[code].count++;
		    stats[code].accum = add_times(stats[code].accum, delta);
		    stats[code].max = MAX(stats[code].max, delta.tv_sec);
#endif STATS
		}
    }
 /* Close databases. */
    database_shutdown();
    if (Server_Reboot) {
	server_reboot();
	return 1;
    } else {
	syslog(LOG_INFO, "shutting down");
	return 0;
    }
}

#ifdef STATS
struct timeval delta_time(struct timeval t1, struct timeval t2)
{
    struct timeval delta;

    delta.tv_sec = t2.tv_sec - t1.tv_sec;
    delta.tv_usec = t2.tv_usec - t1.tv_usec;
    if (delta.tv_usec < 0) {
	delta.tv_usec += 1000000;
	delta.tv_sec -= 1;
    }
    return delta;
}

struct timeval add_times(struct timeval t1, struct timeval t2)
{
    struct timeval sum;

    sum.tv_sec = t2.tv_sec + t1.tv_sec;
    sum.tv_usec = t2.tv_usec + t1.tv_usec;
    if (sum.tv_usec >= 1000000) {
	sum.tv_usec -= 1000000;
	sum.tv_sec += 1;
    }
    return sum;
}

signal_t report_times()
{
    int i;

    signal(SIGPROF, report_times);

    for (i = CLIENT_SIGNON; i <= CLIENT_GROUPLIST; i++) {
	if (stats[i].count > 0) {
	    int t = stats[i].accum.tv_sec / stats[i].count;

	    syslog(LOG_DEBUG, "code %3d: %8d (%d:%02d / 0:%02d)",
		   i, stats[i].count,
		   t / 60, t % 60,
		   stats[i].max);
	}
    }
    return (signal_t)0;
}

#endif STATS
